/*
 * compression.c
 *
 * stuff to deal with compression (mainly VQ for now)
 */
#include <stdio.h>
#include <stdlib.h>
#ifdef sgi
#include <bstring.h>
#endif
#include <string.h>

#include "vq_iface.h"
#include "lightfield.h"

static void
fillVQStruct(LFOps *ops, LFInternOp *op)
{
    LFSlab **slabs = ops->op_slabs;
    int nslab = ops->op_slab_cnt;
    float *descr = op->op_descr;
    LFVQCodebook *vq;
    int tile_u_bits = 0;
    int tile_v_bits = 0;
    int tile_s_bits = 2;
    int tile_t_bits = 2;
    int tile_p_bits[LF_MAXPARAMS] = { 0 };
    float train_pct = .01f;
    int code_size = 1 << 16;
    int i;

    while (*descr != LF_NULL) {
        switch((int)*descr++) {
        case LF_VQ_TILESIZE:
	    tile_u_bits = log2(*descr++);
	    tile_v_bits = log2(*descr++);
	    tile_s_bits = log2(*descr++);
	    tile_t_bits = log2(*descr++);
	    break;

        case LF_VQ_TRAIN_SIZE:
	    train_pct = *descr++;
	    break;

        case LF_VQ_CODESIZE:
	    code_size = *descr++;
	    break;

	case LF_VQ_PTILESIZE:
	  i = *descr++;
	  if(i >= 0 && i < LF_MAXPARAMS) {
	    tile_p_bits[i] = log2(*descr++);
	  } else {
	    if(log2(*descr++)) {
	      lfError("fillVQStruct: parameter #%d not in valid range [0, %d]\n", i, LF_MAXPARAMS);
	      exit(1);
	    }
	  }
	  break;

        default:
	    descr++;
	    break;
        }
    }

    for ( i=0 ; i<nslab ; i++ )
	if  (slabs[i]->shared->vq)
	    free(slabs[i]->shared->vq);

    vq = calloc(1, sizeof(LFVQCodebook));
    vq->ubits = tile_u_bits;
    vq->vbits = tile_v_bits;
    vq->sbits = tile_s_bits;
    vq->tbits = tile_t_bits;
    memcpy(vq->pbits, tile_p_bits, sizeof(tile_p_bits));
    vq->size  = code_size;
    vq->train_pct = train_pct;
    vq->codebook = NULL;
    vq->sample_size = slabs[0]->shared->sample_size;

    for ( i=0 ; i<nslab ; i++ )  {
	slabs[i]->shared->vq = vq;
	/* lightfield data stores data now */
	slabs[i]->sample_size = 2;
    }
}

static int p_power(int *p, int *pbits)
{
  int res;
  int p_level;

  for(p_level = LF_MAXPARAMS - 2, res = p[LF_MAXPARAMS - 1];
      p_level >= 0;
      p_level--) {
    res <<= pbits[p_level];
    res += p[p_level];
  }
  return res;
}
  

/* function to reshuffle data for VQ processing */
void
lfMemShuffle(LFSlab **slabs, int nslab, const int *pos,
	const float *descr, const void *input, int input_size,
	void *output, int output_size, void *aux_ptr)
{
    LFSlab *slab = slabs[pos[LF_SLAB_POS]];
    LFShared *shared = slab->shared;
    LFVQCodebook *vq = shared->vq;
    unsigned char *src_ptr, *dst_ptr;
    int tblk_mul = shared->nt >> vq->tbits;
    int pbits_sum = lfSumPBits(vq);
    int tile_shift = pbits_sum + vq->ubits + vq->vbits + vq->sbits + vq->tbits;
    int smsk = (1 << vq->sbits) - 1;
    int tmsk = (1 << vq->tbits) - 1;
    int sblk, tblk;
    int ssmp, tsmp;
    int u, v, s, t;
    int i, idx;
    int p[LF_MAXPARAMS];
    int block_size = (shared->ns*shared->nt*shared->sample_size) <<
		     (pbits_sum+vq->ubits+vq->vbits);
    bool_t cont;

    assert(block_size == input_size);
    assert(block_size == output_size);

    /*
     * repack training set so that all samples of the same VQ block
     * sit near one another
     */
    src_ptr=(char *)input;
    memset(p, 0, sizeof(p));
    cont = TRUE;
    do {
      for ( u=0; u<(1<<vq->ubits) ; u++ ) {
	for ( v=0 ; v<(1<<vq->vbits) ; v++ ) {
	  for ( s=0 ; s<shared->ns ; s++ )  {
	    sblk = s >> vq->sbits;
	    ssmp = s & smsk;
	    for ( t=0 ; t<shared->nt ; t++ )  {
	      tblk = t >> vq->tbits;
	      tsmp = t & tmsk;
	      idx = ((sblk * tblk_mul + tblk) << tile_shift) +
		(((((((p_power(p, vq->pbits)<<vq->ubits) + u)
		     <<vq->vbits)+v)
		   <<vq->sbits)+ssmp)
		 <<vq->tbits)+tsmp;
	      assert(idx < ((shared->ns * shared->nt) <<
			    (pbits_sum + vq->ubits + vq->vbits)));
	      dst_ptr  = (char *)output + shared->sample_size * idx;
	      for ( i=0 ; i<shared->sample_size ; i++ )
		*dst_ptr++ = *src_ptr++;
	    }
	  }
	}
      }

      /* increment p array */
      for(i = LF_MAXPARAMS - 1; i >= 0; i--) {
	if(++p[i] == 1<<vq->pbits[i]) {
	  p[i] = 0;
	} else {
	  break;
	}
      }
      if(i < 0) {
	cont = FALSE;
      }
    } while(cont);
}

/* check that we can iterate over all source images with the
   current VQ tile size */
static void
bailCheckTile(char *variable, int tile, int lf) {
  lfError("For VQ compression, the size of the light field has to be\n"
	  "a multiple of the tile size. This is not the case for\n"
	  "variable %s: tile size %d, light field size %d\n",
	  variable, tile, lf);
  exit(1);
}
static void
checkTileSize(LFSlab **slabs, int nslab)
{
  LFParams *params = __LF_GET_PARAMS();

  while(--nslab >= 0) {
    LFShared *sh = slabs[nslab]->shared;
    LFVQCodebook *vq = sh->vq;
    int i;

    if(sh->nu % (1<<vq->ubits)) {
      bailCheckTile("u", (1<<vq->ubits), sh->nu);
    }
    if(sh->nv % (1<<vq->vbits)) {
      bailCheckTile("v", (1<<vq->vbits), sh->nv);
    }
    if(sh->ns % (1<<vq->sbits)) {
      bailCheckTile("s", (1<<vq->sbits), sh->ns);
    }
    if(sh->nt % (1<<vq->tbits)) {
      bailCheckTile("t", (1<<vq->tbits), sh->nt);
    }
    for(i = 0; i < params->num; i++) {
      if(vq->pbits[i] &&
	 params->num > 1 &&
	 sh->samples) {
	lfError("Extending VQ compression on parameters that have been\n"
		"sampled at arbitrary positions is not supported. Please\n"
		"use a rectangular parameter sampling or limit the VQ\n"
		"tiles to uvst.\n");
	exit(1);
      }
      if(sh->max_np[i] % (1<<vq->pbits[i])) {
	char buffer[10];
	sprintf(buffer, "p #%d", i);
	bailCheckTile(buffer, (1<<vq->pbits[i]), sh->max_np[i]);
      }
    }
  }
}



/* function to initialize memory shuffling routine */
void *
lfBeginMemShuffle(LFOps *ops, LFInternOp *op)
{
    if  (ops->chain_msk & LF_SET_SLAB) {
	fillVQStruct(ops, op);
	checkTileSize(ops->op_slabs, ops->op_slab_cnt);
    }

    op->op_func  = lfMemShuffle;
    op->op_end   = lfEndDefault;
    op->op_update = lfUpdateNoop;

    return NULL;
}

/* function to generate a codebook */
void
lfVQGenCodebook(LFSlab **slabs, int nslab, const int *pos,
	const float *descr, const void *input, int input_size,
	void *output, int output_size, void *aux_ptr)
{
    LFVQCodebook *vq = slabs[0]->shared->vq;
    VQ_t *tree_vq;

    /* do VQ training */
    tree_vq = vqGenTreeCodebook(input, input_size, 1, vq->size,
	vq->sample_size << (lfSumPBits(vq)+vq->ubits+vq->vbits+vq->sbits+vq->tbits));
    vq->codebook = tree_vq;
    vq->size = tree_vq->nwords;	/* may generate a smaller codebook */
    vq->type = LF_VQ_TREE_VQ;
}

/* function to generate VQ code array */
void
lfVQCompress(LFSlab **slabs, int nslab, const int *pos,
	const float *descr, const void *input, int input_size,
	void *output, int output_size, void *aux_ptr)
{
    LFSlab *slab = slabs[pos[LF_SLAB_POS]];
    LFShared *shared = slab->shared;
    LFVQCodebook *vq = shared->vq;
    int block_size = (shared->ns * shared->nt * shared->sample_size) <<
		     (lfSumPBits(vq) + vq->ubits + vq->vbits);
    int last_bit_offset;
    int arr[(LF_MAXPARAMS + 4) * 2];
    int i;

    assert(input_size == block_size);
    assert(vq->type == LF_VQ_LINEAR_VQ || vq->type == LF_VQ_TREE_VQ);

    /* do VQ classification */
    arr[0] = 1 << vq->tbits;
    arr[1] = 1 << vq->sbits,
    arr[2] = 1 << vq->vbits,
    arr[3] = 1 << vq->ubits;
    for(i = 0; i < LF_MAXPARAMS; i++) {
      arr[4 + i] = 1 << vq->pbits[LF_MAXPARAMS - 1 - i];
    }
    arr[LF_MAXPARAMS + 4] = shared->nt >> vq->tbits,
    arr[LF_MAXPARAMS + 5] = shared->ns >> vq->sbits;
    for(i = 0; i < LF_MAXPARAMS + 2; i++) {
      arr[LF_MAXPARAMS + 6 + i] = 1;
    }

    vqEncode(vq->codebook, (unsigned char *)input, sizeof(arr)/sizeof(arr[0]), arr,
	vq->sample_size /* pixel size */,
	TRUE /* do padding */, 16 /* pad-to-short */, output, &last_bit_offset);

    assert(last_bit_offset == (output_size * 8));

    /* swap bytes - vqEncode() has problems with little endian */
#ifdef IS_LITTLE_ENDIAN
{
    unsigned short *words = output;
    while((output_size -= 2) >= 0) {
      *words = (((unsigned char)*words) << 8) | ((unsigned char)(*words >> 8));
      words++;
    }
}
#endif
}

void
lfEndVQCompress(LFOps *ops, LFInternOp *op, void *aux_ptr)
{
    LFShared *shared;
    int i;

    for ( i=0 ; i<ops->op_slab_cnt ; i++ )  {
	shared = ops->op_slabs[i]->shared;
	if  (shared->vq)
	    shared->sample_size = 2;
    }
}

/* function to make a linear VQ codebook array */
void
lfLinearizeVQCodebook(LFVQCodebook *vq)
{
  if (vq && vq->type != LF_VQ_LINEAR_DATA) {
    VQ_t *veeq = NULL;

    if  (vq->type == LF_VQ_TREE_VQ) {
      veeq = vqTreeCodebookToLinearCodebook(vq->codebook);
      vqFreeCodebook(vq->codebook);
    }
    else  if  (vq->type == LF_VQ_LINEAR_VQ)
      veeq = vq->codebook;
    vq->type = LF_VQ_LINEAR_DATA;
    vq->codebook = veeq->codebook;
  }
}
