%{
/*
 * descr_file_yacc.y
 *
 * parser for a lightfield description file 
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "lightfield.h"

extern LFSlab *__lf_descr_slabs;
extern LFParams   *__lf_descr_params;
extern LFRenderer *__lf_descr_renderer;

static LFSlab *cur_slab;
static LFVtx *cur_vtx;
static int *cur_points, cur_num_points, cur_max_points;

static int max_plans, cur_plan;

static int np[LF_MAXPARAMS];
static int *p_points[LF_MAXPARAMS];
static int sample_points[LF_MAXPARAMS];
static int num_sample_points;

struct Sample;
typedef struct Sample {
  struct Sample *next;
  int p[LF_MAXPARAMS];
} Sample;
static Sample *samples;

static LFParamSamples *cube(int num_params, int *np, int **p_points)
{
  LFParamSamples *psamp;
  int i;

  if(!num_params) {
    return NULL;
  }

  psamp = malloc(sizeof(LFParamSamples) + sizeof(LFSample) * (np[0] - 1));
  psamp->np = np[0];
  for(i = 0; i < np[0]; i++) {
    psamp->samples[i].p = p_points[0][i];
    psamp->samples[i].psamples = cube(num_params - 1, &np[1], &p_points[1]);
  }
  return psamp;
}

static LFParamSamples *addSample(int num_params, LFParamSamples *psamp, int *p)
{
  int e;

  if(!num_params) {
    return NULL;
  }

  if(psamp) {
    int i;

    for(i = 0; i < psamp->np; i++) {
      if(p[0] <= psamp->samples[i].p) {
	if(p[0] == psamp->samples[i].p) {
	  psamp->samples[i].psamples = addSample(num_params - 1, psamp->samples[i].psamples, &p[1]);
	  return(psamp);
	}
	break;
      }
    }
    psamp = realloc(psamp, sizeof(LFParamSamples) + psamp->np * sizeof(LFSample));
    for(e = psamp->np; e > i; e--) {
      psamp->samples[e] = psamp->samples[e - 1];
    }
  } else {
    psamp = calloc(sizeof(*psamp), 1);
    e = 0;
  }

  psamp->np++;
  psamp->samples[e].p = p[0];
  psamp->samples[e].psamples = addSample(num_params - 1, NULL, &p[1]);
  return psamp;
}

void setupSamples(void)
{
  int i;
  bool_t rect_grid = FALSE;
  LFShared *sh = cur_slab->shared;

  /* add rectangular grid? */
  for(i = 0; i < __lf_descr_params->num; i++) {
    if(np[i]) {
      rect_grid = TRUE;
    } else {
      np[i] = 1;
      p_points[i] = malloc(sizeof(int));
      *p_points[i] =  __lf_descr_params->params[i].def;
    }
  }
  if(rect_grid || !samples) {
    sh->psamples = cube(__lf_descr_params->num, np, p_points);
    sh->rectangular = TRUE;
  } else {
    sh->psamples = NULL;
  }

  /* add all samples */
  if(samples) {
    sh->samples = TRUE;
  }
  while(samples) {
    struct Sample *tmp = samples;

    sh->psamples = addSample(__lf_descr_params->num, sh->psamples, samples->p);
    samples = tmp->next;
    free(tmp);
  }

  /* now calculate maxima for each parameter */
  lfCalcMaxNP(__lf_descr_params->num, sh->psamples, sh->max_np);

  for(i = 0; i < __lf_descr_params->num; i++) {
    free(p_points[i]);
    p_points[i] = NULL;
  }
}

static void
initInterpolation(void)
{
  LFShared *sh = cur_slab->shared;

  if(sh->plans) {
    return;
  }

  if(__lf_descr_params->num > 2) {
    lfError("interpolation plan not supported for more than two parameters");
    exit(10);
  }
  max_plans = (__lf_descr_params->params[1].max - __lf_descr_params->params[1].min + 1) *
    (__lf_descr_params->params[0].max - __lf_descr_params->params[0].min + 1);
  cur_plan = 0;
  sh->plans = calloc(1, max_plans * sizeof(LFInterpolationPlan));
  if(!sh->plans) {
    lfError("no mem");
    exit(10);
  }
}

%}
%union {
	float	f;
	int	i;
	char	*s;
}
%token	<i> INTEGER
%token	<f> REAL
%token	<s> FILEPATH
%token  RENDERER VOLPACK VOLVIS RENDERMAN DUMMY QUALITY
%type   <i> RendererType
%token  PARAMETER
%token	RGBA RGB
%token	SLAB FORMAT SLAB_UV SLAB_ST SAMPLES_UV SAMPLES_ST
%token  SAMPLES_P SAMPLE_POINTS_P SAMPLE_POINT
%token  INTERPOLATION
%type	<i> ColorFormat
%type 	<f> Number
%%
File            : Params Renderer Field
                ;

Params          : /* empty */
                | Params Param
                ;

Param           : PARAMETER INTEGER INTEGER INTEGER FILEPATH
                {
		  if(__lf_descr_params->num >= LF_MAXPARAMS) {
		    lfError("lfBeginReadSliceFile: too many slabs\n");
		    exit(1);
		  }
		  __lf_descr_params->params[__lf_descr_params->num].min   = $2;
		  __lf_descr_params->params[__lf_descr_params->num].max   = $3;
		  __lf_descr_params->params[__lf_descr_params->num].current =
		  __lf_descr_params->params[__lf_descr_params->num].def   = $4;
		  __lf_descr_params->params[__lf_descr_params->num].descr = strdup($5);
		  __lf_descr_params->num++;
                }
                ;

Renderer        : /* empty */
                | RENDERER RendererType FILEPATH RendererArgs RendererOpts
                {
		        __lf_descr_renderer->type = $2;
			__lf_descr_renderer->config = strdup($3);
                }
                ;

RendererType	: VOLPACK
		{
			$$ = LF_RENDERER_VOLPACK;
		}
                | VOLVIS
		{
			$$ = LF_RENDERER_VOLVIS;
		}
                | RENDERMAN
		{
			$$ = LF_RENDERER_RENDERMAN;
		}
                | DUMMY
                {
		        $$ = LF_RENDERER_DUMMY;
                }
		;


RendererOpts    : /* empty */
                {
		  __lf_descr_renderer->quality = 0xFFFF;
		}
                | QUALITY INTEGER RendererOpts
                {
		  __lf_descr_renderer->quality = $2;
		}
                ;

RendererArgs    : /* empty */
                | RendererArgs INTEGER 
                {
		  if(__lf_descr_renderer->num_args >= __lf_descr_params->num) {
		    lfError("lfBeginReadSliceFile: only %d parameters available to renderer\n", __lf_descr_params->num);
		    exit(1);
		  }
		  __lf_descr_renderer->arg_indices[__lf_descr_renderer->num_args] = $2;
		  __lf_descr_renderer->num_args++;
		}
                ;


Field		: /* empty */
		| Field Slab
		;

Slab		: SlabHdr SlabAttribs
                {
                        setupSamples();
		}
		;

SlabAttribs	: /* empty */
		| SlabAttribs SlabAttrib
		;

SlabAttrib	: SlabFormat
		| SlabGeomUV
		| SlabGeomST
		| SlabSampleUV
		| SlabSampleST
                | SlabParams
		;

SlabHdr		: SLAB INTEGER FILEPATH
                {
		        int e;

			if  ($2 >= LF_MAXSLABS)  {
				lfError("lfBeginReadSliceFile: too many slabs\n");
				exit(1);
			}
			cur_slab = &__lf_descr_slabs[$2];
			cur_slab->lightfield = strdup($3);

			for(e = 0; e < LF_MAXPARAMS; e++) {
			  np[e] = 0;
			  p_points[e] = NULL;
			}
		}
		;

SlabFormat	: FORMAT ColorFormat
		{
			cur_slab->shared->sample_size = $2;
/*				($2 == 4) ? 4 : 3;
*/
		}
		;

SlabGeomUV	: SLAB_UV	{ cur_vtx = cur_slab->uv; }
		Vert Vert Vert Vert
		;

SlabGeomST	: SLAB_ST	{ cur_vtx = cur_slab->st; }
		Vert Vert Vert Vert
		;

SlabSampleUV	: SAMPLES_UV INTEGER INTEGER
		{
			cur_slab->shared->nu = $2;
			cur_slab->shared->nv = $3;
		}
		;

SlabSampleST	: SAMPLES_ST INTEGER INTEGER
		{
			cur_slab->shared->ns = $2;
			cur_slab->shared->nt = $3;
		}
		;

SlabParams      : /* empty */
                | SlabParam SlabParams
                ;

SlabParam       : SAMPLES_P INTEGER INTEGER
                {
		        int i;

			if(__lf_descr_params->num <= $2) {
			  lfError("lfBeginReadSliceFile: invalid SAMPLES_P %d - no such parameter\n", $2);
			  exit(1);
			}

		        np[$2] = $3;
			p_points[$2] = malloc($3 * sizeof(int));
			if($3 <= 1) {
			  p_points[$2][0] = __lf_descr_params->params[$2].def;
			} else {
			  int span = __lf_descr_params->params[$2].max - __lf_descr_params->params[$2].min;

			  for(i = 0; i < $3; i++) {
			    p_points[$2][i] = __lf_descr_params->params[$2].min +
			      span * i / ($3 - 1);
			  }
			}
		}
                | SAMPLE_POINTS_P INTEGER SlabParamPoints
                {
			if(__lf_descr_params->num <= $2) {
			  lfError("lfBeginReadSliceFile: invalid SAMPLE_POINTS_P %d - no such parameter\n", $2);
			  exit(1);
			}
			if(cur_num_points) {
			  if(cur_points[0] < __lf_descr_params->params[$2].min ||
			     cur_points[cur_num_points - 1] > __lf_descr_params->params[$2].max) {
			    lfError("lfBeginReadSliceFile: parameter %d sample points not within [%d, %d]\n",
				    $2,
				    __lf_descr_params->params[$2].min,
				    __lf_descr_params->params[$2].max);
			    exit(1);
			  }

			  p_points[$2] = malloc(cur_num_points * sizeof(int));
			  memcpy(p_points[$2], cur_points, cur_num_points * sizeof(int));
			  np[$2] = cur_num_points;
			  cur_num_points = 0;
			}
		}
                | SAMPLE_POINT SlabSamplePoints
                {
			Sample *sample;

			for(; num_sample_points < __lf_descr_params->num; num_sample_points++) {
			  sample_points[num_sample_points] = __lf_descr_params->params[num_sample_points].def;
			}
			sample = malloc(sizeof(*sample));
			memcpy(sample->p, sample_points, sizeof(sample->p));
			sample->next = samples;
			samples = sample;
			num_sample_points = 0;
		}
                | INTERPOLATION 
                { initInterpolation(); }
                  InterpPlan
                {
		        if(cur_plan != max_plans) {
			  lfError("incomplete plan!\n");
			  exit(10);
			}
		}
                ;

InterpPlan      : /* empty */
                | InterpPlan FILEPATH
                {
		  

		        /* doing our own parsing is easier than using yacc here */
			if(cur_plan < max_plans) {
			  LFInterpolationPlan *plan = &cur_slab->shared->plans[cur_plan];
			  char *in = $2;
			  int i;

#define MOVE in = strchr(in, '_'); if(!in) {lfError("bad plan format: %s\n", $2); exit(10);} in++;

			  if(*in) {
			    in++;
			  }
			  plan->num_accesses = atoi(in);
			  if(plan->num_accesses > LF_MAX_SOURCES) {
			    lfError("invalid number of sources (%d > %d)\n", plan->num_accesses, LF_MAX_SOURCES);
			    exit(10);
			  }
			  for(i = 0; i < plan->num_accesses; i++) {
			    int e;

			    for(e = 0; e < __lf_descr_params->num; e++) {
			      MOVE;
			      plan->p_indices[i][e] = atoi(in);
			    }
			  }

			  MOVE;
			  plan->num_interp = atoi(in);
			  if(plan->num_interp > LF_MAX_INTERP) {
			    lfError("invalid number of interpolations (%d > %d)\n", plan->num_interp, LF_MAX_INTERP);
			    exit(10);
			  }
			  for(i = 0; i < plan->num_interp; i++) {
			    MOVE;
			    plan->source[i] = atoi(in);
			    MOVE;
			    plan->target[i] = atoi(in);
			    MOVE;
			    plan->fp[i] = atof(in) * (1 << FP_BITS);
			    plan->fp_bits[i] = FP_BITS;
			  }
			  cur_plan++;
			}
		}
                ;

SlabSamplePoints : /* empty */
                | SlabSamplePoints INTEGER
                {
			bool_t dump_points = TRUE;

		        if(num_sample_points >= __lf_descr_params->num) {
			  lfError("more than %d parameter values in SAMPLE_POINT: ",
				  __lf_descr_params->num);
			} else if($2 > __lf_descr_params->params[num_sample_points].max ||
				  $2 < __lf_descr_params->params[num_sample_points].min) {
			  lfError("value %d for parameter #%d in SAMPLE_POINT not in [%d,%d]: ",
				  $2, num_sample_points,
				  __lf_descr_params->params[num_sample_points].min,
				  __lf_descr_params->params[num_sample_points].max);
			} else {
                           dump_points = FALSE;
			}

                        if(dump_points) {
		          int i;
			  for(i = 0; i < num_sample_points; i++) {
			    lfError("%d ", sample_points[i]);
			    exit(1);
			  }
			} else {
			  sample_points[num_sample_points++] = $2;
			}
		}
                ;

SlabParamPoints : /* empty */
                | INTEGER SlabParamPoints
                {
                        int i;

			if(cur_num_points == cur_max_points) {
			  cur_max_points += 10;
			  cur_points = realloc(cur_points, cur_max_points * sizeof(int));
			}
			for(i = cur_num_points - 1; i >= 0; i--) {
			  if($1 < cur_points[i]) {
			    cur_points[i + 1] = cur_points[i];
			  } else {
			    break;
			  }
			}
			cur_points[i + 1] = $1;
			cur_num_points++;
		}
                ;


ColorFormat	: RGBA
		{
			$$ = 4;
		}
		| RGB
		{
			$$ = 3;
		}
		;

Vert		: Number Number Number Number Number Number
		{
			cur_vtx->ox = $1;
			cur_vtx->oy = $2;
			cur_vtx->oz = $3;
			cur_vtx->ow = $4;
			cur_vtx->s  = $5;
			cur_vtx->t  = $6;
			cur_vtx++;
		}
		;

Number		: REAL
		| INTEGER	{ $$ = (float)$1; }
		;
%%
void
yyerror(char *s)
{
    extern int yylinenumber;

    lfError("Error on line %d: %s\n", yylinenumber, s);
    exit(1);
}
