/*
 *  fileIO.c: read / write a lightfield in the new .lif file format
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "lightfield.h"
#include "fileIO.h"
#include "lerp.h"


/* Local helper functions */
void *SlabSlice(LFSlab *slab, int u, int v);
void ClearHdr(LFFileHdr *hdr, FILE *file);
int GetHdrLen(FILE *file);
void WriteRawBlock(FILE *outFile, int offset, char *buffer, int elemCount,
		   int elemInSize, int elemOutSize);
void ReadRawBlock(FILE *inFile, int offset, char *buffer, int elemCount,
		  int elemInSize, int elemOutSize);
int NextHdrToken(FILE *file, char *buffer, int maxTokenSize);
int AssertToken(FILE *file, char *token);
int SkipSegment(FILE *inFile);
int ParseSlabHdr(FILE *inFile, LFFileHdr *hdr, LFSlab *slabs[], int nslabs);
int ParseVQHdr(FILE *inFile, LFFileHdr *hdr, LFSlab *slabs[], int nslabs);
static int ParseLFParams(FILE *inFile, LFFileHdr *hdr, LFParams *params);
static int ParseLFRenderer(FILE *inFile, LFFileHdr *hdr, LFRenderer *renderer);
int Log2(unsigned long n);
static LFParamSamples *readPSamples(FILE *inFile, int num_params, char *buffer);
static void swapPixels(char *buffer, int byte_count, int elem_size);


/* ----------------------------------------------------------------------
 * -         High-level Functions removed, because not adapted       ---- 
 * ----------------------------------------------------------------------
 */


/* ----------------------------------------------------------------------
 * ---------------          Low-level Functions           --------------- 
 * ----------------------------------------------------------------------
 */



/* Open a Lightslab file, malloc the LFField and LFFileHdr structs,
 * and initialize them from the file header info.
 */
FILE *lfOpenInFile(char *fileName, LFFileHdr **hdr, LFSlab *slabs[], 
		   int nslabs)
{
  FILE *inFile;
  char buffer[200];
  int slabn;
  long currPos;

  /* Open the input file */
#ifdef sgi
  inFile = fopen(fileName, "rb");
#else /* WIN32 */
  inFile = fopen(fileName, "r+b");
#endif 
  if (!inFile) {
    lfError("Error: cannot open file %s\n", fileName);
    return(NULL);
  }
  /* Malloc the slabHdr array */
  *hdr = (LFFileHdr *) calloc(1, sizeof(LFFileHdr));
  if (*hdr == NULL) {
    lfError("Error: No memory for slabHdr struct.\n");
    return(NULL);
  }

  /* Figure out hdr length */
  (*hdr)->hdrSize = GetHdrLen(inFile);
  lfOutput("Reading header of length %d\n", (*hdr)->hdrSize);
  
  /* Allocate a shared structure */
  for (slabn = 0; slabn < LF_MAXSLABS; slabn++) {
    slabs[slabn]->shared = (LFShared *) calloc(1, sizeof(LFShared));
  }

  /* Make sure we're at beginning of the file */
  rewind(inFile);

  /* Read the hdr, and use it to fill in the data for enabled slabs */
  AssertToken(inFile, "LIF1.0");
  AssertToken(inFile, "datasize");
  NextHdrToken(inFile, buffer, 200);
  (*hdr)->dataSize = atoi(buffer);
  AssertToken(inFile, "bgnlightfield");
  NextHdrToken(inFile, buffer, 200);
  /* Note: ignore lightfield number for now */

  /* First, read all the codebook headers */
  currPos = ftell(inFile);
  while (NextHdrToken(inFile, buffer, 200)) {
    if (!strcmp(buffer, "endlightfield")) break;
    if (!strcmp(buffer, "bgnsegment")) {
      NextHdrToken(inFile, buffer, 200);
      /* if it's a vq segment, parse the header */
      if (!strcmp(buffer, "vq")) {
	ParseVQHdr(inFile, *hdr, slabs, nslabs);
	/* If it's not a vq segment, skip it for now... */
      } else {
	SkipSegment(inFile);
      }
    }
  }
  /* Restore file position to start of the lightfield */
  fseek(inFile, currPos, SEEK_SET);
	
  /* Now we can parse all the slabs */
  while (1) {
    /* Read next token */
    if (!NextHdrToken(inFile, buffer, 200)) {
      lfError("Error: Unexpected end of file header.\n");
      return(0);
    }
    /* Check if endlightfield */
    if (!strcmp(buffer, "endlightfield")) {
      break;
    }
    /* Check if bgnsegment */
    if (!strcmp(buffer, "bgnsegment")) {
      NextHdrToken(inFile, buffer, 200);
      if (!strcmp(buffer, "slab")) {
	ParseSlabHdr(inFile, *hdr, slabs, nslabs);
      } else if (!strcmp(buffer, "params")) {
	ParseLFParams(inFile, *hdr, __LF_GET_PARAMS());
      } else if (!strcmp(buffer, "renderer")) {
	ParseLFRenderer(inFile, *hdr, __LF_GET_RENDERER());
      } else {
	SkipSegment(inFile);
      }
    } else {
      /* default -- unrecognized token */
      lfError("Error: unexpected token: \"%s\"\n", buffer);
      return 0;
    }
  }
  /* Return file */
  return(inFile);
}

/* This routine reads the rest of the segment. */
int SkipSegment(FILE *inFile)
{
  char buffer[200];

  while (NextHdrToken(inFile, buffer, 200)) {
    if (!strcmp(buffer, "endsegment")) return 1;
  }
  /* if we get here, ran out of header without endsegment */
  return 0;
}

/* Parse a VQ codebook header.  It assumes that "bgnsegment vq" has already
 * been parsed.
 */
int ParseVQHdr(FILE *inFile, LFFileHdr *hdr, LFSlab *slabs[], int nslabs)
{
  char buffer[200];
  int vqn;

  /* get codebook number */
  NextHdrToken(inFile, buffer, 200);
  vqn = atoi(buffer);
  lfOutput("ParseVQHdr %d\n", vqn);
  /* Note: does not check to see if slab is enabled... */
  lfOutput("reading slab hdr info for VQ codebook %d\n", vqn);

  /* Allocate the codebook structure */
  if (!((hdr->vqHdr[vqn].codebook = calloc(1, sizeof(LFVQCodebook))))) {
    lfError("Error: Out of memory in ParseVQHdr\n");
    return 0;
  }
  /* Parse the rest of the segment */
  while (NextHdrToken(inFile, buffer, 200)) {
    if (!strcmp(buffer, "endsegment")) return 1;

    if (!strcmp(buffer, "format")) {
      NextHdrToken(inFile, buffer, 200);
      if (!strcmp(buffer, "index")) {
	lfError("Error: Codebook %d has format index?!?\n", vqn);
	hdr->vqHdr[vqn].format = LF_INDEXCHANNEL;
	return 0;
      } else if (!strcmp(buffer, "rgb")) {
	hdr->vqHdr[vqn].format = LF_RGBCHANNEL;
	hdr->vqHdr[vqn].codebook->sample_size = 3;
      } else if (!strcmp(buffer, "rgba")) {
	hdr->vqHdr[vqn].format = LF_RGBACHANNEL;
	hdr->vqHdr[vqn].codebook->sample_size = 4;
      }
    } 

    if (!strcmp(buffer, "bgnchannel")) {
      NextHdrToken(inFile, buffer, 200);
      /* VQ codebook shouldn't be vq encoded! */
      AssertToken(inFile, "type");
      NextHdrToken(inFile, buffer, 200);
      if (!strcmp(buffer, "int8x3")) {
	hdr->vqHdr[vqn].type = LF_INT8x3;
      } else if (!strcmp(buffer, "int8x4")) {
	hdr->vqHdr[vqn].type = LF_INT8x4;
      }
      /* offset */
      AssertToken(inFile, "offset");
      NextHdrToken(inFile, buffer, 200);
      hdr->vqHdr[vqn].offset = atoi(buffer);
      /* size */
      AssertToken(inFile, "size");
      NextHdrToken(inFile, buffer, 200);
      hdr->vqHdr[vqn].size = atoi(buffer);
      lfOutput("vq %d size %d\n", vqn, hdr->vqHdr[vqn].size);
      /* end channel */
      AssertToken(inFile, "endchannel");
    }

    /* number of tiles in codebook */
    if (!strcmp(buffer, "tiles")) {
      NextHdrToken(inFile, buffer, 200);
      hdr->vqHdr[vqn].codebook->size = atoi(buffer);
    }

    /* Size of tiles in codebook */
    if (!strcmp(buffer, "tilesize")) {
      NextHdrToken(inFile, buffer, 200);
      hdr->vqHdr[vqn].codebook->ubits = Log2(atoi(buffer));
      NextHdrToken(inFile, buffer, 200);
      hdr->vqHdr[vqn].codebook->vbits = Log2(atoi(buffer));
      NextHdrToken(inFile, buffer, 200);
      hdr->vqHdr[vqn].codebook->sbits = Log2(atoi(buffer));
      NextHdrToken(inFile, buffer, 200);
      hdr->vqHdr[vqn].codebook->tbits = Log2(atoi(buffer));
    }

    /* Size of tiles in additional parameters */
    if (!strcmp(buffer, "ptilesize")) {
      int i, n;

      NextHdrToken(inFile, buffer, 200);
      n = atoi(buffer);
      if(n < 0 || n > LF_MAXPARAMS) {
	lfError("cannot handle codebook with %d parameters\n", n);
	return(0);
      }
      for(i = 0; i < n; i++) {
	NextHdrToken(inFile, buffer, 200);
	hdr->vqHdr[vqn].codebook->pbits[i] = Log2(atoi(buffer));
      }
    }
  }
  /* if we get here, then we ran out of header without endsegment */
  return(0);
}

/* Parse a slab header.  It assumes that "bgnsegment slab" has already
 * been parsed.
 */
int ParseSlabHdr(FILE *inFile, LFFileHdr *hdr, LFSlab *slabs[],
		 int nslabs)
{
  char buffer[200];
  int slabn, vqn;
  int vtx;

  /* get slab number */
  NextHdrToken(inFile, buffer, 200);
  slabn = atoi(buffer);
  /* Note: does not check to see if slab is enabled... */
  lfOutput("reading slab hdr info for slab %d\n", slabn);
  slabs[slabn]->id = slabn;

  /* Parse the rest of the segment */
  while (NextHdrToken(inFile, buffer, 200)) {
    if (!strcmp(buffer, "endsegment")) return 1;

    if (!strcmp(buffer, "compression")) {
      NextHdrToken(inFile, buffer, 200);
      if (!strcmp(buffer, "vq")) {
	NextHdrToken(inFile, buffer, 200);
	vqn = atoi(buffer);
	slabs[slabn]->shared->vq = hdr->vqHdr[vqn].codebook;
      } else {
	/* no compression */
	slabs[slabn]->shared->vq = NULL;
      }
    }

    if (!strcmp(buffer, "format")) {
      NextHdrToken(inFile, buffer, 200);
      if (!strcmp(buffer, "index")) {
	hdr->slabHdr[slabn].format = LF_INDEXCHANNEL;
	slabs[slabn]->shared->sample_size = 
	  (slabs[slabn]->shared->vq->size < 257) ? 1 :
	  (slabs[slabn]->shared->vq->size < 65537) ? 2 : 4;
      } else if (!strcmp(buffer, "rgb")) {
	hdr->slabHdr[slabn].format = LF_RGBCHANNEL;
	slabs[slabn]->shared->sample_size = 3;
      } else if (!strcmp(buffer, "rgba")) {
	hdr->slabHdr[slabn].format = LF_RGBACHANNEL;
	slabs[slabn]->shared->sample_size = 4;
      }
    } 

    if (!strcmp(buffer, "bgnchannel")) {
      NextHdrToken(inFile, buffer, 200);
      if (!strcmp(buffer, "index")) {
	AssertToken(inFile, "type");
	NextHdrToken(inFile, buffer, 200);
	if (!strcmp(buffer, "int8")) {
	  hdr->slabHdr[slabn].type = LF_INT8;
	} else if (!strcmp(buffer, "int16")) {
	  hdr->slabHdr[slabn].type = LF_INT16;
	} else if (!strcmp(buffer, "int32")) {
	  hdr->slabHdr[slabn].type = LF_INT32;
	}
      } else {
	/* Not vq encoded */
	AssertToken(inFile, "type");
	NextHdrToken(inFile, buffer, 200);
	if (!strcmp(buffer, "int8x3")) {
	  hdr->slabHdr[slabn].type = LF_INT8x3;
	} else if (!strcmp(buffer, "int8x4")) {
	  hdr->slabHdr[slabn].type = LF_INT8x4;
	}
      }
      /* offset */
      AssertToken(inFile, "offset");
      NextHdrToken(inFile, buffer, 200);
      hdr->slabHdr[slabn].offset = atoi(buffer);
      /* size */
      AssertToken(inFile, "size");
      NextHdrToken(inFile, buffer, 200);
      hdr->slabHdr[slabn].size = atoi(buffer);
      lfOutput("slab %d size %d\n", slabn, hdr->slabHdr[slabn].size);
      /* end channel */
      AssertToken(inFile, "endchannel");
    }
	  
    /* samples size */
    if (!strcmp(buffer, "samples_uv")) {
      NextHdrToken(inFile, buffer, 200);
      slabs[slabn]->shared->nu = atoi(buffer);
      NextHdrToken(inFile, buffer, 200);
      slabs[slabn]->shared->nv = atoi(buffer);
    } else if (!strcmp(buffer, "samples_st")) {
      NextHdrToken(inFile, buffer, 200);
      slabs[slabn]->shared->ns = atoi(buffer);
      NextHdrToken(inFile, buffer, 200);
      slabs[slabn]->shared->nt = atoi(buffer);
    } else if (!strcmp(buffer, "samples_p")) {
      int n;

      NextHdrToken(inFile, buffer, 200);
      n = atoi(buffer);
      if(n < 0 || n > LF_MAXPARAMS) {
	lfError("cannot handle slab with %d parameters\n", n);
	return(0);
      }
      slabs[slabn]->shared->psamples = readPSamples(inFile, n, buffer);
      lfCalcMaxNP(n, slabs[slabn]->shared->psamples, slabs[slabn]->shared->max_np);
    } else if (!strcmp(buffer, "rectangular")) {
      slabs[slabn]->shared->rectangular = TRUE;
    } else if (!strcmp(buffer, "samples")) {
      slabs[slabn]->shared->samples = TRUE;
    }

    /* interpolation */
    if (!strcmp(buffer, "interpolation")) {
      int n, params, i;
      LFInterpolationPlan *plan;

      NextHdrToken(inFile, buffer, 200);
      n = atoi(buffer);
      NextHdrToken(inFile, buffer, 200);
      params = atoi(buffer);
      if(!(plan = calloc(sizeof(*plan), n))) {
	lfError("no memory for %d plans\n");
	return(0);
      }
      slabs[slabn]->shared->plans = plan;
      for(i = 0; i < n; i++) {
	int j;

	NextHdrToken(inFile, buffer, 200);
	plan->num_accesses = atoi(buffer);
	for(j = 0; j < plan->num_accesses; j++) {
	  int e;

	  for(e = 0; e < params; e++) {
	    NextHdrToken(inFile, buffer, 200);
	    plan->p_indices[j][e] = atoi(buffer);
	  }
	}

	NextHdrToken(inFile, buffer, 200);
	plan->num_interp = atoi(buffer);
	for(j = 0; j < plan->num_interp; j++) {
	  NextHdrToken(inFile, buffer, 200);
	  plan->source[j] = atoi(buffer);
	  NextHdrToken(inFile, buffer, 200);
	  plan->target[j] = atoi(buffer);
	  NextHdrToken(inFile, buffer, 200);
	  plan->fp[j] = atof(buffer) * (1<<FP_BITS);
	  plan->fp_bits[j] = FP_BITS;
	}

	plan++;
      }
    }
	
    /* geometry */
    if (!strcmp(buffer, "geometry_uv")) {
      for (vtx=0; vtx < 4; vtx++) {
	NextHdrToken(inFile, buffer, 200);
	slabs[slabn]->uv[vtx].ox = atof(buffer);
	NextHdrToken(inFile, buffer, 200);
	slabs[slabn]->uv[vtx].oy = atof(buffer);
	NextHdrToken(inFile, buffer, 200);
	slabs[slabn]->uv[vtx].oz = atof(buffer);
	NextHdrToken(inFile, buffer, 200);
	slabs[slabn]->uv[vtx].ow = atof(buffer);
	NextHdrToken(inFile, buffer, 200);
	slabs[slabn]->uv[vtx].s = atof(buffer);
	NextHdrToken(inFile, buffer, 200);
	slabs[slabn]->uv[vtx].t = atof(buffer);
      }
    }
    if (!strcmp(buffer, "geometry_st")) {
      for (vtx=0; vtx < 4; vtx++) {
	NextHdrToken(inFile, buffer, 200);
	slabs[slabn]->st[vtx].ox = atof(buffer);
	NextHdrToken(inFile, buffer, 200);
	slabs[slabn]->st[vtx].oy = atof(buffer);
	NextHdrToken(inFile, buffer, 200);
	slabs[slabn]->st[vtx].oz = atof(buffer);
	NextHdrToken(inFile, buffer, 200);
	slabs[slabn]->st[vtx].ow = atof(buffer);
	NextHdrToken(inFile, buffer, 200);
	slabs[slabn]->st[vtx].s = atof(buffer);
	NextHdrToken(inFile, buffer, 200);
	slabs[slabn]->st[vtx].t = atof(buffer);
      }
    }
  }
  /* if we get here, then we ran out of header without endsegment */
  return(0);
}

/* recursively read all psample values */
static LFParamSamples *
readPSamples(FILE *inFile, int num_params, char *buffer)
{
  int np, p;
  LFParamSamples *psamples;

  if(!num_params) {
    return NULL;
  }

  NextHdrToken(inFile, buffer, 200);
  np = atoi(buffer);
  psamples = malloc(sizeof(LFParamSamples) + (np - 1) * sizeof(LFSample));
  psamples->np = np;
  for(p = 0; p < np; p++) {
    NextHdrToken(inFile, buffer, 200);
    psamples->samples[p].p = atoi(buffer);
    psamples->samples[p].psamples = readPSamples(inFile, num_params - 1, buffer);
  }
  return psamples;
}


/* Parse a parameters header.  It assumes that "bgnsegment params" has already
 * been parsed.
 */
static int
ParseLFParams(FILE *inFile, LFFileHdr *hdr, LFParams *params)
{
  char buffer[200];
  int i, n = 0;

  NextHdrToken(inFile, buffer, 200);
  n = atoi(buffer);
  if(n < 0 || n > LF_MAXPARAMS) {
    lfError("cannot handle %d parameters\n", n);
    return(0);
  }
  params->num = n;

  /* Parse the rest of the segment */
  while (NextHdrToken(inFile, buffer, 200)) {  
    if (!strcmp(buffer, "endsegment")) return 1;

    if (!strcmp(buffer, "params")) {
      for(i = 0; i < n; i++) {
	NextHdrToken(inFile, buffer, 200);
	params->params[i].min = atoi(buffer);
	NextHdrToken(inFile, buffer, 200);
	params->params[i].max = atoi(buffer);
	NextHdrToken(inFile, buffer, 200);
	params->params[i].current =
	  params->params[i].def = atoi(buffer);
	NextHdrToken(inFile, buffer, 200);
	params->params[i].descr = strdup(buffer);
      }
    }
  }

  return 0;
}


/* Parse a renderer header.  It assumes that "bgnsegment renderer" has already
 * been parsed.
 */
static int
ParseLFRenderer(FILE *inFile, LFFileHdr *hdr, LFRenderer *renderer)
{
  char buffer[200];
  int i, n = 0;

  NextHdrToken(inFile, buffer, 200);
  if (!strcmp(buffer, "volpack")) {
    renderer->type = LF_RENDERER_VOLPACK;
  } else if (!strcmp(buffer, "volvis")) {
    renderer->type = LF_RENDERER_VOLVIS;
  } else if (!strcmp(buffer, "renderman")) {
    renderer->type = LF_RENDERER_RENDERMAN;
  } else if (!strcmp(buffer, "dummy")) {
    renderer->type = LF_RENDERER_DUMMY;
  }

  /* Parse the rest of the segment */
  while (NextHdrToken(inFile, buffer, 200)) {  
    if (!strcmp(buffer, "endsegment")) return 1;

    if (!strcmp(buffer, "args")) {
      NextHdrToken(inFile, buffer, 200);
      n = atoi(buffer);
      if(n < 0 || n > LF_MAXPARAMS) {
	lfError("cannot handle %d args\n", n);
	return(0);
      }
      renderer->num_args = n;
      for(i = 0; i < n; i++) {
	NextHdrToken(inFile, buffer, 200);
	renderer->arg_indices[i] = atoi(buffer);
      }
    }

    if (!strcmp(buffer, "config")) {
      NextHdrToken(inFile, buffer, 200);
      renderer->config = strdup(buffer);
    }

    if (!strcmp(buffer, "quality")) {
      NextHdrToken(inFile, buffer, 200);
      renderer->quality = atoi(buffer);
    }
  }

  return 0;
}


/* Open a Lightslab file, malloc the LFFileHdr struct,
 * and fill it in with data from the LFField structure.
 */
FILE *lfOpenOutFile(char *fileName, LFFileHdr **hdr)
{
  FILE *outFile;

  /* Open the output file */
#ifdef sgi
  outFile = fopen(fileName, "w");
#else /* WIN32 */
  outFile = fopen(fileName, "w+r");
#endif  
  if (!outFile) {
    lfError("Error: cannot open file %s\n", fileName);
    return(NULL);
  }
  /* Malloc the slabHdr array */
  *hdr = (LFFileHdr *) calloc(1, sizeof(LFFileHdr));
  if (*hdr == NULL) {
    lfError("Error: No memory for slabHdr struct.\n");
    return(NULL);
  }
  /* reserve enough hdr space to hold the header */
  (*hdr)->hdrSize = LF_DEFAULTHDRLEN;
  /* Initialize the header (clear it, and write terminating character) */
  ClearHdr(*hdr, outFile);

  return outFile;
}


/* Write the header (if necessary) and close the file.
 * NOTE -- This frees the LFFileHdr Structure! 
 */
int lfCloseInFile(LFFileHdr **hdr, FILE *inFile)
{
  /* Free/ Clean-up the hdr structure */
  free(*hdr);
  *hdr = NULL;

  /* Close the output file */
  fclose(inFile);

  return 1;
}


int lfCloseOutFile(LFFileHdr **hdr, FILE *outFile)
{
  LFSlab *slab;
  LFVQCodebook *vq;
  LFSlabHdr *slabHdr;
  LFVQHdr *vqHdr;
  int slabno, vqno;
  int i;
  char *uvgeomStr[4] = {"geometry_uv", "           ", 
			"           ", "           "};
  char *stgeomStr[4] = {"geometry_st", "           ", 
			"           ", "           "};
  LFParams *params = __LF_GET_PARAMS();
  LFRenderer *renderer = __LF_GET_RENDERER();

  /* ---- Write the header ---- */
  rewind(outFile);
  fprintf(outFile, "LIF1.0\n");

  /* Print the datasize */
  fprintf(outFile, "datasize %d\n\n", (*hdr)->dataSize);

  /* print the lightfield */
  fprintf(outFile, "bgnlightfield 1\n");

  /* optionally write params */
  if(params->num) {
    fprintf(outFile, "\n  bgnsegment params %d\n", params->num);
    fprintf(outFile, "    params\n");
    for(i = 0; i < params->num; i++) {
      fprintf(outFile, "      %d %d  %d  \"%s\"\n",
	      params->params[i].min,
	      params->params[i].max,
	      params->params[i].def,
	      params->params[i].descr);
    }
    fprintf(outFile, "  endsegment\n");
  }

  /* optionally write renderer */
  if(renderer->type != LF_RENDERER_NONE) {
    fprintf(outFile, "\n  bgnsegment renderer %s\n",
	    renderer->type == LF_RENDERER_VOLPACK   ? "volpack" :
	    renderer->type == LF_RENDERER_VOLVIS    ? "volvis" :
	    renderer->type == LF_RENDERER_RENDERMAN ? "renderman" :
	    renderer->type == LF_RENDERER_DUMMY     ? "dummy" : "unknown");
    fprintf(outFile, "    config \"%s\"\n", renderer->config);
    fprintf(outFile, "    args %d  \n", renderer->num_args);
    for(i = 0; i < renderer->num_args; i++) {
      fprintf(outFile, "%d ", renderer->arg_indices[i]);
    }
    fprintf(outFile, "    quality %d  \n", renderer->quality);
    fprintf(outFile, "  endsegment\n");
  }

  /* Go through the slabs one by one */
  for (slabno=0; slabno < LF_MAXSLABS; slabno++) {
    slabHdr = &((*hdr)->slabHdr[slabno]);
    slab = slabHdr->slab;
    if (slab==NULL) continue;
    fprintf(outFile, "  bgnsegment slab %d\n", slabno); 
    if ((vq=slab->shared->vq) != NULL) {
      /* Output VQ index channel */
      fprintf(outFile, "    compression vq %d\n", slab->shared->vq->id);
      fprintf(outFile, "    format index\n");
      fprintf(outFile, "    bgnchannel index    # vq codebook indices\n");
      fprintf(outFile, "      type %s\n", 
	      (slabHdr->type == LF_INT8) ? "int8" :
	      (slabHdr->type == LF_INT16) ? "int16" :
	      /*(slabHdr->type==LF_INT32)?*/ "int32");
      fprintf(outFile, "      offset %d\n", slabHdr->offset);
      fprintf(outFile, "      size %d\n", slabHdr->size);
      fprintf(outFile, "    endchannel\n");
    } else { 
      fprintf(outFile, "    compression none\n");
      switch (slabHdr->format) {
      case LF_RGBCHANNEL:	
	/* red, green, and blue */
	fprintf(outFile, "    format rgb\n");
	fprintf(outFile, "    bgnchannel rgb\n");
	fprintf(outFile, "      type int8x3\n");
	fprintf(outFile, "      offset %d\n", slabHdr->offset);
	fprintf(outFile, "      size %d\n", slabHdr->size);
	fprintf(outFile, "    endchannel\n");
	break;
      case LF_RGBACHANNEL:	
	/* red, green, blue, and alpha */
	fprintf(outFile, "    format rgba\n");
	fprintf(outFile, "    bgnchannel rgba\n");
	fprintf(outFile, "      type int8x4\n");
	fprintf(outFile, "      offset %d\n", slabHdr->offset);
	fprintf(outFile, "      size %d\n", slabHdr->size);
	fprintf(outFile, "    endchannel\n");
	break;
      default:
        /* LF_INDEXCHANNEL, LF_CODEBOOKCHANNEL? */
	break;
      }
    }	
    /* Print out dimension in p, u, v, s, t */
    fprintf(outFile, "    samples_uv %d %d\n", slab->shared->nu,
	    slab->shared->nv);
    fprintf(outFile, "    samples_st %d %d\n", slab->shared->ns, 
	    slab->shared->nt);
    if(params->num) {
      fprintf(outFile, "    samples_p %d\n", params->num);
      lfWritePSamples(outFile, 4, 4, 0, params->num, slab->shared->psamples);
    }
    if(slab->shared->rectangular) {
      fprintf(outFile, "    rectangular\n");
    }
    if(slab->shared->samples) {
      fprintf(outFile, "    samples\n");
    }
      
    /* Print out plan, if available */
    if(slab->shared->plans) {
      int n, i;
      LFInterpolationPlan *plan = slab->shared->plans;

      for(n = 1, i = 0; i < params->num; i++) {
	n *= params->params[i].max - params->params[i].min + 1;
      }
      fprintf (outFile, "    interpolation %d %d\n", n, params->num);
      for(i = 0; i < n; i++) {
	int j;
	fprintf (outFile, "%d ", plan->num_accesses);
	for(j = 0; j < plan->num_accesses; j++) {
	  int e;
	  for(e = 0; e < params->num; e++) {
	    fprintf (outFile, "%d ", plan->p_indices[j][e]);
	  }
	}

	fprintf(outFile, "%d ", plan->num_interp);
	for(j = 0; j < plan->num_interp; j++) {
	  fprintf (outFile, "%d %d %e ", plan->source[j], plan->target[j], ((float)plan->fp[j]) / (1<<plan->fp_bits[j]));
	}
	fprintf (outFile, "\n");
	plan++;
      }
    }


    /* Print out uv slab coordinates */
    for (i=0; i < 4; i++) {
      fprintf (outFile, "    %s % f % f % f % f % f % f\n", uvgeomStr[i],
	       slab->uv[i].ox, slab->uv[i].oy, slab->uv[i].oz, slab->uv[i].ow,
	       slab->uv[i].s, slab->uv[i].t);
    }
    /* Print out st slab coordinates */
    for (i=0; i < 4; i++) {
      fprintf (outFile, "    %s % f % f % f % f % f % f\n", stgeomStr[i],
	       slab->st[i].ox, slab->st[i].oy, slab->st[i].oz, slab->st[i].ow,
	       slab->st[i].s, slab->st[i].t);
    }

    /* End of segment */
    fprintf(outFile, "  endsegment # %d\n", slabno); /* +1 to start at 1 */
  }

  /* Go through the vqs one by one */
  for (vqno=0; vqno < LF_MAXSLABS; vqno++) {
    int i;
 
    vqHdr = &((*hdr)->vqHdr[vqno]);
    vq = vqHdr->codebook;
    if (vq==NULL) continue;
    fprintf(outFile, "  bgnsegment vq %d\n", vqno); 
    switch (vqHdr->format) {
    case LF_RGBCHANNEL:	
      /* red, green, and blue */
      fprintf(outFile, "    format rgb\n");
      fprintf(outFile, "    bgnchannel rgb\n");
      fprintf(outFile, "      type int8x3\n");
      fprintf(outFile, "      offset %d\n", vqHdr->offset);
      fprintf(outFile, "      size %d\n", vqHdr->size);
      fprintf(outFile, "    endchannel\n");
      break;
    case LF_RGBACHANNEL:	
      /* red, green, blue, and alpha */
      fprintf(outFile, "    format rgba\n");
      fprintf(outFile, "    bgnchannel rgba\n");
      fprintf(outFile, "      type int8x4\n");
      fprintf(outFile, "      offset %d\n", vqHdr->offset);
      fprintf(outFile, "      size %d\n", vqHdr->size);
      fprintf(outFile, "    endchannel\n");
      break;
    default:
      /* LF_INDEXCHANNEL, LF_CODEBOOKCHANNEL? */
      break;
    }
    /* print tile info */
    fprintf(outFile, "    tiles %d\n", vq->size);
    fprintf(outFile, "    tilesize %d %d %d %d\n",
	    1 << vq->ubits, 1 << vq->vbits, 1 << vq->sbits, 1 << vq->tbits);
    if(params->num) {
      fprintf(outFile, "    ptilesize %d  ", params->num);
      for(i = 0; i < params->num; i++) {
	fprintf(outFile, "%d ", 1<<vq->pbits[i]);
      }
    }

    /* End of segment */
    fprintf(outFile, "  endsegment # %d\n", vqno); /* +1 to start at 1 */
  }

  /* End of lightfield entry */
  fprintf(outFile, "endlightfield # 1\n");
  fprintf(outFile, "endheader\n");

  /* Free/ Clean-up the hdr structure */
  free(*hdr);
  *hdr = NULL;

  /* Close the output file */
  fclose(outFile);
  return 1;
}


/*
** calculate (1+last[max]-first[max])* ... *(1+last[0]-first[0])
*/
static int diffMul(int pFirstLast[2][LF_MAXPARAMS])
{
  int i, res;

  for(i = LF_MAXPARAMS - 1, res = 1;
      i >= 0;
      i--) {
    res *= 1 + pFirstLast[1][i] - pFirstLast[0][i];
  }
  return res;
}


/* volume of slice (or rather hypercube) in p */
static int pTileSize(int pbits[LF_MAXPARAMS])
{
  int i, res;

  for(i = 0, res = 1; i < LF_MAXPARAMS; i++) {
    res *= 1<<pbits[i];
  }
  return res;
}

/*
** calculate informations to read or write tiles in in .lid file
*/
static int
getPosition(LFSlab *slab, LFFileHdr *hdr, LFSlabHdr *slabHdr,
	    int tilePixels, int indexBytes,
	    int uFirst, int uLast, int vFirst, int vLast,
	    int pFirstLast[2][LF_MAXPARAMS],
	    int *rawOffset, int *rawLen, int *inElemBytes, int *outElemBytes)
{
  int blockn, blockLen; /* used for VQ index arrays */
  LFParams *params = __LF_GET_PARAMS();
  int i;

  switch (slabHdr->format) {
  case LF_RGBCHANNEL:
  case LF_RGBACHANNEL:
    *rawOffset = hdr->hdrSize + slabHdr->offset +
      ((slabHdr->type == LF_INT8x3) ? 3 : 4) * 
      (lfIndex(params->num, pFirstLast[0], slab->shared->psamples) * slab->shared->nu * slab->shared->nv +
       uFirst * slab->shared->nv +
       vFirst) *
      slab->shared->ns * slab->shared->nt;
    *rawLen = ((slabHdr->type == LF_INT8x3) ? 3 : 4)
      * diffMul(pFirstLast)
      * (1+uLast-uFirst) * (1+vLast-vFirst)
      * slab->shared->ns * slab->shared->nt;
    *inElemBytes = *outElemBytes = 1;
    break;
  case LF_INDEXCHANNEL:
    /* Note: this requires that the slabslice in u and v is a multiple
     * of the tile size.  (So, for example, if you have 2x2x2x2 blocks,
     * then uFirst and vFirst must be even numbers, and uLast and vLast
     * must be odd numbers.)
     */
    
    /* Check to make sure that uFirst and vFirst are divisible by 
     * the tile size in p, u and v and that slice size (in p, u and v) is divisible
     * by the tile size.
     */
    for(i = 0; i < LF_MAXPARAMS; i++) {
      if(pFirstLast[0][i] % (1<<slab->shared->vq->pbits[i])) {
	lfError(
		"Error:  for reading/writing VQ indices, the slice must start\n"
		"        on an even boundary (Divisible by the tile size)\n"
		"        (violated by parameter #%d: %d / %d).\n",
		i, pFirstLast[0][i], (1<<slab->shared->vq->pbits[i]));
	return(0);
      }
      if((pFirstLast[1][i]+1) % (1<<slab->shared->vq->pbits[i])) {
	lfError(
		"Error:  for reading/writing VQ indices, the size of the slice\n"
		"        must be evenly divisible by the tile size\n"
		"        (violated by parameter #%d: %d / %d).\n",
		i, pFirstLast[1][i]+1, (1<<slab->shared->vq->pbits[i]));
	return(0);
      }
    }
    if ((uFirst % (1<<slab->shared->vq->ubits)) ||
	  (vFirst % (1<<slab->shared->vq->vbits))) {
      lfError(
	      "Error:  for reading/writing VQ indices, the slice must start\n"
	      "        on an even boundary (Divisible by the tile size)\n");
      return(0);
    }
    if (((uLast+1) % (1<<slab->shared->vq->ubits)) ||
	((vLast+1) % (1<<slab->shared->vq->vbits))) {
      lfError(
	      "Error:  for reading/writing VQ indices, the size of the slice\n"
	      "        must be evenly divisible by the tile size.\n");
      return(0);
    }

    /* Calculate the block offset */
    blockn = ((lfVQIndex(params->num, slab->shared->vq->pbits, pFirstLast[0], slab->shared->psamples) * (slab->shared->nu >> slab->shared->vq->ubits)) +
	      (uFirst >> slab->shared->vq->ubits)) * (slab->shared->nv >> slab->shared->vq->vbits) +
      (vFirst >> slab->shared->vq->vbits);
    /* The length of a single tile slice in uv */
    blockLen = (pTileSize(slab->shared->vq->pbits) *
		(1<<slab->shared->vq->ubits) *
		(1<<slab->shared->vq->vbits)) *
      slab->shared->ns * slab->shared->nt / tilePixels;
    /* offset */
    *rawOffset = hdr->hdrSize + slabHdr->offset + 
      indexBytes * blockn * blockLen;
    
    /* Calculate the length of the block */
    *rawLen = diffMul(pFirstLast) * (1+uLast-uFirst) * (1+vLast-vFirst) * 
      slab->shared->ns * slab->shared->nt / tilePixels;
    
    *inElemBytes = *outElemBytes = indexBytes;
    break;
  default:
    lfError("Error: unrecognized data format %d\n", 
	    slabHdr->format);
    *rawOffset =
      *rawLen =
      *inElemBytes =
      *outElemBytes = 0;
    return(0);
    break;
  }
#ifdef DEBUG
  lfOutput("getPosition: 0x%x => %d @%d in/out %d/%d [",
	   slabHdr->format, *rawLen, *rawOffset, *inElemBytes, *outElemBytes);
  for(i = LF_MAXPARAMS - 1; i >= 0; i--)
    lfOutput("%d-%d, ", pFirstLast[0][i], pFirstLast[1][i]);
  lfOutput("%d-%d, %d-%d]\n", uFirst, uLast, vFirst, vLast);
#endif

  return(1);
}


/* Write a portion (possibly all) of a slab of a lightfield to file */
int lfWriteBlock(void *block, int slabn, LFSlab *slab, LFFileHdr *hdr, 
		 FILE *file, int uFirst, int uLast, int vFirst, int vLast,
		 int pFirstLast[2][LF_MAXPARAMS],
 		 LF_ChannelFormat format, LF_ChannelType type)
{
  LFShared *shared = slab->shared;
  LFSlabHdr *slabHdr = &(hdr->slabHdr[slabn]);
  int indexBytes = 0;  /* Number of bytes necessary to index codebook */
  int tilePixels = 0;  /* Number of pixels contained in each tile */
  LFVQCodebook *vq;		/* vq codebook */
  int channelSize;
  int rawOffset, rawLen, inElemBytes, outElemBytes;

  if ((vq = shared->vq) != NULL) {
    /* indexBytes = # of bytes needed to index into the codebook array */
    indexBytes = (vq->size > 256) ? (vq->size > 65536) ? 4 : 2 : 1;
    /* tilePixels is the number of pixels that are in each tile */
    tilePixels = (1 << (lfSumPBits(vq) + vq->ubits + vq->vbits + vq->sbits + vq->tbits));
  }

  if (format == LF_CODEBOOKCHANNEL) {
    /* Codebook*/
    lfError("Error: lfWriteBlock called on a codebook!\n");
    return 0;
  } else {
    /* Handle slab segment (not codebook) */
    /* Set hdr info, if this is the first write */
    if (slabHdr->size == 0) {
      slabHdr->format = format;
      slabHdr->type   = type;
      slabHdr->slab   = slab;
      slabHdr->offset = hdr->dataSize;
      if (vq != NULL) {
	/* Channelsize */
	channelSize =  MAX_PINDEX(slab->shared) * slab->shared->nu * slab->shared->nv * 
	  slab->shared->ns * slab->shared->nt * indexBytes / tilePixels;
	slabHdr->size = channelSize;
      } else {
	/* Set size big enough to hold raw file */
	slabHdr->size = 
	  ((type == LF_INT8x4) ? 4 : /*(type == LF_INT8x3)?*/  3) *
	  (MAX_PINDEX(slab->shared) * shared->nu * shared->nv * shared->ns * shared->nt);
      }
      lfOutput("upping datasize by %d\n", slabHdr->size);
      hdr->dataSize += slabHdr->size;
    }

    /* Now write the binary section to file */
    if(getPosition(slab, hdr, slabHdr, tilePixels, indexBytes,
		   uFirst, uLast, vFirst, vLast, pFirstLast,
		   &rawOffset, &rawLen, &inElemBytes, &outElemBytes)) {
#ifndef ABGR_LF
      if(!vq) {
	swapPixels(block, rawLen, type == LF_INT8x4 ? 4 : 3);
      }
#endif
      WriteRawBlock(file, rawOffset, block,
		    rawLen, inElemBytes, outElemBytes);
#ifndef ABGR_LF
      if(!vq) {
	swapPixels(block, rawLen, type == LF_INT8x4 ? 4 : 3);
      }
#endif
    }
  }	
  return 1;
}

/* Read an entire codebook from the file */
/* (This will allocate the memory needed for the codebook array) */
int lfReadCodebook(LFVQCodebook *vq, LFShared *shared, LFFileHdr *hdr, 
		   FILE *file)
{
  LFVQHdr *vqHdr = &(hdr->vqHdr[vq->id]);
  int indexBytes = 0;  /* Number of bytes necessary to index codebook */
  int tilePixels = 0;  /* Number of pixels contained in each tile */
  int rawOffset, rawLen, inElemBytes, outElemBytes;
  void *block;

  lfOutput("lfReadCodebook %d\n", vq->id);

  /* indexBytes = # of bytes needed to index into the codebook array */
  indexBytes = (vq->size > 256) ? (vq->size > 65536) ? 4 : 2 : 1;
  /* tilePixels is the number of pixels that are in each tile */
  tilePixels = (1 << (lfSumPBits(vq) + vq->ubits + vq->vbits + vq->sbits + vq->tbits));
  /* Allocate space for the codebook's array */
  if (vq->codebook != NULL) {
    lfError("Warning: lfReadCodebook:"
	    "codebook already contains data!\n");
  }
  if ((vq->codebook = (void *) 
       malloc(vq->size * tilePixels * vq->sample_size)) == NULL) {
    lfError("Error:  Not enough memory to allocate codebook %d!\n",
	    vq->id);
    return 0;
  }

  block = vq->codebook;
  rawOffset = hdr->hdrSize + vqHdr->offset;
  /* vqHdr->size is already in bytes */
  rawLen = vqHdr->size;
  inElemBytes = 1;
  outElemBytes = 1;
  ReadRawBlock(file, rawOffset, block, rawLen, inElemBytes, outElemBytes);
#ifndef ABGR_LF
  swapPixels(block, rawLen, vq->sample_size);
#endif
  lfOutput("Done reading %d bytes of codebook.\n", rawLen);
  return 1;
}

/* Write an entire codebook to file */
int lfWriteCodebook(void *block, int vqn, LFSlab *slab, LFFileHdr *hdr, 
		 FILE *file, LF_ChannelFormat format, LF_ChannelType type)
{
  LFShared *shared = slab->shared;
  LFVQHdr *vqHdr = &(hdr->vqHdr[vqn]);
  int indexBytes = 0;  /* Number of bytes necessary to index codebook */
  int tilePixels = 0;  /* Number of pixels contained in each tile */
  LFVQCodebook *vq;		/* vq codebook */
  int rawOffset, rawLen, inElemBytes, outElemBytes;

  lfOutput("lfWriteCodebook, format %d\n", format);

  if ((vq = shared->vq) != NULL) {
    /* indexBytes = # of bytes needed to index into the codebook array */
    indexBytes = (vq->size > 256) ? (vq->size > 65536) ? 4 : 2 : 1;
    /* tilePixels is the number of pixels that are in each tile */
    tilePixels = (1 << (lfSumPBits(vq) + vq->ubits + vq->vbits + vq->sbits + vq->tbits));
  }

  if (format == LF_CODEBOOKCHANNEL) {
    /* Handle codebook */
    vq = slab->shared->vq;

    /* Don't re-set the hdr info (which would allocate *another* block
     * for the codebook), if the codebook is already used by another
     * slab.
     */
    if (vqHdr->size) {
      /* debug message */
      lfOutput("skipping codebook %d -- it's shared by another slab.\n",
	     slab->shared->vq->id);
      return 1;
    }

    /* Set hdr info, since we assume that we write the entire codebook
       * out at one time.
       */
    vqHdr->format = (type == LF_INT8x3) ? LF_RGBCHANNEL : LF_RGBACHANNEL;
    vqHdr->type = type;
    vqHdr->codebook = vq;
    vqHdr->offset = hdr->dataSize;
    vqHdr->size = vq->size * 
      (vq->sample_size << (lfSumPBits(vq) + vq->ubits + vq->vbits + vq->sbits + vq->tbits));
    lfOutput("upping datasize by VQ-codebook: %d\n", vqHdr->size);
    hdr->dataSize += vqHdr->size;
    rawOffset = hdr->hdrSize + vqHdr->offset;
    block = vq->codebook;
    rawLen = vqHdr->size;
    inElemBytes = 1;
    outElemBytes = 1;
#ifndef ABGR_LF
    swapPixels(block, rawLen, type == LF_INT8x4 ? 4 : 3);
#endif
    WriteRawBlock(file, rawOffset, block,
		  rawLen, inElemBytes, outElemBytes);
#ifndef ABGR_LF
    swapPixels(block, rawLen, type == LF_INT8x4 ? 4 : 3);
#endif
  } else {
    lfError("Error: lfWriteCodebook called on a non-codebook.\n");
    return 0;
  }
  return 1;
}


/* **********************************************************************
 *
 *	       Unwritten Functions Below
 *
 * **********************************************************************
 */


/* Read a portion (possibly all) of a slab of a lightfield from file */
int lfReadBlock(void *block, int slabn, LFSlab *slab, LFFileHdr *hdr, 
		FILE *file, int uFirst, int uLast, int vFirst, int vLast,
		int pFirstLast[2][LF_MAXPARAMS],
		LF_ChannelFormat format, LF_ChannelType type)
{
  LFShared *shared = slab->shared;
  LFSlabHdr *slabHdr = &(hdr->slabHdr[slabn]);
  int indexBytes = 0;  /* Number of bytes necessary to index codebook */
  int tilePixels = 0;  /* Number of pixels contained in each tile */
  LFVQCodebook *vq;		/* vq codebook */
  int rawOffset, rawLen, inElemBytes, outElemBytes;

  if ((vq = shared->vq) != NULL) {
    /* indexBytes = # of bytes needed to index into the codebook array */
    indexBytes = (vq->size > 256) ? (vq->size > 65536) ? 4 : 2 : 1;
    /* tilePixels is the number of pixels that are in each tile */
    tilePixels = (1 << (lfSumPBits(vq) + vq->ubits + vq->vbits + vq->sbits + vq->tbits));
  }

  if (format == LF_CODEBOOKCHANNEL) {
    /* Codebook*/
    lfError("Error: lfReadBlock called on a codebook!\n");
    return 0;
  } else {
    /* Now read the binary section from file */
    getPosition(slab, hdr, slabHdr, tilePixels, indexBytes,
		uFirst, uLast, vFirst, vLast, pFirstLast,
		&rawOffset, &rawLen, &inElemBytes, &outElemBytes);
    ReadRawBlock(file, rawOffset, block, rawLen, inElemBytes, outElemBytes);
#ifndef ABGR_LF
    if(!vq) {
      swapPixels(block, rawLen, type == LF_INT8x4 ? 4 : 3);
    }
#endif
  }	
  return 1;
}


/* **********************************************************************
 *
 *	       local helper functions
 *
 * **********************************************************************
 */


void *SlabSlice(LFSlab *slab, int u, int v)
{
  LFShared *shared = slab->shared;
  int *uindex = shared->uindex;
  int *vindex = shared->vindex;
  int *sindex = shared->sindex;
  int *tindex = shared->tindex;

  /* We always want to call LIGHTPTR, not VQLIGHTPTR, because
   * we want a pointer into the actual lightfield array, not into
   * the codebook.
   */
  if (shared->vq != NULL) {
    /* Figure out how the codebook indices are packed into the lightfield:
     *    codebook size less than 257      -> 1-byte indices
     *    codebook size less than 65537    -> 2-byte indices
     *    codebook size greater than 65536 -> 4-byte indices
     */
    if (slab->shared->vq->size <= 256) {
      unsigned char *lightfield = slab->lightfield;
      return LIGHTPTR(lightfield, u, v, 0, 0);
    } else if (slab->shared->vq->size <= 65536) {
      unsigned short *lightfield = slab->lightfield;
      return LIGHTPTR(lightfield, u, v, 0, 0);
    } else {
      unsigned int *lightfield = slab->lightfield;
      return LIGHTPTR(lightfield, u, v, 0, 0);
    }
  } else {
    /* Figure out how the data is packed, and get a pointer to it */
    switch (slab->shared->sample_size) {
    case 1: {
      unsigned char *lightfield = slab->lightfield;
      return LIGHTPTR(lightfield, u, v, 0, 0);
      break;
    }
    case 2: {
      unsigned short *lightfield = slab->lightfield;
      return LIGHTPTR(lightfield, u, v, 0, 0);
      break;
    }
    case 4: {
      LFPixelRGBA8 *lightfield = slab->lightfield;
      return LIGHTPTR(lightfield, u, v, 0, 0);
      break;
    }
    case 8: {
      unsigned long *lightfield = slab->lightfield;
      return LIGHTPTR(lightfield, u, v, 0, 0);
      break;
    }
    default: {
      lfError("Error:  slab packed in memory, %d bytes per pixel.\n"
	      "How the heck can I typecast that array???\n",
	      slab->shared->sample_size);
      return 0;
      break;
    }
    }
  }
}

void ClearHdr(LFFileHdr *hdr, FILE *file)
{
  int i;
  
  /* Go to start of file */
  rewind(file);

  /* Clear hdr space */
  for (i=0; i < (hdr->hdrSize-1); i++) {
    fputc(LF_HDRFILLER, file);
  }
  /* Terminating character */
  fputc(LF_HDRTERMINATOR, file);

  /* Now go back to start for writing hdr over it. */
  rewind(file);


}

int GetHdrLen(FILE *file)
{
  int i=0;
  int c;

  rewind(file);
  /* Count up characters in header until we hit terminator character */
  while ((c=fgetc(file)) != EOF) {
    i++;
    if (c == LF_HDRTERMINATOR) {
      break;
    }
  }
  rewind(file);
  return i;
}


/* Write out raw block of data.  It always writes little-endian,
 * for cross-platform compatibility.
 */
void WriteRawBlock(FILE *outFile, int offset, char *buffer, int elemCount,
		  int elemInSize, int elemOutSize)
{
  int i, j;
  unsigned long data;

  /*  lfOutput("WriteRawBlock off %d, size %d (%d,%d) into buf %d\n", 
	 offset, elemCount, elemInSize, elemOutSize, (int) buffer);
	 */

  /* Sanity check to avoid bus errors */
  if (elemInSize != 1 && elemInSize != 2 && elemInSize != 4 &&
      elemInSize != 8) {
    lfError(
	    "Error: WriteRawBlock:  What kind of number uses %d bytes??\n", 
	    elemInSize);
    return;
  }

  /* Go to starting offset in the file */
  fseek(outFile, offset, SEEK_SET);

  /* Now write */
  for (i=0; i<elemCount; i++) {
    /* Set data = element value */
    switch (elemInSize) {
    case 1:
      data = * (buffer);
      break;
    case 2:
      data = * ((unsigned short *) ((void *) buffer));
      break;
    case 4:
      data = * ((unsigned int *)   ((void *) buffer));
      break;
    case 8:
      data = * ((unsigned long *)  ((void *) buffer));
      break;
    default:
      /* should not happen */
      data = 0;
      break;
    }
    /*    lfOutput("w %d\n", data); */
    buffer += elemInSize;

    /* Write the data, little-endian */
    for (j=0; j < elemOutSize; j++) {
      fputc(data & 0x0ff, outFile);
      data >>= 8;
    }
  }
}
  
/* Read in raw block of data.  It reads writes little-endian,
 * for cross-platform compatibility.
 */
void ReadRawBlock(FILE *inFile, int offset, char *buffer, int elemCount,
		  int elemInSize, int elemOutSize)
{
  int i, j;
  unsigned long data;
  unsigned int tmpdata;

  /* Sanity check to avoid bus errors */
  if (elemOutSize != 1 && elemOutSize != 2 && elemOutSize != 4 &&
      elemOutSize != 8) {
    lfError(
	    "Error: ReadRawBlock:  What kind of number uses %d bytes??\n", 
	    elemOutSize);
    return;
  }

  /* Go to starting offset in the file */
  fseek(inFile, offset, SEEK_SET);

  /* Now read */
  for (i=0; i<elemCount; i++) {
    /* Read the data, little-endian */
    data = tmpdata = 0;
    for (j=0; j < elemInSize; j++) {
      tmpdata = fgetc(inFile);
      /* put the character in the jth byte */
      data += tmpdata << (j<<3);
    }

    /* Set data = element value */
    switch (elemOutSize) {
    case 1:
      * buffer = data;
      break;
    case 2:
      * ((short *) ((void *) buffer)) = data;
      break;
    case 4:
      * ((int *)   ((void *) buffer)) = data;
      break;
    case 8:
      * ((long *)  ((void *) buffer)) = data;
      break;
    }
    buffer += elemOutSize;
  }
}
  
/* This routine returns 1 on success, 0 on failure (end-of-hdr).
 * It assumes that buffer has been allocated and is large enough
 * to hold a token of maxTokenSize-1 + null-terminator.
 */
int NextHdrToken(FILE *file, char *buffer, int maxTokenSize)
{
  int index=0;		/* Index into the buffer */
  int c;
  /* State machine:
   * 0 = haven't seen start of token yet
   * 1 = reading token
   * 2 = token finished
   * 3 = inside a comment
   */
  int state = 0;
  int quotes = 0;    /* state 1: token is terminated by quotes, not space */

  while (state != 2) {
    c = fgetc(file);

    /* Handle EOF */
    if (c == EOF || c == LF_HDRTERMINATOR) {
      buffer[index] = '\0';
      /* Return 1 iff we've seen part of a token, 0 otherwise */
      if (state == 1) {
	return 1;
      } else {
	return 0;
      }
    }
    
    /* Handle comments */
    if (c == '#') {
      if (state == 1) {
	buffer[index] = '\0';
	/* Finish reading the comment, so we start on next line next time */
	while ((c != '\n') && (c != EOF)) {
	  c = fgetc(file);
	}
	/* now return token that we read. */
	return 1;
      } else {
	/* We're inside a comment now */
	state=3;
	continue;
      }
    }

    /* Handle end-of-comment? */
    if (c == '\n') {
      if (state == 1) {
	buffer[index] = '\0';
	return 1;
      } else {
	/* End of comment, back into whitespace */
	state=0;
	continue;
      }
    }

    /* Handle whitespace */
    if (isspace(c)) {
      if (state == 1) {
	if(quotes) {
	  if(index < maxTokenSize-1) {
	    buffer[index++] = c;
	  }
	} else {
	  buffer[index] = '\0';
	  return 1;
	}
      } else {
	/* do nothing -- go on to next character */
	continue;
      }
    }
	
    /* Default case */
    {
      if (state == 0) {
	state = 1;
	if(c == '\"') {
	  quotes = 1;
	} else {
	  buffer[index++] = c;
	  quotes = 0;
	}
      } else if (state == 1) {
	if(quotes && c == '\"') {
	  buffer[index] = '\0';
	  return 1;
	} else {
	  if(index < maxTokenSize-1) {
	    buffer[index++] = c;
	  }
	}
      }
    }
  }
  return 1;
}

/* Read a token from input file, check to make sure it matches what
 * we were expecting.
 */
int AssertToken(FILE *file, char *token)
{
  char buf[200];
  /* Read the next token */
  if (!NextHdrToken(file, buf, 200)) {
    lfError("Error: expected token \"%s\", got EOF.\n", token);
    return 0;
  }
  /* Check to make sure it matches what we expect */
  if (strcmp(token, buf)) {
    lfError("Error: expected token \"%s\", got \"%s\".\n", 
	    token, buf);
    return 0;
  }
  return 1;
}

/* This function takes the log (base2) of n, assuming n is a power of 2 */
int Log2(unsigned long n)
{
  int log;

  for (log=0; log < 32; log++) {
    if (1 & (n >> log)) return log;
  }
  /* if we get here, n was 0 */
  return(-1);
}

/*
** RGB(A) pixels in a .lif file are always in (A)BGR order, so swap if necessary.
*/
static void swapPixels(char *buffer, int byte_count, int elem_size)
{
  int elem_count = byte_count / elem_size;

  if(elem_size == 3) {
    while(--elem_count >= 0) {
      char tmp = buffer[0];
      buffer[0] = buffer[2];
      buffer[2] = tmp;
      buffer += 3;
    }
  } else {
    while(--elem_count >= 0) {
      char tmp = buffer[0];
      buffer[0] = buffer[3];
      buffer[3] = tmp;
      tmp = buffer[1];
      buffer[1] = buffer[2];
      buffer[2] = tmp;
      buffer += 4;
    }
  }
}
