/*
 * io.c
 *
 * stuff related to file I/O
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <unistd.h>
#ifndef NO_INTERNAL_READ_RGB
#include "readrgb.h"
#else
#include <libgen.h>
#include <gl/image.h>
#endif

#include "lightfield.h"
#include "lerp.h"
#include "renderer.h"

LFSlab *__lf_descr_slabs;
LFParams   *__lf_descr_params;
LFRenderer *__lf_descr_renderer;

typedef struct {
  char *slab_names[LF_MAXSLABS];
  char *slab_suffices[LF_MAXSLABS];
  float uvmat[LF_MAXSLABS][3][3];   /* mat * (u,v,1)T = (x,y,z)T */
  float stmat[LF_MAXSLABS][3][3];
  bool_t perspective;
  bool_t create_images;
  bool_t no_images;
  bool_t uv_filter;
  int nslab;
} LFSliceHeader;

typedef struct {
  int u_size, v_size;
  int s_size, t_size;
  int sample_size;
} SlabDefaults;

static void
get_xyz(float mat[3][3], float s, float t, vpVector3 v)
{
  int i;

  for(i = 0; i < 3; i++) {
    v[i] = mat[i][0]*s + mat[i][1]*t + mat[i][2];
  }
}

static void
set_matrix(float a[3][3], LFVtx st[4])
{
  float x[4], y[4], z[4], s[4], t[4];
  int i,j,k;

  /* copy and normalize if possible */
  for(i = 0; i < 4; i++) {
    x[i] = st[i].ox;
    y[i] = st[i].oy;
    z[i] = st[i].oz;
    s[i] = st[i].s;
    t[i] = st[i].t;
    if(st[i].w != 0.0) {
      x[i] /= st[i].w;
      y[i] /= st[i].w;
      z[i] /= st[i].w;
    }
  }
  /* check for uncorrelated s/t values */
  for(i = 0; i < 4; i++) {
    for(j = i + 1; j < 4; j++) {
      for(k = j + 1; k < 4; k++) {
	float diff = (-s[k]*t[j]+s[k]*t[i]+t[j]*s[i]+t[k]*s[j]-t[k]*s[i]-t[j]*s[j]);

	if(diff <= 0.0000001)
	  continue;

	/* taken from maple... */
	a[0][1] = -(s[k]*x[j]-s[k]*x[i]-s[i]*x[j]-x[k]*s[j]+x[k]*s[i]+x[i]*s[j])/diff;
	a[1][2] = (-t[j]*s[k]*y[i]+y[i]*t[k]*s[j]-y[j]*t[k]*s[i]+t[j]*y[k]*s[i]+y[j]*s[k]*t[i]-t[i]*y[k]*s[j])/diff;
	a[2][1] = -(s[k]*z[j]-s[k]*z[i]-z[j]*s[i]-z[k]*s[j]+z[k]*s[i]+s[j]*z[i])/diff;
	a[1][1] = (-s[k]*y[j]+s[k]*y[i]+y[j]*s[i]+y[k]*s[j]-y[k]*s[i]-s[j]*y[i])/diff;
	a[2][0] = (z[j]*t[k]-t[j]*z[k]+z[i]*t[j]-z[i]*t[k]-t[i]*z[j]+t[i]*z[k])/diff;
	a[1][0] = -(-y[j]*t[k]+t[j]*y[k]-y[i]*t[j]+y[i]*t[k]+t[i]*y[j]-t[i]*y[k])/diff;
	a[2][2] = (-t[j]*s[k]*z[i]+z[i]*t[k]*s[j]-z[j]*t[k]*s[i]+t[j]*z[k]*s[i]+z[j]*s[k]*t[i]-t[i]*z[k]*s[j])/diff;
	a[0][0] = (x[j]*t[k]-t[j]*x[k]+t[j]*x[i]-x[i]*t[k]-x[j]*t[i]+t[i]*x[k])/diff;
	a[0][2] = (-t[j]*s[k]*x[i]+x[i]*t[k]*s[j]-x[j]*t[k]*s[i]+t[j]*x[k]*s[i]+x[j]*s[k]*t[i]-t[i]*x[k]*s[j])/diff;
	return;
      }
    }
  }
  lfError("set_matrix: no solution found for plane equation, check slab planes!\n");
  exit(1);
}

static void
fillSlabDefault(LFInternOp *op, SlabDefaults *defaults)
{
    LFShared *shared;
    int i;

    for ( i=0 ; i<LF_MAXSLABS ; i++ )  {
	shared = __lf_descr_slabs[i].shared;
	shared->sample_size = defaults->sample_size;
	shared->nu = defaults->u_size;
	shared->nv = defaults->v_size;
	shared->ns = defaults->s_size;
	shared->nt = defaults->t_size;
    }
}

/*
 * function to begin inputing all slice files in a lightfield
 */
extern FILE *yyin;
void *
lfBeginReadSliceFile(LFOps *ops, LFInternOp *op)
{
  char *index_fnm = op->op_extra;
  FILE *fp;
  LFSliceHeader *hdr = calloc(1, sizeof(LFSliceHeader));
  LFSlab *slab;
  char fnm[256];
  char name[128], *dir;
  int i, n, id;
  float *descr = op->op_descr;
  int no_light = 0;
  SlabDefaults defaults;
  int quality = -1;
  char *old_dir;
  int cwdsize = 256;

  defaults.u_size = defaults.v_size = 16;
  defaults.s_size = defaults.t_size = 256;
  defaults.sample_size = 3;
  hdr->create_images = 0;

  while(*descr != LF_NULL) {
    switch((int)*descr++) {
    case LF_READ_SAMPLES_UV:
      defaults.u_size = *descr++;
      defaults.v_size = *descr++;
      break;
      
    case LF_READ_SAMPLES_ST:
      defaults.s_size = *descr++;
      defaults.t_size = *descr++;
      break;
      
    case LF_READ_FORMAT:
      switch((int)*descr++) {
      case LF_RGB:
	defaults.sample_size = 3;
	break;
	
      case LF_RGBA:
	defaults.sample_size = 4;
	break;
      }
      break;
    case LF_READ_NO_LIGHT:
      no_light = 1;
      break;
    case LF_READ_NO_IMAGES:
      hdr->no_images = 1;
      break;
    case LF_READ_UVFILTER:
      hdr->uv_filter = TRUE;
      break;
    case LF_READ_CREATE_IMAGES:
      hdr->create_images = 1;
      break;
    case LF_READ_QUALITY:
      quality = *descr++;
      break;
    }
  }

  __lf_descr_renderer = &__LF_GET_CONTEXT()->field->renderer;
  __lf_descr_params   = &__LF_GET_CONTEXT()->field->params;
  bzero(__lf_descr_renderer, sizeof(*__lf_descr_renderer));
  bzero(__lf_descr_params, sizeof(*__lf_descr_params));
  __lf_descr_slabs = calloc(LF_MAXSLABS, sizeof(LFSlab));
  for ( i=0 ; i<LF_MAXSLABS ; i++ )
    __lf_descr_slabs[i].shared = calloc(1, sizeof(LFShared));
  
  /* fill in the defaults */
  fillSlabDefault(op, &defaults);
  
  /* determine available slabs */
  if  ((fp=fopen(index_fnm, "r")) == NULL) {
    lfError("lfBeginReadSliceFile: cannot read %s\n", index_fnm);
    exit(1);
  }

  /* change directory */
  while(1) {
    if((old_dir = getcwd(NULL, cwdsize))) {
      break;
    }
    if(errno != ERANGE) {
      perror("getcwd");
      exit(10);
    }
    cwdsize += 256;
  }
  strcpy(name, index_fnm);
  dir = dirname(name);
  if(chdir(dir)) {
    perror(dir);
  }

#ifdef YYRESTART
  yyrestart(fp);
#else
  yyin = fp;
#endif
  yyparse();
  fclose(fp);
  chdir(old_dir);
  free(old_dir);
  if(quality != -1) {
    __lf_descr_renderer->quality = quality;
  }

  /* create a file header to access slab slices */
  for ( i=0 ; i<LF_MAXSLABS ; i++ ) {
    char *suffix;

    slab = &__lf_descr_slabs[i];
    if  (! slab->lightfield) 
      continue;
    if (((char *)slab->lightfield)[0] == '/' ||
	!dir[0]) {
      sprintf(fnm, "%s", (char *)slab->lightfield);
    } else {
      sprintf(fnm, "%s/%s", dir, (char *)slab->lightfield);
    }
    suffix = strrchr(fnm, '.');
    if(suffix) {
      hdr->slab_suffices[i] = strdup(suffix);
      *suffix = 0;
    } else {
      hdr->slab_suffices[i] = strdup(".rgb");
    }
    hdr->slab_names[i] = strdup(fnm);
    free(slab->lightfield);
    slab->lightfield = 0;

    set_matrix(hdr->uvmat[i], slab->uv);
    set_matrix(hdr->stmat[i], slab->st);
    if(slab->uv[0].w == 0.0 ||
       slab->st[0].w == 0.0) {
      hdr->perspective = FALSE;
    } else {
      hdr->perspective = TRUE;
    }

#ifdef DETAILED_PLANE_LOG
    {
      vpVector3 xyz;
      int j;

      lfOutput("uv plane:\n");
      for(j = 0; j < 4; j++) {
	get_xyz(hdr->uvmat[i], slab->uv[j].s, slab->uv[j].t, xyz);
	lfOutput("   %d. %f/%f -> %f/%f/%f\n",
		 j, slab->uv[i].s, slab->uv[i].t,
		 xyz[0], xyz[1], xyz[2]);
      }
      lfOutput("st plane:\n");
      for(j = 0; j < 4; j++) {
	get_xyz(hdr->stmat[i], slab->st[j].s, slab->st[j].t, xyz);
	lfOutput("   %d. %f/%f -> %f/%f/%f\n",
		 j, slab->st[j].s, slab->st[j].t,
		 xyz[0], xyz[1], xyz[2]);
      }
    }
#endif
  }

  /* determine what slabs are operable */
  if  (ops->chain_msk & LF_SET_SLAB)  {
    for ( i=n=0 ; i<ops->op_slab_cnt ; i++ ) {
      id = ops->op_slabs[i]->id;
      if  (! hdr->slab_names[id])
	continue;
      
      if  (ops->op_slabs[i]->shared)  {
	if  (ops->op_slabs[i]->shared->vq)
	  free(ops->op_slabs[i]->shared->vq);
	free(ops->op_slabs[i]->shared);
      }
      *ops->op_slabs[i] = __lf_descr_slabs[id];
      ops->op_slabs[i]->id = id;
      ops->op_slabs[n++] = ops->op_slabs[i];
    }
    ops->op_slab_cnt = n;
  }
  free(__lf_descr_slabs);
  
  /* fill in functions */
  op->op_func  = lfReadSliceFile;
  op->op_end   = lfEndReadSliceFile;
  op->op_update = ops->chain_msk & (LF_GEN_VQ_TRAINSET|LF_GEN_VQ_CODEARRAY) ?
    lfSerialUpdateBlockSlice : lfUpdateNoop;
  
  return hdr;
}

/*
 * render correct image for floating point (uv) position
 */
static GLubyte *
RenderUV(float u, float v, LFSlab *slab, const int *pos, LFSliceHeader *hdr)
{
  LFView *view = &__LF_GET_CONTEXT()->view;
  LFRenderer *renderer = __LF_GET_RENDERER();
  LFShared *shared = slab->shared;
  LFParams *params = __LF_GET_PARAMS();
  GLubyte *image;
  int slab_id = slab->id; 
  vpVector3 lowerleftuv, /* lower left corner of uv/st planes */
    lowerleftst;
  vpVector3 uvec,        /* unit vectors along edge of uv/st planes */
    vvec,
    svec,
    tvec;
  vpVector3 obs, center; /* current position in uv plane = observer,
			    projection onto st plane = center */
  vpVector3 wincenter;   /* center relativ to lower left corner */
  float win_width,       /* length of svec and tvec from origin of */
    win_height;          /* plane to ns/nt */
  vpVector3 stnormal;    /* unit normal vector of st plane */
  vpVector3 lloffsetst,  /* projection of lowerleftst onto st plane, */
    llnormal;    /* lowerleftst = lloffsetst + llnormal */
  vpVector3 tmp;
  float tmp2;
  int values[LF_MAXPARAMS];
  int i;

  if(!(image = malloc(shared->sample_size * shared->ns * shared->nt))) {
    lfError("lfReadSliceFile: no memory for renderer.\n");
    exit(1);
  }

  /* setup view for renderer (using volpack tools) */
  get_xyz(hdr->uvmat[slab_id], 0, 0, lowerleftuv);
  get_xyz(hdr->stmat[slab_id], 0, 0, lowerleftst);
  get_xyz(hdr->uvmat[slab_id], 1, 0, tmp);
  vpDiffVector3(uvec, tmp, lowerleftuv);
  vpNormalize3(uvec);
  get_xyz(hdr->uvmat[slab_id], 0, 1, tmp);
  vpDiffVector3(vvec, tmp, lowerleftuv);
  vpNormalize3(vvec);
  get_xyz(hdr->stmat[slab_id], 1, 0, tmp);
  vpDiffVector3(svec, tmp, lowerleftst);
  win_width = sqrt(vpDotProduct3(svec, svec));
  vpNormalize3(svec);
  get_xyz(hdr->stmat[slab_id], 0, 1, tmp);
  vpDiffVector3(tvec, tmp, lowerleftst);
  win_height = sqrt(vpDotProduct3(tvec, tvec));
  vpNormalize3(tvec);
  vpCrossProduct(stnormal, svec, tvec);
  vpNormalize3(stnormal);
  tmp2 = vpDotProduct3(lowerleftst, stnormal);
  vpSetVector3(llnormal, tmp2 * stnormal[0],
	       tmp2 * stnormal[1],
	       tmp2 * stnormal[2]);
  vpDiffVector3(lloffsetst, lowerleftst, llnormal);
  get_xyz(hdr->uvmat[slab_id], u, v, obs);
  tmp2 = vpDotProduct3(obs, stnormal);
  vpSetVector3(center, obs[0] - tmp2 * stnormal[0],
	       obs[1] - tmp2 * stnormal[1],
	       obs[2] - tmp2 * stnormal[2]);
  vpDiffVector3(wincenter, center, lloffsetst);
  view->screen_left  = -vpDotProduct3(wincenter, svec);
  view->screen_lower = -vpDotProduct3(wincenter, tvec);
  view->screen_right = win_width + view->screen_left;
  view->screen_upper = win_height + view->screen_lower;
  vpDiffVector3(tmp, obs, center);
  view->screen_d = sqrt(vpDotProduct3(tmp, tmp)) -
    sqrt(vpDotProduct3(llnormal, llnormal));
  view->fov = 0.0;
  view->perspective = hdr->perspective;
  view->xobs = obs[0],
    view->yobs = obs[1],
    view->zobs = obs[2];
  view->xcenter = center[0],
    view->ycenter = center[1],
    view->zcenter = center[2];
  view->yrot = view->xrot = 0;
  view->nu = shared->nu;
  view->n = .001;
  view->f = 1000.0;
  
  /* calculate by projecting the vup vector (i.e. the y vector)
     onto the st plane and calculate angle to tvec
     - if normal parallel to y, the use negative z instead */
  if(fabs(stnormal[1] - 1.0) <= 0.0000001) {
    tmp2 = -1 * stnormal[2]; /* dot product of -z and stnormal */
    vpSetVector3(tmp, 0, 0, -1);
  } else {
    tmp2 = stnormal[1];      /* dot product of y and stnormal */
    vpSetVector3(tmp, 0, 1, 0);
  }
  vpSetVector3(tmp, tmp[0] - tmp2 * stnormal[0],
	       tmp[1] - tmp2 * stnormal[1],
	       tmp[2] - tmp2 * stnormal[2]);
  vpNormalize3(tmp);
  view->zrot = acos(vpDotProduct3(tvec, tmp)) * 180 / PI;
  /* on left side of tvec? */
  if(vpDotProduct3(svec, tmp) < 0) {
    view->zrot -= 180;
  }
  
  /* set all arguments that the renderer can accept */
  lfExtractParams(params->num, &pos[LF_P_POS(0)], shared->psamples, values);
  for(i = 0; i < renderer->num_args; i++) {
    int arg_index = renderer->arg_indices[i];
    params->params[arg_index].current = values[arg_index];
  }
  
  if(!lfRDRender(renderer, NULL, NULL, shared->ns, shared->nt,
		shared->sample_size == 3 ? LF_RGB : LF_RGBA,
		image)) {
    lfError("lfReadSliceFile: renderer failed.\n");
    exit(1);
  }
  return image;
}



/*
 * function to input a slice file
 */
bool_t __LFRowMajor; /* xxx: hack for panalysis: don't rearrange image in memory */
void
lfReadSliceFile(LFSlab **slabs, int nslab, const int *pos,
	const float *descr, const void *input, int input_size,
	void *output, int output_size, void *aux_ptr)
{
#ifndef NO_INTERNAL_READ_RGB
  char slice_name[256];
  LFSliceHeader *hdr = aux_ptr;
  LFSlab *slab = slabs[pos[LF_SLAB_POS]];
  LFShared *shared = slab->shared;
  LFParams *params = __LF_GET_PARAMS();
  int slab_id = slab->id;
  char *ip = output, *sp;
  int x, y, line_size;
  LFPixelRGBA8 *lpa;
  LFPixelRGB8 *lp;
  GLubyte *image;
  GLint xsize, ysize;
  GLenum format;
  int i;
  int values[LF_MAXPARAMS];

  /* no slabs? */
  if(!shared)
    return;

  strcpy(slice_name, hdr->slab_names[slab_id]);
  lfExtractParams(params->num, &pos[LF_P_POS(0)], shared->psamples, values);
  for(i = 0; i < params->num; i++) {
    sprintf(&slice_name[strlen(slice_name)], ".%d", values[i]);
  }
  sprintf(&slice_name[strlen(slice_name)], ".%d.%d%s",
	  pos[LF_U_POS], pos[LF_V_POS],
	  hdr->slab_suffices[slab_id]);
  lfOutput("image %s... ", slice_name);
  if(hdr->no_images) {
    image = NULL;
    lfOutput("ignored.\n");
  } else {
    image = LoadRGBImage(slice_name, &xsize, &ysize, &format);
    if(image) {
      lfOutput("loaded.\n");
    } else {
      lfOutput("failed.\n");
    }
  }

  if (!image) {
    LFRenderer *renderer = __LF_GET_RENDERER();

    if(!hdr->no_images && !lfVerbose) {
      lfError("lfReadSliceFile: cannot open image %s for slab %d\n",
	      slice_name, slab_id);
    }
    if(renderer->type == LF_RENDERER_NONE) {
      exit(1);
    } else {
      float u, v;

      if(shared->nu <= 1) {
	u = 0.5;
      } else {
	u = 1.0 - (2 * pos[LF_U_POS] + 1)/(float)(2 * shared->nu);
      }
      if(shared->nv <= 1) {
	v = 0.5;
      } else {
	v = 1.0 - (2 * pos[LF_V_POS] + 1)/(float)(2 * shared->nv);
      }
      image = RenderUV(u,v,slab,pos,hdr);
      if(hdr->uv_filter) {
	GLubyte *edges[4];
	float u_diff, v_diff;

	u_diff = shared->nu > 1 ? 1/(float)(4*shared->nu) : 0;
	v_diff = shared->nv > 1 ? 1/(float)(4*shared->nv) : 0;
	edges[0] = RenderUV(u-u_diff,v-v_diff,slab,pos,hdr);
	edges[1] = RenderUV(u-u_diff,v+v_diff,slab,pos,hdr);
	edges[2] = RenderUV(u+u_diff,v-v_diff,slab,pos,hdr);
	edges[3] = RenderUV(u+u_diff,v+v_diff,slab,pos,hdr);
	i = shared->sample_size * shared->ns * shared->nt;
	while(--i >= 0) {
	  image[i] = (image[i] + edges[0][i] + edges[1][i] + edges[2][i] + edges[3][i])
	    / 5;
	}
	for(i = 0; i < 4; i++) {
	  free(edges[i]);
	}
      }
      xsize = shared->ns;
      ysize = shared->nt;
      format = shared->sample_size == 3 ? GL_RGB : GL_RGBA;
      if(hdr->create_images) {
	WriteRGBImage(slice_name, xsize, ysize, format, image);
      }
    }
  }

  if  (xsize != shared->ns || ysize != shared->nt) {
    lfError("lfReadSliceFile: size of %s mismatches size of slab %d\n",
	    slice_name, slab_id);
    exit(1);
  }

  if (!((shared->sample_size == 3 || shared->sample_size == 4) &&
	(format == GL_RGB || format == GL_RGBA))) {
    lfError("lfReadSliceFile: %s has wrong format (does not contain %ld samples)\n",
	    slice_name, shared->sample_size);
    exit(1);
  }

  sp = image;
  if(__LFRowMajor) {
    if(shared->sample_size == 3) {
      lp = (LFPixelRGB8 *)ip;
      for ( y=0 ; y<ysize ; y++) {
	for ( x=0 ; x<xsize ; x++) {
	  lp->r = *sp++;
	  lp->g = *sp++;
	  lp->b = *sp++;
	  lp++;
	  if(format == GL_RGBA) {
	    sp++;
	  }
	}
      }	  
    } else {
      lpa = (LFPixelRGBA8 *)ip;
      for ( y=0 ; y<ysize ; y++) {
	for ( x=0 ; x<xsize ; x++) {
	  lpa->r = *sp++;
	  lpa->g = *sp++;
	  lpa->b = *sp++;
	  lpa++;
	  if(format == GL_RGBA) {
	    lpa->a = *sp++;
	  } else {
	    lpa->a = 0xFF;
	  }
	}
      }
    }
  } else {
    line_size = shared->nt * shared->sample_size;
    if  (shared->sample_size == 3)  {
      for ( y=0 ; y<ysize ; y++,ip+=shared->sample_size) {
	for ( x=0,lp=(LFPixelRGB8 *)ip; x<xsize ; x++,lp=(LFPixelRGB8 *)&((char *)lp)[line_size]) {
	  lp->r = *sp++;
	  lp->g = *sp++;
	  lp->b = *sp++;
	  if(format == GL_RGBA) {
	    sp++;
	  }
	}
      }
    }
    else {
      for ( y=0 ; y<ysize ; y++,ip+=shared->sample_size ) {
	for ( x=0,lpa=(LFPixelRGBA8 *)ip; x<xsize ; x++,lpa=(LFPixelRGBA8 *)&((char *)lpa)[line_size]) {
	  lpa->r = *sp++;
	  lpa->g = *sp++;
	  lpa->b = *sp++;
	  if(format == GL_RGBA) {
	    lpa->a = *sp++;
	  } else {
	    lpa->a = 0xFF;
	  }
	}
      }
    }
  }
  free(image);

#else /* NO_INTERNAL_READ_RGB */
  IMAGE *image;
  LFSliceHeader *hdr = aux_ptr;
  LFSlab *slab = slabs[pos[LF_SLAB_POS]];
  LFShared *shared = slab->shared;
  char *ip = output, *lp;
  int slab_id = slab->id;
  short rbuf[8192];
  short gbuf[8192];
  short bbuf[8192];
  short abuf[8192];
  int line_size;
  int x, xsize;
  int y, ysize;
  int zsize;
  char slice_name[256];

  sprintf(slice_name, "%s.%d.%d.rgb",
	  hdr->slab_names[slab_id], pos[LF_U_POS], pos[LF_V_POS]);

  if  ((image=iopen(slice_name, "r")) == NULL) {
    lfError("lfReadSliceFile: cannot read %s\n", slice_name);
    exit(1);
  }

  xsize = image->xsize;
  ysize = image->ysize;
  zsize = image->zsize;

  if  (zsize < 3) {
    lfError("lfReadSliceFile: %s is not an RGB image file\n",
	    slice_name);
    exit(1);
  }

  if  (xsize != shared->ns || ysize != shared->nt) {
    lfError("lfReadSliceFile: size of %s mismatches size of slab %d\n",
	    slice_name, slab_id);
    exit(1);
  }

  line_size = shared->nt * shared->sample_size;
  if  (shared->sample_size == 3)  {
    for ( y=0 ; y<ysize ; y++,ip+=shared->sample_size ) {
      getrow(image,rbuf,y,0);
      getrow(image,gbuf,y,1);
      getrow(image,bbuf,y,2);
      for ( x=0,lp=ip ; x<xsize ; x++,lp+=line_size ) {
	lp[2] = rbuf[x];
	lp[1] = gbuf[x];
	lp[0] = bbuf[x];
      }
    }
  }
  else {
    for ( y=0 ; y<ysize ; y++,ip+=shared->sample_size ) {
      getrow(image,rbuf,y,0);
      getrow(image,gbuf,y,1);
      getrow(image,bbuf,y,2);
      getrow(image,abuf,y,3);
      for ( x=0,lp=ip ; x<xsize ; x++,lp+=line_size ) {
	lp[3] = rbuf[x];
	lp[2] = gbuf[x];
	lp[1] = bbuf[x];
	lp[0] = abuf[x];
      }
    }
  }
  iclose(image);
#endif /* NO_INTERNAL_READ_RGB */
}
 
/*
 * function to end inputing all slice files in a lightfield
 */
void
lfEndReadSliceFile(LFOps *ops, LFInternOp *op, void *aux_ptr)
{
    LFSliceHeader *hdr = aux_ptr;
    int i;

    for ( i=0 ; i<hdr->nslab ; i++ ) {
	free(hdr->slab_names[i]);
    }
    free(hdr);
}
