/*
 * Read an SGI .rgb or .tiff image file.
 * Much of this code was borrowed from SGI's tk OpenGL toolkit for GLUT
 * and copied once more from there.
 */


#include <tiff.h>
#include <tiffio.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#include <ctype.h>


#ifndef SEEK_SET
#  define SEEK_SET 0
#endif


/*
** RGB Image Structure
*/

typedef struct _TK_RGBImageRec {
   GLint sizeX, sizeY;
   GLint components;
   unsigned char *data;
} TK_RGBImageRec;



/******************************************************************************/

typedef struct _rawImageRec {
    unsigned short imagic;
    unsigned short type;
    unsigned short dim;
    unsigned short sizeX, sizeY, sizeZ;
    unsigned long min, max;
    unsigned long wasteBytes;
    char name[80];
    unsigned long colorMap;
    FILE *file;
    unsigned char *tmp, *tmpR, *tmpG, *tmpB, *tmpA;
    unsigned long rleEnd;
    GLuint *rowStart;
    GLint *rowSize;
} rawImageRec;

/******************************************************************************/

static void ConvertShort(unsigned short *array, long length)
{
   unsigned long b1, b2;
   unsigned char *ptr;

   ptr = (unsigned char *)array;
   while (length--) {
      b1 = *ptr++;
      b2 = *ptr++;
      *array++ = (b1 << 8) | (b2);
   }
}

static void ConvertLong(GLuint *array, long length)
{
   unsigned long b1, b2, b3, b4;
   unsigned char *ptr;

   ptr = (unsigned char *)array;
   while (length--) {
      b1 = *ptr++;
      b2 = *ptr++;
      b3 = *ptr++;
      b4 = *ptr++;
      *array++ = (b1 << 24) | (b2 << 16) | (b3 << 8) | (b4);
   }
}

static rawImageRec *RawImageOpen(const char *fileName)
{
   union {
      int testWord;
      char testByte[4];
   } endianTest;
   rawImageRec *raw;
   GLenum swapFlag;
   int x;

   endianTest.testWord = 1;
   if (endianTest.testByte[0] == 1) {
      swapFlag = GL_TRUE;
   } else {
      swapFlag = GL_FALSE;
   }

   raw = (rawImageRec *)malloc(sizeof(rawImageRec));
   if (raw == NULL) {
      fprintf(stderr, "Out of memory!\n");
      return NULL;
   }
   if ((raw->file = fopen(fileName, "rb")) == NULL) {
      perror(fileName);
      return NULL;
   }

   fread(raw, 1, 12, raw->file);

   if (swapFlag) {
      ConvertShort(&raw->imagic, 6);
   }

   raw->tmp = (unsigned char *)malloc(raw->sizeX*256);
   raw->tmpR = (unsigned char *)malloc(raw->sizeX*256);
   raw->tmpG = (unsigned char *)malloc(raw->sizeX*256);
   raw->tmpB = (unsigned char *)malloc(raw->sizeX*256);
   if (raw->sizeZ==4) {
      raw->tmpA = (unsigned char *)malloc(raw->sizeX*256);
   }
   if (raw->tmp == NULL || raw->tmpR == NULL || raw->tmpG == NULL ||
       raw->tmpB == NULL) {
      fprintf(stderr, "Out of memory!\n");
      return NULL;
   }

   if ((raw->type & 0xFF00) == 0x0100) {
      x = raw->sizeY * raw->sizeZ * sizeof(GLuint);
      raw->rowStart = (GLuint *)malloc(x);
      raw->rowSize = (GLint *)malloc(x);
      if (raw->rowStart == NULL || raw->rowSize == NULL) {
         fprintf(stderr, "Out of memory!\n");
         return NULL;
      }
      raw->rleEnd = 512 + (2 * x);
      fseek(raw->file, 512, SEEK_SET);
      fread(raw->rowStart, 1, x, raw->file);
      fread(raw->rowSize, 1, x, raw->file);
      if (swapFlag) {
         ConvertLong(raw->rowStart, x/sizeof(GLuint));
         ConvertLong((GLuint *)raw->rowSize, x/sizeof(GLint));
      }
   }
   return raw;
}

static void RawImageClose(rawImageRec *raw)
{

   fclose(raw->file);
   free(raw->tmp);
   free(raw->tmpR);
   free(raw->tmpG);
   free(raw->tmpB);
   if (raw->sizeZ>3) {
      free(raw->tmpA);
   }
   free(raw);
}

static void RawImageGetRow(rawImageRec *raw, unsigned char *buf, int y, int z)
{
   unsigned char *iPtr, *oPtr, pixel;
   int count;

   if ((raw->type & 0xFF00) == 0x0100) {
      fseek(raw->file, raw->rowStart[y+z*raw->sizeY], SEEK_SET);
      fread(raw->tmp, 1, (unsigned int)raw->rowSize[y+z*raw->sizeY],
            raw->file);
      
      iPtr = raw->tmp;
      oPtr = buf;
      while (1) {
         pixel = *iPtr++;
         count = (int)(pixel & 0x7F);
         if (!count) {
            return;
         }
         if (pixel & 0x80) {
            while (count--) {
               *oPtr++ = *iPtr++;
            }
         } else {
            pixel = *iPtr++;
            while (count--) {
               *oPtr++ = pixel;
            }
         }
      }
   } else {
      fseek(raw->file, 512+(y*raw->sizeX)+(z*raw->sizeX*raw->sizeY),
            SEEK_SET);
      fread(buf, 1, raw->sizeX, raw->file);
   }
}


static void RawImageGetData(rawImageRec *raw, TK_RGBImageRec *final)
{
   unsigned char *ptr;
   int i, j;

   final->data = (unsigned char *)malloc((raw->sizeX+1)*(raw->sizeY+1)*4);
   if (final->data == NULL) {
      fprintf(stderr, "Out of memory!\n");
   }

   ptr = final->data;
   for (i = 0; i < (int)(raw->sizeY); i++) {
      RawImageGetRow(raw, raw->tmpR, i, 0);
      RawImageGetRow(raw, raw->tmpG, i, 1);
      RawImageGetRow(raw, raw->tmpB, i, 2);
      if (raw->sizeZ>3) {
         RawImageGetRow(raw, raw->tmpA, i, 3);
      }
      for (j = 0; j < (int)(raw->sizeX); j++) {
         *ptr++ = *(raw->tmpR + j);
         *ptr++ = *(raw->tmpG + j);
         *ptr++ = *(raw->tmpB + j);
         if (raw->sizeZ>3) {
            *ptr++ = *(raw->tmpA + j);
         }
      }
   }
}


static TK_RGBImageRec *tkRGBImageLoad(const char *fileName)
{
   rawImageRec *raw;
   TK_RGBImageRec *final;

   raw = RawImageOpen(fileName);
   if(raw == NULL)
     return NULL;
   final = (TK_RGBImageRec *)malloc(sizeof(TK_RGBImageRec));
   if (final == NULL) {
      fprintf(stderr, "Out of memory!\n");
      return NULL;
   }
   final->sizeX = raw->sizeX;
   final->sizeY = raw->sizeY;
   final->components = raw->sizeZ;
   RawImageGetData(raw, final);
   RawImageClose(raw);
   return final;
}


static void FreeImage( TK_RGBImageRec *image )
{
   free(image->data);
   free(image);
}



/*
 * Load an SGI .rgb file and return a pointer to the image data.
 * Input:  imageFile - name of .rgb to read
 * Output:  width - width of image
 *          height - height of image
 *          format - format of image (GL_RGB or GL_RGBA)
 * Return:  pointer to image data or NULL if error
 */
static GLubyte *LoadTIFFImage( const char *imageFile, GLint *width, GLint *height,
			       GLenum *format )
{
  GLubyte *buffer = NULL;
  TIFF* tif = TIFFOpen(imageFile, "r");
  if (tif) {
    uint32 w, h;
    size_t npixels;
    
    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
    TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h);
    npixels = w * h;
    *width = w;
    *height = h;
    *format = GL_RGBA;
    buffer = malloc(npixels * sizeof (uint32));
    if (buffer != NULL) {
      if (!TIFFReadRGBAImage(tif, w, h, (uint32 *)buffer, 0)) {
	free(buffer);
	buffer = NULL;
      }
    }
    TIFFClose(tif);
  }
  return(buffer);
}

static GLubyte *LoadSGIImage( const char *imageFile, GLint *width, GLint *height,
			      GLenum *format )
{
   TK_RGBImageRec *image;
   GLint bytes;
   GLubyte *buffer = NULL;

   image = tkRGBImageLoad( imageFile );
   if (!image) {
      return NULL;
   }

   if (image->components==3) {
      *format = GL_RGB;
   }
   else if (image->components==4) {
      *format = GL_RGBA;
   }
   else {
      /* not implemented */
      fprintf(stderr,
              "Error in LoadRGBImage %d-component images not implemented\n",
              (int)image->components );
      return NULL;
   }

   *width = image->sizeX;
   *height = image->sizeY;
   bytes = image->sizeX * image->sizeY * image->components;
   buffer = (GLubyte *) malloc(bytes);
   if (!buffer)
      return NULL;

   buffer = image->data;
   image->data = NULL;
   FreeImage(image);

   return buffer;
}

static void skip_spaces(FILE *in) {
  int next;

  while((next = fgetc(in)) != EOF) {
    if(!isspace(next)) {
      ungetc(next, in);
      break;
    }
  }
}

static GLubyte *LoadPPMImage( const char *imageFile, GLint *width, GLint *height,
			      GLenum *format )
{
   GLubyte *buffer = NULL;
   FILE *in;
   char id[2];

   if((in = fopen(imageFile, "rb"))) {
     if(fread(id, 1, 2, in) == 2 &&
	id[0] == 'P' && id[1] == '6') {
       int max_val;

       skip_spaces(in);
       fscanf(in, "%d", width);
       skip_spaces(in);
       fscanf(in, "%d", height);
       *format = GL_RGB;
       fscanf(in, "%d", &max_val);
       skip_spaces(in);

       if((buffer = malloc(*width * *height * 3))) {
	 if(fread(buffer, 1, *width * *height * 3, in) != *width * *height * 3) {
	   perror(imageFile);
	   free(buffer);
	   buffer = NULL;
	 }
       } else {
	 fprintf(stderr, "no memory for ppm pixmap of %dx%dx3!\n", *width, *height);
       }
     }
     fclose(in);
   }
   return buffer;
}

GLubyte *LoadRGBImage( const char *imageFile, GLint *width, GLint *height,
                       GLenum *format )
{
   GLubyte *buffer = NULL;
   
   /* try TIFF */
   if((buffer = LoadTIFFImage(imageFile, width, height, format))) {
     return buffer;
   }

   /* try PPM */
   if((buffer = LoadPPMImage(imageFile, width, height, format))) {
     return buffer;
   }

   /* try .rgb */
   if((buffer = LoadSGIImage(imageFile, width, height, format))) {
     return buffer;
   }

   return NULL;
}



/*
 * Write a .tiff file.
 * Input:  imageFile - name of .tiff to write
 * Output:  width - width of image
 *          height - height of image
 *          format - format of image (GL_RGB or GL_RGBA)
 *          image - the data in the given format
 * Return:  success
 */
int WriteRGBImage( const char *imageFile, GLint width, GLint height,
		   GLenum format, GLubyte *image)
{
  TIFF* tif = TIFFOpen(imageFile, "w");
  int success = 0;
  int samples = format == GL_RGBA ? 4 : 3;
  int i;

  if (tif) {
    TIFFSetField(tif, TIFFTAG_IMAGEWIDTH, width);
    TIFFSetField(tif, TIFFTAG_IMAGELENGTH, height);
    TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, 8);
    TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
    TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, samples);
    TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
/*    TIFFSetField(tif, TIFFTAG_TILEWIDTH, width);
    TIFFSetField(tif, TIFFTAG_TILELENGTH, height);*/
    TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, 1);

    /*    TIFFWriteEncodedTile(tif, TIFFComputeTile(tif, 0,0, 0,0),
			 (uint32 *)image, height * width * samples);*/
    for(i = height - 1; i >= 0; i--, image += width * samples) {
      TIFFWriteEncodedStrip(tif, i, (uint32 *)image, width * samples);
    }
    TIFFClose(tif);
  } else {
    fprintf(stderr, "WriteRGBImage: couldn't create %s\n", imageFile);
  }

  return(success);
}
