/*
** renderman.c - interface to RIB renderer
*/

#include "renderer.h"
#include "readrgb.h"

#include <math.h>
#include <stdlib.h>

/*
** qualities:
*/
enum {
  QUALITY_RGL,             /* crude  */
  QUALITY_RENDRIB_CRUDE,   /* medium = 1 pixel sample */
  QUALITY_RENDRIB_NORMAL,  /* medium = default pixel sample */
  QUALITY_RENDRIB_BLUR     /* high = default pixel sample + aperture */
};


bool_t lfRIBAlloc(LFRenderer *renderer)
{
  renderer->sup_qualities = (1<<QUALITY_RGL) | (1<<QUALITY_RENDRIB_CRUDE)
    | (1<<QUALITY_RENDRIB_NORMAL) | (1<<QUALITY_RENDRIB_BLUR);
  return(TRUE);
}
void lfRIBDone(LFRenderer *renderer)
{
}
bool_t lfRIBRender(LFRenderer *renderer,
		   Matrix4 mat_model, Matrix4 mat_proj, int width, int height,
		   int format, void *pixels)
{
  FILE *header;
  bool_t success = FALSE;
  LFView *view = &__LF_GET_CONTEXT()->view;

  if((header = fopen("/tmp/rib_header.rib", "w")) != 0) {
    vpVector3 vpn;
    float distance, xrot, yrot;
    char command[512], fnm[256];
    float fov = 0, frameaspectratio, screen_left, screen_right, screen_lower, screen_upper, screen_d;
    float shear_x, shear_y;

    lfOutput("=> center %f/%f/%f, obs %f/%f/%f\n", view->xcenter, view->ycenter, view->zcenter, view->xobs, view->yobs, view->zobs);
    vpSetVector3(vpn, view->xobs - view->xcenter,
		 view->yobs - view->ycenter,
		 view->zobs - view->zcenter);
    distance = sqrt(vpDotProduct3(vpn, vpn));
    lfOutput("=> dx %lf, dy %lf, dz %lf, d %lf\n", vpn[0], vpn[1], vpn[2], distance);
    xrot = atan2(vpn[1], sqrt(vpn[0]*vpn[0] + vpn[2]*vpn[2])) * 180 / PI
      + view->xrot;
    yrot = -atan2(vpn[0], vpn[2]) * 180 / PI
      + view->yrot;
    lfOutput("=> x-rot %lf, y-rot %lf\n", xrot, yrot);

    frameaspectratio = width / height;
    if(view->fov == 0.0) {
      shear_x = atan2(view->xcenter,view->screen_d) * 180 / PI;
      shear_y = atan2(view->ycenter,view->screen_d) * 180 / PI;
      screen_d = view->screen_d;
      lfOutput("=> shear %f/%f\n", shear_x, shear_y);
    } else {
      shear_x = 0;
      shear_y = 0;
      screen_d = distance;
    }

    if(view->perspective) {
      if(view->fov == 0.0) {
	fov = atan2((view->screen_upper - view->screen_lower)/2, view->screen_d)
	  * 2 * 180 / PI;
      } else {
	fov = view->fov;
      }
      if(frameaspectratio >= 1.0) {
	screen_left  = -frameaspectratio;
	screen_right = frameaspectratio;
	screen_lower = -1;
	screen_upper = +1;
      } else {
	screen_left  = -1;
	screen_right = +1;
	screen_lower = -frameaspectratio;
	screen_upper = +frameaspectratio;
      }
    } else {
      if(view->fov == 0.0) {
	float width2 = (view->screen_right - view->screen_left) / 2;
	float height2 = (view->screen_upper - view->screen_lower) / 2;

	screen_left = -width2;
	screen_right = width2;
	screen_lower = -height2;
	screen_upper =  height2;
      } else {
	lfError("orthogonal viewing not implemented");
	exit(1);
      }
    }

    fprintf(header, "Display \"/tmp/rib.tiff\" \"file\" \"%s\"\n", format == GL_RGB ? "rgb" : "rgba");
    fprintf(header, "##RenderMan RIB-Structure 1.0\n");
    fprintf(header, "Format %d %d 1.0\n", width, height);
    if(renderer->quality <= QUALITY_RENDRIB_CRUDE) {
      fprintf(header, "PixelSamples 1 1\n");
    }
    if(renderer->quality >= QUALITY_RENDRIB_BLUR && view->nu) {
      fprintf(header, "DepthOfField %f %f %f\n",
	      view->nu * screen_d / (screen_right - screen_left),
	      screen_d, screen_d);
    }
    if(view->perspective) {
      fprintf(header, "Projection \"perspective\" \"fov\" [ %f ] \n", fov);
    } else {
      fprintf(header, "Projection \"orthographic\"\n");
    }
    fprintf(header, "ScreenWindow %f %f %f %f\n",
	    screen_left,
	    screen_right,
	    screen_lower,
	    screen_upper);
    fprintf(header, "Skew [ %f   0 0 1   1 0 0 ]\n", shear_x);
    fprintf(header, "Skew [ %f   0 0 1   0 1 0 ]\n", shear_y);
    fprintf(header, "Translate 0 0 %f\n", distance);
    fprintf(header, "Rotate %f 1 0 0\n", -xrot);
    fprintf(header, "Rotate %f 0 1 0\n", -yrot);
    fprintf(header, "Translate %f %f %f\n", -view->xcenter, -view->ycenter, view->zcenter);
    fclose(header);

    if(renderer->config[0] == '/' ||
       renderer->lidDir[0] == 0) {
      strcpy(fnm, renderer->config);
    } else {
      sprintf(fnm, "%s/%s", renderer->lidDir, renderer->config);
    }
    sprintf(command, "cat /tmp/rib_header.rib \"%s\" | %s", fnm,
	    renderer->quality < QUALITY_RENDRIB_CRUDE ? "rgl -dumprgba" : "rendrib");
    lfOutput("%s ...", command);
    if(!system(command)) {
      GLubyte *image;
      int xsize, ysize;
      GLenum gl_format;

      image = LoadRGBImage("/tmp/rib.tiff", (GLint *)&xsize, (GLint *)&ysize, &gl_format);

      if(format == GL_RGB) {
	lfError("Renderman renderer cannot generate RGB only data\n");
	exit(1);
      }
      if(xsize != width || ysize != height) {
	lfError("Renderman generated picture of %dx%d pixels instead of %dx%d.\n"
		"Please remove Format command from .rib file.\n",
		xsize, ysize, width, height);
	exit(1);
      }
      if(image) {
	memcpy(pixels, image, 4 * xsize * ysize);
	success = TRUE;
      }
    }
    lfOutput(" done\n");
  }
  return(success);
}

void lfRIBRenderParams(LFRenderer *renderer)
{
}
