/*
 * resample.c
 *
 * various routines for resampling the lightfield when drawing
 */

#include "lightfield.h"
#include "lerp.h_"       /* lerp.h cannot be used because its macros stretch over multiple lines,
			    which would prevent their use in the definion of another macro */

#define	NEEDCLAMP(x, xlo, xhi)	\
	((x < xlo) || (x > xhi))

#define CLAMP(x, xlo, xhi)      \
	if  (x > xhi) x = xhi;  \
	else if  (x < xlo) x = xlo

#define	NEEDCLAMP_SAMPLE(mmap)	\
	(  NEEDCLAMP((mmap)->u, 0, uclamp) \
	|| NEEDCLAMP((mmap)->v, 0, vclamp) \
	|| NEEDCLAMP((mmap)->s, 0, sclamp) \
	|| NEEDCLAMP((mmap)->t, 0, tclamp) )

static void
clampScanline(LFView *view, LFPixelUVST *scanline, int w,
	int uclamp, int vclamp, int sclamp, int tclamp)
{
    int cur_x;

#ifdef	STATS
    view->scanline_clamped++;
#endif
    for ( cur_x=0 ; cur_x<w ; cur_x++,scanline++ ) {
	CLAMP(scanline->u, 0, uclamp);
	CLAMP(scanline->v, 0, vclamp);
	CLAMP(scanline->s, 0, sclamp);
	CLAMP(scanline->t, 0, tclamp);
    }
}


void
lfResampleRGBA8(const LFSlab *slab, LFPixelUVST *scanline, int x, int y, int w)
{
    LFView *view = &__LF_GET_CONTEXT()->view;
    LFPixelUVST *mapp  = scanline;
#ifndef ALPHA_OUT
    LFPixelOutputRGB8 *imgp = ((LFPixelOutputRGB8 *)view->pixel_buffer) + y*view->width + x;
#else
    LFPixelOutputRGBA8 *imgp = ((LFPixelOutputRGBA8 *)view->pixel_buffer) + y*view->width + x;
#endif
#ifndef	SCAN_ONLY
    bool_t uvlerp = (view->filter & LF_LERP_UV) == LF_LERP_UV;
    bool_t stlerp = (view->filter & LF_LERP_ST) == LF_LERP_ST;
    LFShared *shared = slab->shared;
    LFVQCodebook *vq = shared->vq;
    int frac_ubits = shared->u_frac_bits;
    int frac_vbits = shared->v_frac_bits;
    int frac_sbits = shared->s_frac_bits;
    int frac_tbits = shared->t_frac_bits;
    int frac_umsk = (1 << frac_ubits) - 1;
    int frac_vmsk = (1 << frac_vbits) - 1;
    int frac_smsk = (1 << frac_sbits) - 1;
    int frac_tmsk = (1 << frac_tbits) - 1;
    int u, v, s, t;
    int fu, fv, fs, ft;
    int num_interp = shared->plan.num_interp;
    int *fp = shared->plan.fp;
    int *fp_bits = shared->plan.fp_bits;
    /* BEGIN stuff used inside macros */
    LFPixelRGBA8 *p00, *p10, *p01, *p11;
    LFPixelRGBA8 c00,c10,c01,c11;
    int *uindex = shared->uindex;
    int *vindex = shared->vindex;
    int *sindex = shared->sindex;
    int *tindex = shared->tindex;
    int uclamp = (shared->nu << shared->u_frac_bits) - 1;
    int vclamp = (shared->nv << shared->v_frac_bits) - 1;
    int sclamp = (shared->ns << shared->s_frac_bits) - 1;
    int tclamp = (shared->nt << shared->t_frac_bits) - 1;
    /* END for macros */
    LFPixelOutputRGB8 interp[1 << LF_MAXPARAMS];
    int i, max_samples = shared->plan.num_accesses;
    const int *source = shared->plan.source;
    const int *target = shared->plan.target;

    if  (NEEDCLAMP_SAMPLE(mapp) || NEEDCLAMP_SAMPLE(mapp+w-1)) {
	clampScanline(view, scanline, w, uclamp, vclamp, sclamp, tclamp);
    }

    /* the body is the same in both versions of this function,
       so chooselerp.h defines it and maps xxxBOTHRGB()
       to the correct macros in lerp.h */
    if(!vq) {
      LFPixelRGBA8 **lfs = (LFPixelRGBA8 **)slab->lfs;

#undef VQLIGHTSOURCE
#include "chooselerp.h"
      switch(num_interp) {
      case 0:
	/* simple light field rendering */
	RESAMPLE_BODY(0, 1);
	break;
      default:
	RESAMPLE_BODY(num_interp, max_samples);
	break;
      }
    } else {
      unsigned short **lfs = (unsigned short **)slab->lfs;
      LFPixelRGBA8 **vqtiles_p = slab->vqtiles_p;
      int *vqtile_uindex = shared->vqtile_uindex;
      int *vqtile_vindex = shared->vqtile_vindex;
      int *vqtile_sindex = shared->vqtile_sindex;
      int *vqtile_tindex = shared->vqtile_tindex;
      int tile_shift = lfSumPBits(vq) + vq->ubits + vq->vbits + vq->sbits + vq->tbits;

#define VQLIGHTSOURCE
#include "chooselerp.h"
      switch(num_interp) {
      case 0:
	/* simple light field rendering */
	RESAMPLE_BODY(0, 1);
	break;
      case 1:
	switch(max_samples) {
	case 2:
	  RESAMPLE_BODY(1, 2);
	  break;
	default:
	  lfError("interp 1, max_samples %d???\n", max_samples);
	  break;
	}
	break;
      case 2:
	switch(max_samples) {
	case 3:
	  RESAMPLE_BODY(2, 3);
	  break;
	default:
	  lfError("interp 2, max_samples %d???\n", max_samples);
	  break;
	}
	break;
      case 3:
	switch(max_samples) {
	case 4:
	  RESAMPLE_BODY(3, 4);
	  break;
	default:
	  lfError("interp 3, max_samples %d???\n", max_samples);
	  break;
	}
	break;
      default:
	RESAMPLE_BODY(num_interp, max_samples);
	break;
      }
    }

#else
    for ( x=0 ; x<w ; x++,imgp++ )
	imgp->r = imgp->g = imgp->b = 0xff;
#endif
}

