/*
 * util.c
 *
 * utility stuff
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "lightfield.h"

/* Global variable, controls whether or not to print debug messages */
bool_t lfVerbose = 0;

/* Print error messages (if lfVerbose) */
void
lfError(char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    fprintf(stderr, "Error: ");
    vfprintf(stderr, fmt, args);
    fprintf(stderr, "\n");
    fflush(stderr);
    va_end(args);
}

/* Print output messages */
void
lfOutput(char *fmt, ...)
{
    va_list args;

    if (lfVerbose) {
      va_start(args, fmt);
      vfprintf(stdout, fmt, args);
      va_end(args);
      fflush(stdout);
    }
}

/* Calculate the floor of the log (base 2) of a number */
int
log2(int size)
{
    int i = 0;
    while ((size >>= 1) > 0) i++;
    return i;
}

/* Set up the indices for the offset tables for indexing into the
 * light field data array */
void
lfMakeIndices(LFShared *shared)
{
    int nu = shared->nu;
    int nv = shared->nv;
    int ns = shared->ns;
    int nt = shared->nt;
    int u, v, s, t;
    int subsize;

    if  (shared->uindex)
	free(shared->uindex);

    if  (! shared->vq)  {
	shared->uindex = calloc(nu+nv+ns+nt+4, sizeof(int));
	if  (! shared->uindex)  {
	    fprintf(stderr,"lfMakeIndices: can't allocate storage\n");
	    exit(1);
	}

	shared->vindex = shared->uindex + nu + 1;
	shared->sindex = shared->vindex + nv + 1;
	shared->tindex = shared->sindex + ns + 1;
	shared->vqtile_uindex = NULL;
	shared->vqtile_vindex = NULL;
	shared->vqtile_sindex = NULL;
	shared->vqtile_tindex = NULL;

	for ( t=0 ; t<nt ; t++ ) {
	    shared->tindex[t] = t;
	}
	shared->tindex[t] = shared->tindex[t-1];

	subsize = nt;
	for ( s=0 ; s<ns ; s++ ) {
	    shared->sindex[s] = s * subsize;
	}
	shared->sindex[s] = shared->sindex[s-1];

	subsize *= ns;
	for ( v=0 ; v<nv ; v++ ) {
	    shared->vindex[v] = v * subsize;
	}
	shared->vindex[v] = shared->vindex[v-1];

	subsize *= nv;
	for ( u=0 ; u<nu ; u++ )  {
	    shared->uindex[u] = u * subsize;
	}
	shared->uindex[u] = shared->uindex[u-1];
    }
    else {
	int vq_ubits = shared->vq->ubits;
	int vq_vbits = shared->vq->vbits;
	int vq_sbits = shared->vq->sbits;
	int vq_tbits = shared->vq->tbits;
	int vq_umsk = (1 << vq_ubits) - 1;
	int vq_vmsk = (1 << vq_vbits) - 1;
	int vq_smsk = (1 << vq_sbits) - 1;
	int vq_tmsk = (1 << vq_tbits) - 1;

	/* XXX: need to patch in 1 extra entry to avoid checking later */
	shared->uindex = calloc((nu+nv+ns+nt+4)<<1, sizeof(int));
	if  (! shared->uindex)  {
	    fprintf(stderr,"lfMakeIndices: can't allocate storage\n");
	    exit(1);
	}

	shared->vindex = shared->uindex + nu + 1;
	shared->sindex = shared->vindex + nv + 1;
	shared->tindex = shared->sindex + ns + 1;
	shared->vqtile_uindex = shared->tindex + nt + 1;
	shared->vqtile_vindex = shared->vqtile_uindex + nu + 1;
	shared->vqtile_sindex = shared->vqtile_vindex + nv + 1;
	shared->vqtile_tindex = shared->vqtile_sindex + ns + 1;

	for ( u=0 ; u<nu ; u++ ) {
	    shared->uindex[u] = ((u>>vq_ubits)*nv*ns*nt) >> (vq_vbits+vq_sbits+vq_tbits);
	    shared->vqtile_uindex[u] = (u&vq_umsk) << (vq_vbits+vq_sbits+vq_tbits);
	}
	shared->uindex[u] = shared->uindex[u-1];
	shared->vqtile_uindex[u] = shared->vqtile_uindex[u-1];

	for ( u=0 ; u<nu ; u++ ) {
	    shared->uindex[u] = ((u>>vq_ubits)*nv*ns*nt) >> (vq_vbits+vq_sbits+vq_tbits);
	    shared->vqtile_uindex[u] = (u&vq_umsk) << (vq_vbits+vq_sbits+vq_tbits);
	}
	shared->uindex[u] = shared->uindex[u-1];
	shared->vqtile_uindex[u] = shared->vqtile_uindex[u-1];

	for ( v=0 ; v<nv ; v++ ) {
	    shared->vindex[v] = ((v>>vq_vbits)*ns*nt) >> (vq_sbits+vq_tbits);
	    shared->vqtile_vindex[v] = (v&vq_vmsk) << (vq_sbits+vq_tbits);
	}
	shared->vindex[v] = shared->vindex[v-1];
	shared->vqtile_vindex[v] = shared->vqtile_vindex[v-1];

	for ( s=0 ; s<ns ; s++ ) {
	    shared->sindex[s] = ((s>>vq_sbits)*nt) >> vq_tbits;
	    shared->vqtile_sindex[s] = (s&vq_smsk) << vq_tbits;
	}
	shared->sindex[s] = shared->sindex[s-1];
	shared->vqtile_sindex[s] = shared->vqtile_sindex[s-1];

	for ( t=0 ; t<nt ; t++ ) {
	    shared->tindex[t] = (t>>vq_tbits);
	    shared->vqtile_tindex[t] = (t&vq_tmsk);
	}
	shared->tindex[t] = shared->tindex[t-1];
	shared->vqtile_tindex[t] = shared->vqtile_tindex[t-1];
    }
}

/*
 * routine to pad data.
 * We do this to improve cache performance.  If the 
 * image size is a multiple of the cache size (or vice
 * versa), we may get excessive cache swapping.
 */
void
lfPadData(LFSlab *slab, int pad_byte, int new_size, int mode)
{
    LFShared *shared = slab->shared;
    LFVQCodebook *vq = shared->vq;
    char *freed, *src;
    char *alloced, *dst;
    int total_size;
    int old_size;
    int pad_size;
    int i, j;

    if  (vq) {
	if  (vq->sample_size == new_size)
	    return;
	freed = src = vq->codebook;
	old_size = vq->sample_size;
	total_size = (vq->size << (lfSumPBits(vq) + vq->ubits+vq->vbits+vq->sbits+vq->tbits))
			* new_size;
    }
    else {
        slab->sample_size = shared->sample_size;
	if  (slab->sample_size == new_size)
	    return;
	freed = src = slab->lightfield;
	old_size = shared->sample_size;
	total_size = MAX_PINDEX(shared)
	  * shared->nu * shared->nv * shared->ns * shared->nt
	  * new_size;
    }
    pad_size = new_size - old_size;
    dst = alloced = malloc(total_size);

    if  (mode == LF_PAD_FIRST)  {
	for ( i=0 ; i<total_size ; i+=new_size ) {
	    for ( j=0 ; j<pad_size ; j++ )
		*dst++ = pad_byte;
	    for ( j=0 ; j<old_size ; j++ )
		*dst++ = *src++;
	}
    }
    else {
	for ( i=0 ; i<total_size ; i+=new_size ) {
	    for ( j=0 ; j<old_size ; j++ )
		*dst++ = *src++;
	    for ( j=0 ; j<pad_size ; j++ )
		*dst++ = pad_byte;
	}
    }

    free(freed);
    if  (vq) {
	vq->codebook = alloced;
	vq->sample_size = new_size;
    }
    else {
	slab->lightfield = alloced;
	slab->sample_size = new_size;
    }
}


int lfSum(const int *values, int min_p, int max_p)
{
  int sum, p;

  for(p = min_p, sum = 0;
      p < max_p;
      p++) {
    sum += values[p];
  }
  return(sum);
}

int lfMul(const int *values, int min_p, int max_p)
{
  int mul, p;

  for(p = min_p, mul = 1;
      p < max_p;
      p++) {
    mul *= values[p];
  }
  return(mul);
}

char *dirname(char *name)
{
  char *end = strrchr(name, '/');
  if(end)
    *end = 0;
  else
    *name = 0;
  return(name);
}

/* expects RGBA byte order: */
float *lfIntensity;
float lfCalcIntensity(unsigned char *pixels, int width, int height, int sample_size)
{
  long double result = 0;
  int num_pixels = width * height;

  if(sample_size == 4) {
    while(--num_pixels >= 0) {
      result += *pixels++;
      result += *pixels++;
      result += *pixels++;
      pixels++;
    }
  } else {
    num_pixels *= sample_size;
    while(--num_pixels >= 0) {
      result += *pixels++;
    }
  }
  return(result/width/height/(sample_size == 4 ? 3 : sample_size));
}
float lfCalcIntensityStride(unsigned char *pixels, int width, int height, int stride)
{
  long double result = 0;
  int num_pixels = width * height;

  while(--num_pixels >= 0) {
    result += *pixels;
    pixels += stride;
  }
  return(result/width/height);
}

/* calculate turn parameter indices into linear index, with and without vq tiling */
static int lfVQIndexTileStartRec(int num_params, bool_t prev_limit,
				 int *pbits, int *j, LFParamSamples *psamples,
				 int *tile_start)
{
  int index = 0;
  int i, next_i;

  switch(num_params) {
  case 0:
    return 0;
    break;
  case 1:
    if(prev_limit) {
      if(tile_start) {
	tile_start[0] = (j[0] >> pbits[0]) << pbits[0];
      }
      return j[0];
    } else {
      return psamples[0].np;
    }
    break;
  default:
    for(i = 0;
	i < (prev_limit ? min(j[0] + 1, psamples[0].np) : psamples[0].np);
	i = next_i) {
      next_i = i + (1<<pbits[0]);
      index += lfVQIndexTileStartRec(num_params - 1, prev_limit && next_i > j[0],
				     &pbits[1], &j[1], psamples[0].samples[i].psamples,
				     tile_start ? &tile_start[1] : NULL);
    }
    if(prev_limit && tile_start) {
      tile_start[0] = (j[0] >> pbits[0]) << pbits[0];
    }
    return index;
    break;
  }
}

int lfVQIndexTileStart(int num_params, int pbits[LF_MAXPARAMS], int j[LF_MAXPARAMS], LFParamSamples *psamp,
		       int tile_start[LF_MAXPARAMS])
{
  return lfVQIndexTileStartRec(num_params, TRUE, pbits, j, psamp, tile_start);
}

int lfIndex(int num_params, int j[LF_MAXPARAMS], LFParamSamples *psamp)
{
  static int pbits[LF_MAXPARAMS];

  return(lfVQIndex(num_params, pbits, j, psamp));
}

int lfMaxVQIndex(int num_params, int pbits[LF_MAXPARAMS], LFParamSamples *psamples)
{
  int j[LF_MAXPARAMS];

  if(!num_params)
    return 1;

  j[0] = psamples->np;
  return lfVQIndex(num_params, pbits, j, psamples);
}
int lfMaxIndex(int num_params, LFParamSamples *psamples)
{
  int j[LF_MAXPARAMS];

  if(!num_params)
    return 1;

  j[0] = psamples->np;
  return lfIndex(num_params, j, psamples);
}

void lfExtractParams(int num_params, const int p[LF_MAXPARAMS], LFParamSamples *psamples, int values[LF_MAXPARAMS])
{
  int i;

  for(i = 0; i < num_params; i++) {
    values[i] = psamples->samples[p[i]].p;
    psamples = psamples->samples[p[i]].psamples;
  }
}

/* write complete sample array */
static void indent(FILE *outFile, int num_spaces)
{
  fprintf(outFile, "    ");
  while(--num_spaces >= 0) {
    fprintf(outFile, " ");
  }
}

void lfWritePSamples(FILE *outFile, int first_indent, int level_indent,
		     int pnum, int num_params, LFParamSamples *psamples)
{
  int i;

  if(pnum >= num_params) {
    return;
  }

  indent(outFile, pnum * level_indent + first_indent);
  fprintf(outFile, "%d\n", psamples->np);
  for(i = 0; i < psamples->np; i++) {
    indent(outFile, pnum * level_indent + first_indent);
    fprintf(outFile, "%d\n", psamples->samples[i].p);
    lfWritePSamples(outFile, first_indent, level_indent,
		    pnum + 1, num_params, psamples->samples[i].psamples);
  }
}

void lfCalcMaxNP(int num_params, LFParamSamples *psamples, int max_np[LF_MAXPARAMS])
{
  int i;

  if(!num_params) {
    return;
  }

  if(psamples->np > max_np[0]) {
    max_np[0] = psamples->np;
  }
  for(i = 0; i < psamples->np; i++) {
    lfCalcMaxNP(num_params - 1, psamples->samples[i].psamples, &max_np[1]);
  } 
}
