/*
** interface to volpack library
*/

#include "renderer.h"
#include "volume.h"

#ifndef NO_VP

#include "volpack.h"
#include "denfile.h"

#include <GL/gl.h>

#include <stdlib.h>
#include <math.h>
#ifndef min
#define min(x,y) ((x) < (y) ? (x) : (y))
#endif


typedef struct {		/* contents of a voxel */
    short normal;		/*   encoded surface normal vector */
    unsigned char density;	/*   original density */
    unsigned char gradient;	/*   original gradient */
} RawVoxel;

RawVoxel *dummy_voxel;

#define BYTES_PER_VOXEL	sizeof(RawVoxel)	/* voxel size in bytes */
#define VOXEL_FIELDS	3	/* number of fields in voxel */
#define SHADE_FIELDS	2	/* number of fields used for shading
				   (normal and density); must be the
				   1st fields of RawVoxel */
#define CLSFY_FIELDS	2	/* number of fields used for classifying
				   (density and gradient); can be any fields
				   in the RawVoxel */

#define NORMAL_FIELD	0
#define NORMAL_OFFSET	vpFieldOffset(dummy_voxel, normal)
#define NORMAL_SIZE	sizeof(short)
#define NORMAL_MAX	VP_NORM_MAX

#define DENSITY_FIELD	1
#define DENSITY_OFFSET	vpFieldOffset(dummy_voxel, density)
#define DENSITY_SIZE	sizeof(unsigned char)

#define GRADIENT_FIELD	2
#define GRADIENT_OFFSET	vpFieldOffset(dummy_voxel, gradient)
#define GRADIENT_SIZE	sizeof(unsigned char)
#define GRADIENT_MAX	VP_GRAD_MAX

#define DENSITY_PARAM		0		/* classification parameters */
#define OCTREE_DENSITY_THRESH	4
#define GRADIENT_PARAM		1
#define OCTREE_GRADIENT_THRESH	4
#define OCTREE_BASE_NODE_SIZE	4

#define DENSITY_RAMP_POINTS 3			/* classification ramps */
int DensityRampX[] =    {  0,  24, 255};
float DensityRampY[] =  {0.0, 1.0, 1.0};

#define GRADIENT_RAMP_POINTS 4
int GradientRampX[] =   {  0,   5,  20, 221};
float GradientRampY[] = {0.0, 0.0, 1.0, 1.0};
#define GRADIENT_RAMP_ABD_POINTS 3
int GradientRampAbdX[] =   {  0,   5, 221};
float GradientRampAbdY[] = {0.0, 1.0, 1.0};
float gradient_ramp[GRADIENT_MAX + 1];

/* XXX: hack for supporting smallbrain.den and abdomen.den */
bool_t abdomen_mode;

/*
** set new transfer parameters and return TRUE if they have changed
*/
static int old_position = -1, old_steepness = -1;
bool_t lfVPSetTransfer(vpContext *vpc, int position, int steepness,
		       int *old_pos, int *old_steep)
{
  int orig_stp = steepness, orig_pos = position;

  if(position == *old_pos &&
     steepness == *old_steep)
    return TRUE;

  lfVolSetRampVals(position, steepness);
  vpSetClassifierTable(vpc, DENSITY_PARAM, DENSITY_FIELD, density_ramp,
		       sizeof(density_ramp));

  lfOutput("reclassify volume (pos %ld, steep %ld)... ", position, steepness);
  if(vpClassifyVolume(vpc)) {
    lfError("VolPack error: %s\n", 
	    vpGetErrorString(vpGetError(vpc)));
    return FALSE;
  } else {
    lfOutput("done.\n");
  }

  *old_pos = orig_pos;
  *old_steep = orig_stp;
  return TRUE;
}

void *vpc_density;

bool_t lfVPInitVPC(vpContext *vpc, void *density, int x, int y, int z, int quality)
{
  int dummy;
  vpResult res;
  static int scale[][3] = { {8,8,2}, {4,4,1}, {2,2,1} };

  abdomen_mode = x > 128;
  if(abdomen_mode && quality < sizeof(scale) / sizeof(scale[0])) {
    void *red_density;

    /* resample */
    red_density = malloc((x / scale[quality][0]) * (y / scale[quality][1]) * (z / scale[quality][2]) * sizeof(char));
    vpResample3D(density, x, y, z, red_density, x/scale[quality][0], y/scale[quality][1], z/scale[quality][2], VP_UCHAR, VP_LINEAR_FILTER);
    free(density);
    density = red_density;
    x /= scale[quality][0];
    y /= scale[quality][1];
    z /= scale[quality][2];
  }

  vpSetVolumeSize(vpc, x, y, z);
  vpSetVoxelSize(vpc, BYTES_PER_VOXEL, VOXEL_FIELDS,
		 SHADE_FIELDS, CLSFY_FIELDS);
  vpSetVoxelField(vpc, NORMAL_FIELD, NORMAL_SIZE, NORMAL_OFFSET,
		  NORMAL_MAX);
  vpSetVoxelField(vpc, DENSITY_FIELD, DENSITY_SIZE, DENSITY_OFFSET,
		  DENSITY_MAX);
  vpSetVoxelField(vpc, GRADIENT_FIELD, GRADIENT_SIZE, GRADIENT_OFFSET,
		  GRADIENT_MAX);
  vpSetRawVoxels(vpc, malloc(x*y*z*sizeof(RawVoxel)),
		 x*y*z*sizeof(RawVoxel), sizeof(RawVoxel), 
		 x*sizeof(RawVoxel),
		 x*y*sizeof(RawVoxel));
  lfOutput("lfVPInitVPC: storing data in volume... ");
  vpVolumeNormals(vpc, density, x*y*z, DENSITY_FIELD, GRADIENT_FIELD, NORMAL_FIELD);
  volCalcHistogram(density, x*y*z, volHistogram, &volHisMax);
  lfOutput("done.\n");
  if(abdomen_mode) {
    vpRamp(gradient_ramp, sizeof(float),
	   GRADIENT_RAMP_ABD_POINTS,
	   GradientRampAbdX, GradientRampAbdY);
  } else {
    vpRamp(gradient_ramp, sizeof(float),
	   GRADIENT_RAMP_POINTS,
	   GradientRampX, GradientRampY);
  }
  vpSetClassifierTable(vpc, GRADIENT_PARAM, GRADIENT_FIELD,
		       gradient_ramp, sizeof(gradient_ramp));
  vpSetd(vpc, VP_MIN_VOXEL_OPACITY, 0.05);
  lfVPSetTransfer(vpc, 0, 100, &dummy, &dummy);
  lfOutput("lfVPInitVPC: calculating min/max octree... ");
  vpMinMaxOctreeThreshold(vpc, DENSITY_PARAM, 10);
  vpCreateMinMaxOctree(vpc, 0, 4);
  if((res = vpGetError(vpc)) != VP_OK) {
    lfError("VolPack error: %s", vpGetErrorString(res));
    return FALSE;
  } else {
    lfOutput("done.\n");
    vpc_density = density;
    return TRUE;
  }
}


bool_t lfVPAlloc(LFRenderer *renderer)
{
  vpContext *vpc;	        /* rendering context */
  unsigned char *density;	/* buffer for density data */
  int x, y, z;            /* dimensions of data */
  char fnm[256];

  /* set parameter rendering and return immediately, if requested */
  renderer->aspect = 1.5;
  if(renderer->flags & LFRF_NO_RENDERING) {
    return TRUE;
  }

  vpc = vpCreateContext();
  if(!vpc) {
    lfError("lfVPAlloc: no context\n");
    return FALSE;
  }
  renderer->handle = vpc;
  old_position = -1, old_steepness = -1;

  if(renderer->config[0] == '/' ||
     renderer->lidDir[0] == 0) {
    strcpy(fnm, renderer->config);
  } else {
    sprintf(fnm, "%s/%s", renderer->lidDir, renderer->config);
  }
  lfOutput("lfVPAlloc: loading %s... ", fnm);
  density = read_den(fnm, &x, &y, &z);
  if(!density) {
    lfError("lfVPAlloc: cannot read file %s (as .den)\n", fnm);
    goto failed;
  }
  lfOutput("done.\n");

  if(!lfVPInitVPC(vpc, density, x, y, z, renderer->quality)) {
  failed:
    vpDestroyContext(vpc);
    return FALSE;
  } else {
    return TRUE;
  }
}

void lfVPDone(LFRenderer *renderer)
{
  vpDestroyContext((vpContext *)renderer->handle);
  free(vpc_density);
  vpc_density = NULL;
}

bool_t lfVPRenderVPC(vpContext *vpc,
		     Matrix4 mat_model, Matrix4 mat_proj, int width, int height,
		     int format, void *pixels, LFShading *shading)
{
  LFParams *params = __LF_GET_PARAMS();
  LFRenderer *renderer = __LF_GET_RENDERER();
  vpResult res;
  int size = width * height;
  int line_size_source = width * 4;
  int i;
  void *real_pixels = pixels;
  bool_t result = FALSE;

  /* VolPack can only generate images with alpha channel;
     VP_RGB seems to be ignored and image also has to be mirrored
     => generate with VP_RGBA and throw away A, mirror */
  pixels = malloc(size * 4);
  
  /* check for and set new transfer parameters */
  if(!lfVPSetTransfer(vpc,
		      params->params[renderer->arg_indices[0]].current,
		      params->params[renderer->arg_indices[1]].current,
		      &old_position, &old_steepness)) {
    return FALSE;
  }

  /* set position */
  lfVolSetPosition(vpc, width, height, FALSE);
  vpSetImage(vpc, pixels, width, height, sizeof(LFPixelRGBA8) * width, VP_RGBA);

  /* set the shading parameters */
  if(abdomen_mode) {
    static float weight_table[DENSITY_MAX + 1][3];
    static float shade_table[NORMAL_MAX+1][3][3];	/* shading lookup table */
    int     bone_X[] = {   0,  60,  150, 255 };
    float   bone_Y[] = { 0.0, 0.0,  1.0, 1.0 };
    int   vessel_X[] = {   0,  40,   50,  52,  60, 255 };
    float vessel_Y[] = { 0,0, 0.0,  1.0, 1.0, 0.0, 0.0 };
    int   kidney_X[] = {   0,  30,   45,  60, 255 };
    float kidney_Y[] = { 1.0, 1.0,  0.5, 0.0, 0.0 };

    #define AMBIENT 30
    #define DIFFUSE 100
    #define SPECULAR 130
    #define MAT_BONE VP_MATERIAL0
    #define BONE_COLOR(x) (x)/255.0*238.0/255, (x)/255.0*238.0/255, (x)/255.0*224.0/255
    #define MAT_VESSEL VP_MATERIAL1
    #define VESSEL_COLOR(x) (x)/255.0*234.0/255, (x)/255.0*30.0/255, (x)/255.0*30.0/255
    #define MAT_KIDNEY VP_MATERIAL2
    #if REALLY_KIDNEY
    #define KIDNEY_COLOR(x) (x)/255.0*255.0/255, (x)/255.0*140.0/255, (x)/255.0*75.0/255
    #else
    #define KIDNEY_COLOR(x) (x)/255.0*242.0/255, (x)/255.0*232.0/255, (x)/255.0*159.0/255
    #endif

    vpRamp(&weight_table[0][0], sizeof(float) * 3,
	   sizeof(bone_X) / sizeof(bone_X[0]), bone_X, bone_Y);
    vpRamp(&weight_table[0][1], sizeof(float) * 3,
	   sizeof(vessel_X) / sizeof(vessel_X[0]), vessel_X, vessel_Y);
    vpRamp(&weight_table[0][2], sizeof(float) * 3,
	   sizeof(kidney_X) / sizeof(kidney_X[0]), kidney_X, kidney_Y);
    vpSetLookupShader(vpc, 3, 3,
		      NORMAL_FIELD, (void *)shade_table, sizeof(shade_table),
		      DENSITY_FIELD, (void *)weight_table, sizeof(weight_table));

    /* bone */
    vpSetMaterial(vpc, MAT_BONE, VP_AMBIENT, VP_BOTH_SIDES,
		  BONE_COLOR(AMBIENT));
    vpSetMaterial(vpc, MAT_BONE, VP_DIFFUSE, VP_BOTH_SIDES,
		  BONE_COLOR(DIFFUSE));
    vpSetMaterial(vpc, MAT_BONE, VP_SPECULAR, VP_BOTH_SIDES,
		  BONE_COLOR(SPECULAR));
    vpSetMaterial(vpc, MAT_BONE, VP_SHINYNESS, VP_BOTH_SIDES,10.0,0.0,0.0);

    /* blood vessel */
    vpSetMaterial(vpc, MAT_VESSEL, VP_AMBIENT, VP_BOTH_SIDES,
		  VESSEL_COLOR(AMBIENT));
    vpSetMaterial(vpc, MAT_VESSEL, VP_DIFFUSE, VP_BOTH_SIDES,
		  VESSEL_COLOR(DIFFUSE));
    vpSetMaterial(vpc, MAT_VESSEL, VP_SPECULAR, VP_BOTH_SIDES,
		  VESSEL_COLOR(SPECULAR));
    vpSetMaterial(vpc, MAT_VESSEL, VP_SHINYNESS, VP_BOTH_SIDES,10.0,0.0,0.0);

    /* kidney */
    vpSetMaterial(vpc, MAT_KIDNEY, VP_AMBIENT, VP_BOTH_SIDES,
		  KIDNEY_COLOR(AMBIENT));
    vpSetMaterial(vpc, MAT_KIDNEY, VP_DIFFUSE, VP_BOTH_SIDES,
		  KIDNEY_COLOR(DIFFUSE));
    vpSetMaterial(vpc, MAT_KIDNEY, VP_SPECULAR, VP_BOTH_SIDES,
		  KIDNEY_COLOR(SPECULAR));
    vpSetMaterial(vpc, MAT_KIDNEY, VP_SHINYNESS, VP_BOTH_SIDES,10.0,0.0,0.0);
    
    vpSetLight(vpc, VP_LIGHT0, VP_DIRECTION, -0.3, -0.3, 1.0 /* 0.3, 0.8, 1.0 */);
    vpEnable(vpc, VP_LIGHT_BOTH_SIDES, 0);
  } else {
    static float shade_table[NORMAL_MAX+1][3];	/* shading lookup table */

    vpSetLookupShader(vpc, 3, 1, NORMAL_FIELD, (void *)shade_table,
		      sizeof(shade_table), 0, NULL, 0);
    vpSetMaterial(vpc, VP_MATERIAL0, VP_AMBIENT, VP_BOTH_SIDES,
		  0.18, 0.18, 0.18);
    vpSetMaterial(vpc, VP_MATERIAL0, VP_DIFFUSE, VP_BOTH_SIDES,
		  0.35, 0.35, 0.35);
    vpSetMaterial(vpc, VP_MATERIAL0, VP_SPECULAR, VP_BOTH_SIDES,
		  0.39, 0.39, 0.39);
    vpSetMaterial(vpc, VP_MATERIAL0, VP_SHINYNESS, VP_BOTH_SIDES,10.0,0.0,0.0);
    vpSetLight(vpc, VP_LIGHT0, VP_DIRECTION, 0.3, 0.3, 1.0);
  }
  vpEnable(vpc, VP_LIGHT0, 1);
  vpEnable(vpc, VP_DEPTH_CUE, 0);

  if((res = vpGetError(vpc))) {
    lfError("VolPack error during intialisation of lightning: %s\n", 
	     vpGetErrorString(res));
    goto done;
  }

  /* compute shading lookup table */
  vpSetd(vpc, VP_MAX_RAY_OPACITY, 0.95);
  if (vpShadeTable(vpc) != VP_OK) {
    lfError("VolPack error shade table: %s\n",
	    vpGetErrorString(vpGetError(vpc)));
    goto done;
  }

  /* render */
  if (vpRenderClassifiedVolume(vpc) != VP_OK) {
    lfError("VolPack error rendering: %s\n",
	    vpGetErrorString(vpGetError(vpc)));
    goto done;
  }
  else {
    unsigned char *t = real_pixels;
    for(i = 0; i < height; i++) {
      unsigned char *s = &((unsigned char *)pixels)[(height - i - 1) * line_size_source];
      
      size = width;
      while(--size >= 0) {
	*t++ = *s++;
	*t++ = *s++;
	*t++ = *s++;
	/* copy without A? */
	if(format != LF_RGBA) {
	  s++;
	} else {
	  *t++ = *s++;
	}
      }
    }
    result = TRUE;
  }

 done:
  free(pixels);
  return result;
}

bool_t lfVPRender(LFRenderer *renderer,
		   Matrix4 mat_model, Matrix4 mat_proj, int width, int height,
		   int format, void *pixels)
{
  return lfVPRenderVPC((vpContext *)renderer->handle,
		       mat_model, mat_proj, width, height, format, pixels, NULL);
}




#else  /* NO_VP */


bool_t lfVPInitVPC(vpContext *vpc, void *density, int x, int y, int z, int quality)
{
  return TRUE;
}


bool_t lfVPAlloc(LFRenderer *renderer)
{
  /* only parameter rendering possible */
  renderer->aspect = 1.5;
  if(renderer->flags & LFRF_NO_RENDERING) {
    return TRUE;
  } else {
    return FALSE;
  }
}

void lfVPDone(LFRenderer *renderer)
{
}

bool_t lfVPRenderVPC(vpContext *vpc,
		     Matrix4 mat_model, Matrix4 mat_proj, int width, int height,
		     int format, void *pixels, LFShading *shading)
{
  return TRUE;
}

bool_t lfVPRender(LFRenderer *renderer,
		   Matrix4 mat_model, Matrix4 mat_proj, int width, int height,
		   int format, void *pixels)
{
  return TRUE;
}


#endif /* NO_VP */


void lfVPRenderParams(LFRenderer *renderer)
{
  lfVolRenderParams(renderer);
}




