/*
** code shared by VolVis and volpack renderer
*/

#include "volume.h"
#include "renderer.h"
#include "lerp.h"
#include <GL/gl.h>

#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <math.h>
#ifndef min
#define min(x,y) ((x) < (y) ? (x) : (y))
#endif

/*
** histogram of current volume
*/
int volHistogram[DENSITY_MAX + 1];
int volHisMax;

void volCalcHistogram(unsigned char *density, int size,
		      int histo[UCHAR_MAX],
		      int *max)
{
  int i;

  bzero(histo, sizeof(int) * UCHAR_MAX);
  while(--size>=0) {
    histo[*density++]++;
  }
  for(i = 0, *max = 0; i <= UCHAR_MAX; i++) {
    if(i > 0 && histo[i] > *max) {
      *max = histo[i];
    }
  }
}

float density_ramp[DENSITY_MAX + 1];
int ramp_x[4];
float ramp_y[4];
int steps;

void
ramp(dst, stride, num_points, ramp_x, ramp_y)
float *dst;	/* array to store ramp into */
int stride;	/* stride (in bytes) for dst */
int num_points;	/* number of linear segment endpoints */
int *ramp_x;	/* x coordinates of the endpoints (and indexes into dst) */
float *ramp_y;	/* y coordinates of the endpoints */
{
    int i, x;
    int lastx, nextx;
    double lasty, nexty, dydx, y;

    dst[*ramp_x * stride/sizeof(float)] = *ramp_y;
    if (num_points == 1)
	return;
    for (i = 1; i < num_points; i++) {
	lastx = ramp_x[i-1];
	lasty = ramp_y[i-1];
	nextx = ramp_x[i];
	nexty = ramp_y[i];
	dydx = (double)(nexty - lasty) / (double)(nextx - lastx);
	for (x = lastx+1, y = lasty+dydx; x < nextx; x++, y += dydx)
	    dst[x * stride/sizeof(float)] = y;
	dst[x * stride/sizeof(float)] = nexty;
    }
}



/*
** calculate ramp based on current parameters
*/
static void set_ramp(int position, int steepness)
{
  #define SET_XY(x,y) ramp_x[steps] = x, ramp_y[steps++] = y

  steps = 0;
  SET_XY(0,0);
  if(position > 0) {
    if(position > DENSITY_MAX) {
      SET_XY(DENSITY_MAX, 0);
    } else {
      SET_XY(position, 0);
    }
  } 
  if(position + steepness > DENSITY_MAX) {
    SET_XY(DENSITY_MAX, (float)(DENSITY_MAX - position) * 1.0/steepness);
  } else {
    SET_XY(position + steepness, 1.0);
  }
  if(ramp_x[steps - 1] < DENSITY_MAX) {
    SET_XY(DENSITY_MAX, 1.0);
  }
  bzero(density_ramp, sizeof(density_ramp));
  ramp(density_ramp, sizeof(float), steps, ramp_x, ramp_y);
}
void lfVolSetRampVals(int position, int steepness)
{
  set_ramp(position, steepness);
}
void lfVolSetRamp(LFRenderer *renderer)
{
  LFParams *params = &__LF_GET_CONTEXT()->field->params;
  int position = params->params[renderer->arg_indices[0]].current;
  int steepness = params->params[renderer->arg_indices[1]].current;

  set_ramp(position, steepness);
}

static void draw_ramp(float color[3])
{
  int i;

  glColor3f(color[0], color[1], color[2]);
  glBegin(GL_LINE_STRIP);
  for(i = 0; i < steps; i++) {
    glVertex2f(ramp_x[i]/(float)DENSITY_MAX, ramp_y[i]);
  }
  glEnd();
}


void lfVolRenderParams(LFRenderer *renderer)
{
  int i;
  LFField *field = __LF_GET_CONTEXT()->field;
  LFParams *params = __LF_GET_PARAMS();
  LFView *view = __LF_GET_VIEW();
  LFShared *sh = field->slabs[0] ? field->slabs[0]->shared : NULL;
  float color[3];

  glPushAttrib(GL_CURRENT_BIT);

  /* draw histogram */
  glLineWidth(1.0);
  glColor3f(1.0, .5, .5);
  glBegin(GL_LINE_STRIP);
  for(i = 0; i <= DENSITY_MAX; i++) {
    glVertex2f(i/(float)DENSITY_MAX, i && volHisMax ? volHistogram[i]/(float)volHisMax : 0.0);
  }
  glEnd();

  if(sh && !view->force_rendering) {
    bool_t interpolated[1<<2];
    int position[1<<2];
    int steepness[1<<2];
    int *target = sh->plan.target;
    int *source = sh->plan.source;

    color[0] =
      color[1] =
      color[2] = 1.0;

    /* draw samples in different shades of gray to indicate their level */
    for(i = 0; i < sh->plan.num_accesses; i++) {
      int pvalues[LF_MAXPARAMS];

      interpolated[i] = FALSE;
      lfExtractParams(params->num, sh->plan.p_indices[i], sh->psamples, pvalues);
      position[i] = pvalues[renderer->arg_indices[0]];
      steepness[i] = pvalues[renderer->arg_indices[1]];
      set_ramp(position[i], steepness[i]);
      glLineWidth(1.0);
/*       color[0] = */
/* 	color[1] = */
/* 	color[2] = 1.0; */
      draw_ramp(color);
    }
    /* now draw interpolated ramps */
    for(i = sh->plan.num_interp; --i >= 0;) {
      position[target[i]] = LERP(position[target[i]], position[source[i]], sh->plan.fp[i],sh->plan.fp_bits[i]);
      steepness[target[i]] = LERP(steepness[target[i]], steepness[source[i]], sh->plan.fp[i],sh->plan.fp_bits[i]);
      if(interpolated[target[i]] && interpolated[source[i]]) {
      glLineWidth(4.0);
/* 	color[0] = */
/* 	  color[1] = */
/* 	  color[2] = .25; */
      } else if(interpolated[target[i]] || interpolated[source[i]]) {
      glLineWidth(3.0);
/* 	color[0] = */
/* 	  color[1] = */
/* 	  color[2] = .5; */
      } else {
      glLineWidth(2.0);
/* 	color[0] = */
/* 	  color[1] = */
/* 	  color[2] = .75; */
      }
      interpolated[target[i]] = TRUE;
      set_ramp(position[target[i]], steepness[target[i]]);
      draw_ramp(color);
    }
  }

  /* draw current ramp */
  lfVolSetRamp(renderer);
  color[0] = 0;
  color[1] = 1.0;
  color[2] = 0;
  draw_ramp(color);

  glPopAttrib();
}


#ifndef NO_VP

/*
** view point selection for VolPack;
** also used for VolVis by reading the modelling matrix afterwards
*/

/* XXXX hack: VolVis ignores the warp matrix, so use
   these two variables and a hack in the VolVis source
   to shift the view window */
#ifndef NO_VV
extern float __VVXOff, __VVYOff;
#endif

void lfVolSetPosition(vpContext *vpc, int width, int height, bool_t volvis)
{
  LFView *view = &__LF_GET_CONTEXT()->view;
  vpVector3 vpn;     /* view plane normal vector */
  double win_left,
    win_right,
    win_lower,
    win_upper; /* distances from CW to right/top border of view plane window */
  float distance, screen_distance;
  float invf = volvis ? -1.0 : 1.0;
  double xrot, yrot;
  vpMatrix4 shear_mat;
  float xoff, yoff;   /* offset of screen at 'distance' */

  /* use rotations instead of PHIGS model */
  vpCurrentMatrix(vpc, VP_MODEL);
  vpIdentityMatrix(vpc);
  vpCurrentMatrix(vpc, VP_VIEW);
  vpIdentityMatrix(vpc);

  /* translate first, then orient camera */
  if(!volvis) {
    vpSeti(vpc, VP_CONCAT_MODE, VP_CONCAT_LEFT);
  }
  vpTranslate(vpc, invf * -view->xcenter,
	           view->ycenter,
	           -view->zcenter);
  vpSetVector3(vpn, view->xobs - view->xcenter,
	            view->yobs - view->ycenter,
                    view->zobs - view->zcenter);
  distance = sqrt(vpDotProduct3(vpn, vpn));
  xrot = -atan2(vpn[1], sqrt(vpn[0]*vpn[0] + vpn[2]*vpn[2])) * 180 / PI 
    - view->xrot;
  yrot = -atan2(vpn[0], vpn[2]) * 180 / PI
    + view->yrot;
  vpRotate(vpc, VP_Y_AXIS, invf * yrot);
  vpRotate(vpc, VP_X_AXIS, xrot);
  vpTranslate(vpc, 0.0, 0.0, -distance);

  if(view->fov != 0.0) {
    /* calculate size of window */
    win_right = distance * tan(view->fov/2.0 * PI / 180);
    win_left = -win_right;
    win_upper = win_right * height / width;
    win_lower = -win_upper;
    /* bug!? 
       xoff = -view->xcenter;
       yoff = -view->ycenter; */
    xoff =
      yoff = 0;
    screen_distance = distance;
  } else {
    float width = (view->screen_right - view->screen_left) / 2,
      height = (view->screen_upper - view->screen_lower) / 2;

    xoff = (view->screen_left + view->screen_right) / 2;
    yoff = (view->screen_lower + view->screen_upper) / 2;
    screen_distance = view->screen_d;
    win_left  = -width;
    win_right = width;
    win_lower = -height;
    win_upper = height;
  }

  if(volvis) {
    /* move window */
#ifndef NO_VV
    __VVXOff = xoff / screen_distance;
    __VVYOff = yoff / screen_distance;
#endif
  } else {
    /* shear object back into center and use symetric screen window */
    shear_mat[0][0] = 1;
    shear_mat[0][1] = 0;
    shear_mat[0][2] = invf * xoff/screen_distance;
    shear_mat[0][3] = 0;

    shear_mat[1][0] = 0;
    shear_mat[1][1] = 1;
    shear_mat[1][2] = -yoff/screen_distance;
    shear_mat[1][3] = 0;
  
    shear_mat[2][0] = 0;
    shear_mat[2][1] = 0;
    shear_mat[2][2] = 1;
    shear_mat[2][3] = 0;

    shear_mat[3][0] = 0;
    shear_mat[3][1] = 0;
    shear_mat[3][2] = 0;
    shear_mat[3][3] = 1;

    vpMultMatrix(vpc, shear_mat);
  }

  vpCurrentMatrix(vpc, VP_PROJECT);
  vpIdentityMatrix(vpc);
  vpWindow(vpc, VP_PARALLEL, win_left, win_right,
	   win_lower, win_upper, view->n, min(10.0, view->f));

#if defined(DETAILED_TRANS_LOG)
  lfOutput("=> center %f/%f/%f, obs %f/%f/%f\n", view->xcenter, view->ycenter, view->zcenter, view->xobs, view->yobs, view->zobs);
  lfOutput("=> dx %lf, dy %lf, dz %lf, d %lf\n", vpn[0], vpn[1], vpn[2], distance);
  lfOutput("=> x-rot %lf, y-rot %lf\n", xrot, yrot);
  lfOutput("=> shear x %f y %f\n", xoff, yoff);

  lfOutput("=> window %f/%f/%f/%f\n", 
	   win_left, win_right,
	   win_lower, win_upper);
#endif
}

#endif /* NO_VP */


/* if set, volume rendering will set the average opacity */
float *volOpacity;
float lfVolCalcOpacity(int *hist, float *ramp, int entries)
{
  int voxels = 0, cur_voxels;
  double opacity = 0;

  while(--entries >= 0) {
    cur_voxels = *hist++;
    opacity += cur_voxels * *ramp++;
    voxels += cur_voxels;
  }
  return opacity / voxels;
}
