/*
 *  panalysis.c: analyze effects of additional parameters
 *
 */

#include <GL/glut.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <time.h>
#include <math.h>
#include <ctype.h>
#include <assert.h>
#include <signal.h>
#include <stdarg.h>
#include <limits.h>
#include "lightfield.h"
#include "lerp.h"
#include "readrgb.h"
#include "volume.h"

#ifndef min
#define min(x,y) ((x) < (y) ? (x) : (y))
#endif


/* generate a random grid with/without minimum distance */
static void
RandomGrid(int *grid, int num_entries, int width, int height, bool_t poisson)
{
  bool_t *set;
  int x,y,pos;
  float min_dist = sqrt(((double)(width*height)) / num_entries / PI) * 0.9;
  int idist = min_dist;

  if(!(set = calloc(width * height, sizeof(*set)))) {
    lfError("no mem\n");
    exit(10);
  }

  while(num_entries) {
    x = rand() / (RAND_MAX / width);
    y = rand() / (RAND_MAX / height);
    pos = y * width + x;

    if(!set[pos]) {
      bool_t use = TRUE;
      if(poisson) {
	int x_, y_;
	for(y_ = max(0, y - idist); use && y_ <= min(height - 1, y + idist); y_++) {
	  for(x_ = max(0, x - idist); use && x_ <= min(width - 1, x + idist); x_++) {
	    if(set[y_ * width + x_] &&
	       sqrt((y_ - y) * (y_ - y) +
		    (x_ - x) * (x_ - x)) < min_dist) {
	      use = FALSE;
	    }
	  }
	}
      }
      if(use ) {
	set[pos] = TRUE;
	grid[--num_entries] = pos * 3;
	grid[--num_entries] = pos * 3 + 1;
	grid[--num_entries] = pos * 3 + 2;
      }
    }
  }
}


typedef float err_t;
typedef double lerr_t;
#define ERR_FACTOR 1 /* (3.0 * 100 * 100) */

struct Position {
  unsigned char *image;  /* original */
  #define IMAGE(m) image[Grid[CurrentSet][m]]
  lerr_t sum;            /* the total error of other samples introduced by removing it */
  /* int p0, p1;            the parameter values at this position */
  int p[2];              /* 0 = FALSE = !vert = horizontal, 1 = TRUE = vert */
  int e,  i;             /* and its indices */
  bool_t sampled;        /* this position will be sampled */

  /* for doing interpolation */
  double **cx, **cy;     /* if this position is sampled or interpolated for doing
			    further interpolation, these are the coefficients of
			    the basis functions in x and y direction, given
			    for all possible distances */
  int left, right, top, bottom; /* indices of neighbors */

  /* for being interpolated/sampled */
  struct interp {
    err_t error;           /* the error for using the interpolated one instead */
    bool_t vert;           /* sample within column(s) first, then in row;
			      else the other way around */
    struct sources {
      int num;               /* number of sources that were used for interpolation */
      int sources[4][2];     /* and their indices */
      double sfactor[4];     /* their influence */
      int fnum;              /* number of factors */
      double factor[3];      /* and the factors themselves */
      int source_indices[3][2]; /* plus indices of source/target in sources[4] */
#ifdef DEBUG
      int checksum[4];       /* checksums of the sources */
      err_t error;           /* the error for this entry */
#endif
    } sources[2];
#define ADD_SOURCE(sources, i, e, factor) (sources)->sources[(sources)->num][0] = i, \
	(sources)->sources[(sources)->num][1] = e, (sources)->sfactor[(sources)->num++] = factor
  } interp;
  err_t *contribs;         /* cache for changes in the error if the other entry is being removed;
			      entries are addressed by index of remaining sample */
  #ifdef CACHED
  unsigned int *vert_buffer; /* bit-wise flags that indicate if it it better to use vertical
				interpolation first when dropping the sample */
  #endif

  #define BITS_PER_ENTRY 32
  #define IS_SET(buffer, i) ((buffer)[i / BITS_PER_ENTRY] & (1<<(i%BITS_PER_ENTRY)))
  #define SET(buffer, i) (buffer)[i / BITS_PER_ENTRY] |= (1<<(i%BITS_PER_ENTRY))
  #define CLEAR(buffer, i) (buffer)[i / BITS_PER_ENTRY] &= ~(1<<(i%BITS_PER_ENTRY))

  /* for browsing */
  unsigned char *influence; /* intensity values */
  #define INFLUENCE(position, p1, p0) (position)->influence[(POSITION(0,Columns-1).p[0] - POSITION(0,0).p[0] + 1) * \
							   (p1 - POSITION(0,0).p[1]) + p0 - POSITION(0,0).p[0]]
  #define INFL_SIZE (POSITION(0,Columns-1).p[0] - POSITION(0,0).p[0] + 1) * \
    (POSITION(Rows-1,0).p[1] - POSITION(0,0).p[1] + 1)

  /* double linked list of non-interpolated Positions,
     sorted based on their error;
     NULL if interpolated */
  struct Position *next;

  /* indices as used in the liblight.a for the current sampling -
     calculated by PrintInterpolation() */
  int index[2];
} *Positions, *List;
int *LNeigh, *RNeigh, *TNeigh, *BNeigh; /* rows/columns to the left/right/top/bottom that still have
					   entries */
int MinErrorPos[2];
lerr_t MinError;
int Rows, Columns, NumInterp;
lerr_t Error;

/* abortion limits */
lerr_t MaxError = 1.0e30;
lerr_t MaxAvgError = 1.0e30;
int MaxInterp = INT_MAX;

/* for subsets of the pixels */
int NumSubsets;
#define MAX_NS 10
int CurrentSet;
lerr_t *GErrors[MAX_NS];
int *GDropped[MAX_NS];
int *Dropped;
int *Grid[MAX_NS];
int GSize[MAX_NS];
struct {
  enum {
    REGULAR,
    POISSON,
    RANDOM
  } type;
  /* regular: */
  int dx, dy;
  /* poison: */
  float percent;
} GType[MAX_NS];
bool_t GLinear[MAX_NS];

#define POSITION(p1,p0) Positions[(p1) * Columns + (p0)]
unsigned char *Left, *Right, *Interp, *Buffer, *LeftBasis, *RightBasis;
bool_t NoDepended;

struct PosList
{
  int num;
  int indices[1][2]; /* more entries follow */
} *recalc, *affected;
#define ADD_INDICES(list, i, e) (list)->indices[(list)->num][0] = i, (list)->indices[(list)->num++][1] = e
static void AddIndicesSorted(struct PosList *list, int i, int e)
{
  int j;
  for(j = list->num - 1; j >= 0; j--) {
    if(list->indices[j][0] > i ||
       (list->indices[j][0] == i && list->indices[j][1] > e)) {
      list->indices[j + 1][0] = list->indices[j][0];
      list->indices[j + 1][1] = list->indices[j][1];
    } else if(list->indices[j][0] < i ||
	      (list->indices[j][0] == i && list->indices[j][1] < e)) {
      break;
    } else {
      /* undo copying because entry already there */
      for(j++; j < list->num; j++) {
	list->indices[j][0] = list->indices[j + 1][0];
	list->indices[j][1] = list->indices[j + 1][1];
      }
      return;
    }
  }
  list->indices[j + 1][0] = i;
  list->indices[j + 1][1] = e;
  list->num++;
}


/*
** GLUT window handling
*/
int MW;         /* main window */
int BW;         /* browser window; non-zero if in use */
int P[2];       /* the current parameters */
int OP[2];      /* before clamping */
struct sources PSources; /* the sources and direction for these parameters */
bool_t PVert, PVertOnly, PHorizOnly;
static void InitBrowsing(void);
bool_t MWMouseDown; /* mouse button is pressed in main window */


/* Local helper functions */
static void Usage(char *progName);
static err_t diffMemSQ(unsigned char *a, unsigned char *b, int size);
static err_t diffMemABS(unsigned char *a, unsigned char *b, int size);
static bool_t DoInteraction(char *action, int i, int e, int level, bool_t display_only);
static void SaveState(void);
static void TimeStamp(char *fmt, int arg);
static void InitArray(void);
static void InitOpacity(void);
static bool_t AllocXY(void);
static void PrintAnalysis(void);
static void PrintInterpolation(FILE *res);
static void InterpolatePictures(unsigned char *target,
				unsigned char *source1,
				unsigned char *source2,
				double factor);
enum
{
  LEVEL_NONE,
  LEVEL_INIT_INTERP,
  LEVEL_IMAGES,
  LEVEL_REINTERP,
  LEVEL_REMOVE,
  LEVEL_DONE,
  LEVEL_MAX
};


/* calculate all possible differences */
static void optError(void);

/* options for opError() */
LFOps *ops;
LFInternOp op;
FILE *res;
char *destname;
int Minimum[LF_MAXPARAMS];
int Maximum[LF_MAXPARAMS];
int values[LF_MAXPARAMS];
int *ExactValues[LF_MAXPARAMS];
int Dimensions; /* number of ranges with more than one entry */
int width = -1, height = -1, Pixels;
int image_width, image_height;
int ImageSize;
int Channels = 1 | 2 | 4;
int NumChannels;
int win_width, win_height;
bool_t Swapped; /* Rows and Columns have been exchanged */
err_t(*diff_mem)(unsigned char *a, unsigned char *b, int size);


int Interactive;
bool_t Linear, VertOnly, NoWeights, Clamp, FInterp, NoDisplay, NotCentered;
int ClampCount;
#define FP_BITS 6
int Bits = FP_BITS;
int Quality = -1;
int Threshold = INT_MAX;
bool_t UseOpac;

bool_t Analysis, Histogram, Basis, OpacityOnly;
bool_t NoPivot;
double *AnaChannelsX, *AnaChannelsY;
int *MaxXChange, *MaxYChange;
int *MaxXChangeIndex, *MaxYChangeIndex;
float *Intensity;
float *Opacity, *OpacityPos, *OpacityPosMax, *OpacitySteep, *OpacitySteepMax;
unsigned char *Channel;
int MaxPixel;

int ForceIndices[100];
int NumForce, StartForce;
bool_t ForceImm;

int Dummy;
bool_t Smooth;
typedef unsigned char (*DummyFunc_t)(int p0, int p1);
DummyFunc_t DummyFuncs[10];
char *DummyFuncNames[10];
int NumDummyFuncs;
unsigned char DummyFuncSinExp(int p0, int p1)
{
  return (cos((1 - exp(p0 / 100.0)) * PI) + 1) * 127;
}
char DummyFuncSinExpName[] = "(cos((1 - exp(x / 100)) * pi) + 1) * 127";
unsigned char DummyFuncSinExpMirror(int p0, int p1)
{
  return 254 - DummyFuncSinExp(p0, p1);
}
char DummyFuncSinExpMirrorName[] = "254 - (cos((1 - exp(x / 100)) * pi) + 1) * 127";
unsigned char DummyFuncLinear(int p0, int p1)
{
  return p0 * 254 / 200;
}
char DummyFuncLinearName[] = "x * 254 / 200";
unsigned char DummyFuncSin(int p0, int p1)
{
  return (sin(p0 / 100.0 * PI) + 1) * 127;
}
char DummyFuncSinName[] = "(sin(x / 100 * pi) + 1) * 127";

/* print to result file and to stdout */
static void
logging(char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    fprintf(res, "# ");
    vfprintf(res, fmt, args);
    if(lfVerbose) {
      vprintf(fmt, args);
    }
    va_end(args);
}

static void DivideArray(float *diffs, int minv, int maxv, int *values, int num_values)
{
  int j, step;
  double sum, threshold;

  if(num_values <= 1) {
    *values = (maxv + minv) / 2;
    return;
  }

  for(sum = 0, j = 0;
      j + minv <= maxv;
      j++) {
    sum += diffs[j];
  }
  *values++ = minv;
  num_values--;
  threshold = sum / num_values;
  sum = 0;
  j = 0;
  step = 1;
  while(num_values > 0) {
    while(sum < threshold * step && j + minv < maxv) {
      sum += diffs[j];
      j++;
    }
    *values++ = j + minv;
    num_values--;
    step++;
  }
}

static int DivideRange(float *diffs, int minv, int maxv,
		       int left, int right)
{
  float sum, halve;
  int middle;

  if(left == right) {
    lfError("cannot divide range %d-%d!\n", left, right);
    return(left);
  }

  for(middle = left, sum = 0;
      middle <= right;
      middle++) {
    sum += diffs[middle - minv];
  }
  halve = sum / 2;
  for(middle = left, sum =0;
      middle < right && sum < halve;
      middle++) {
    sum += diffs[middle - minv];
  }
  return middle == right ? middle - 1 : middle;
}

extern bool_t __LFRowMajor; /* xxx hack to get displayable images from liblight.a */

int 
main(int argc, char *argv[])
{
  char *basename;
  char *progname;
  char *suffixPtr;
  char **argv_orig = argv;
  int argc_orig = argc;
  float read_descr[] = {
    LF_IGNORE,
    LF_IGNORE,
    LF_IGNORE,
    LF_NULL,
  };
  int i,e,j;
  LFParams *params;
  LFRenderer *renderer;
  char fnm[256];
  double absv, absh, absb;
  int iabsv, iabsh, iabsb;
  double avgh, avgv, avgb;
  double devh, devv, devb;
  double maxh, maxv, maxb;

  /* Parse the arguments */
  diff_mem = diffMemSQ;
  progname = argv[0];
  argv++;      
  argc--;

  GType[0].type = REGULAR;
  GType[0].dx = 1;
  GType[0].dy = 1;
  NumSubsets = 1;

  for(i = 0; i < LF_MAXPARAMS; i++) {
    Minimum[i] = INT_MIN;
    Maximum[i] = INT_MIN;
    values[i] = 0;
  }

  while (argc > 2) {
    /* -v: verbose */
    if (!strcmp(argv[0], "-v")) {
      lfVerbose = 1;
      argv++;
      argc--;
    } else if (!strcmp(argv[0], "-interactive")) {
      if(!isdigit(argv[1][0])) {
	Interactive = LEVEL_NONE + 1;
	argv++;
	argc--;
      } else {
	Interactive = atoi(argv[1]);
	argv += 2;
	argc -= 2;
      }
    } else if (!strcmp(argv[0], "-no_display")) {
      NoDisplay = TRUE;
      argv++;
      argc--;
    }
    /* picture comparison function */
    else if (!strcmp(argv[0], "-abs")) {
      diff_mem = diffMemABS;
      argv++;
      argc--;
    } else if(!strcmp(argv[0], "-linear")) {
      Linear = TRUE;
      argv++;
      argc--;
    } else if(!strcmp(argv[0], "-vertonly")) {
      VertOnly = TRUE;
      argv++;
      argc--;
    } else if(!strcmp(argv[0], "-noweights")) {
      NoWeights = TRUE;
      argv++;
      argc--;
    } else if(!strcmp(argv[0], "-clamp")) {
      Clamp = TRUE;
      argv++;
      argc--;
    } else if(!strcmp(argv[0], "-bits")) {
      Bits = atoi(argv[1]);
      argv += 2;
      argc -= 2;
    } else if(!strcmp(argv[0], "-force")) {
      argv++;
      argc--;
      NumForce = 0;
      while(argc && isdigit(argv[0][0])) {
	ForceIndices[NumForce++] = atoi(argv[0]);
	argv++;
	argc--;
      }
    } else if(!strcmp(argv[0], "-fimm")) {
      ForceImm = TRUE;
      argv++;
      argc--;
    } else if (!strcmp(argv[0], "-nodep")) {
      NoDepended = TRUE;
      argv++;
      argc--;
    }
    /* abortion limits */
    else if (!strcmp(argv[0], "-maxerror")) {
      MaxError = atof(argv[1]);
      argv += 2;
      argc -= 2;
    } else if (!strcmp(argv[0], "-maxavgerror")) {
      MaxAvgError = atof(argv[1]);
      argv += 2;
      argc -= 2;
    } else if (!strcmp(argv[0], "-maxinterp")) {
      MaxInterp = atoi(argv[1]);
      argv += 2;
      argc -= 2;
    } 
    /* analysis to be done */
    else if (!strcmp(argv[0], "-analysis")) {
      Analysis = TRUE;
      argv++;
      argc--;
      if(isdigit(argv[0][0])) {
	MaxPixel = atoi(argv[0]);
	argv++;
	argc--;
      } else {
	MaxPixel = -1;
      }
    } else if(!strcmp(argv[0], "-opacityonly")) {
      OpacityOnly = TRUE;
      argv++;
      argc--;
    } else if(!strcmp(argv[0], "-histogram")) {
      Histogram = TRUE;
      argv++;
      argc--;
    } else if(!strcmp(argv[0], "-basis")) {
      Basis = TRUE;
      argv++;
      argc--;
    }
    /* number and range of pictures being compared */
    else if (!strcmp(argv[0], "-range")) {
      i = atoi(argv[1]);
      if(i < 0 || i > LF_MAXPARAMS) {
	fprintf(stderr, "invalid parameter number %d\n", i);
	exit(1);
      }
      Minimum[i] = atoi(argv[2]);
      Maximum[i] = atoi(argv[3]);
      values[i]  = atoi(argv[4]) + 1;
      argv += 5;
      argc -= 5;
    } else if (!strcmp(argv[0], "-steps")) {
      i = atoi(argv[1]);
      if(i < 0 || i > LF_MAXPARAMS) {
	fprintf(stderr, "invalid parameter number %d\n", i);
	exit(1);
      }
      values[i] = atoi(argv[2]) + 1;
      argv += 3;
      argc -= 3;
    } else if(!strcmp(argv[0], "-threshold")) {
      Threshold = atof(argv[1]) * ERR_FACTOR;
      argv += 2;
      argc -= 2;
    } else if(!strcmp(argv[0], "-useopac")) {
      UseOpac = TRUE;
      argv++;
      argc--;
    }
    /* source description */
    else if (!strcmp(argv[0], "-size")) {
      width  = atoi(argv[1]);
      height = atoi(argv[2]);
      argv += 3;
      argc -= 3;
    } else if(!strcmp(argv[0], "-nochannel")) {
      Channels &= ~atoi(argv[1]);
      argv += 2;
      argc -= 2;
    } else if(!strcmp(argv[0], "-resetgrids")) {
      NumSubsets = 0;
      argv++;
      argc--;
    } else if(!strcmp(argv[0], "-reggrid")) {
      GType[NumSubsets].type = REGULAR;
      GType[NumSubsets].dx = atoi(argv[1]);
      GType[NumSubsets].dy = atoi(argv[2]);
      argv += 3;
      argc -= 3;
      if(!strcmp(argv[0], "linear")) {
	GLinear[NumSubsets] = TRUE;
	argv++;
	argc--;
      } else {
	GLinear[NumSubsets] = FALSE;
      }
      NumSubsets++;
    } else if(!strcmp(argv[0], "-poisson")) {
      GType[NumSubsets].type = POISSON;
      GType[NumSubsets].percent = atof(argv[1]);
      argv += 2;
      argc -= 2;
      if(!strcmp(argv[0], "linear")) {
	GLinear[NumSubsets] = TRUE;
	argv++;
	argc--;
      } else {
	GLinear[NumSubsets] = FALSE;
      }
      NumSubsets++;
    } else if(!strcmp(argv[0], "-random")) {
      GType[NumSubsets].type = RANDOM;
      GType[NumSubsets].percent = atof(argv[1]);
      argv += 2;
      argc -= 2;
      if(!strcmp(argv[0], "linear")) {
	GLinear[NumSubsets] = TRUE;
	argv++;
	argc--;
      } else {
	GLinear[NumSubsets] = FALSE;
      }
      NumSubsets++;
    } else if (!strcmp(argv[0], "-create_images")) {
      read_descr[0] = LF_READ_CREATE_IMAGES;
      argv++;
      argc--;
    } else if (!strcmp(argv[0], "-no_images")) {
      read_descr[1] = LF_READ_NO_IMAGES;
      argv++;
      argc--;
    } else if (!strcmp(argv[0], "-no_renderer")) {
      read_descr[2] = LF_READ_NO_RENDERER;
      argv++;
      argc--;
    } else if (!strcmp(argv[0], "-dummy")) {
      
      if(!isdigit(argv[1][0])) {
	Dummy = 1;
	argv++;
	argc--;
      } else {
	Dummy = atoi(argv[1]);
	if(Dummy >= 100) {
	  Dummy -= 100;
	  Smooth = TRUE;
	}
	if(Dummy == 3 && !NumForce) {
	  NumForce = 1;
	  ForceIndices[0] = 1;
	}
	argv += 2;
	argc -= 2;
      }
      NoPivot = TRUE;
    } else if(!strcmp(argv[0], "-quality")) {
      Quality = atoi(argv[1]);
      argv += 2;
      argc -= 2;
    } else if(!strcmp(argv[0], "-not_centered")) {
      NotCentered = TRUE;
      argv++;
      argc--;
    }
    /* default */
    else {
      Usage(progname);
    }
  }
  /* Now only the two file names should be left */
  if( argc != 2) {
    Usage(progname);
    exit(-1);
  }

  /* Ok, now do it. */
  basename = argv[0];
  destname = argv[1];
  
  lfNoITiming = TRUE; /* would spoil profiling and is useless anyway */
  lfInitContext();
  params = __LF_GET_PARAMS();
  renderer = __LF_GET_RENDERER();
  
  /* SuffixPtr points to last 4 characters of the file name */
  suffixPtr = &basename[(strlen(basename) > 4) ? (strlen(basename)-4) : 0];
  /* Read in the lightfield */
  lfBegin(LF_FIELD, NULL);
  ops = &__LF_GET_CONTEXT()->ops;
  ops->chain_msk = LF_SET_SLAB;
  op.op_descr = read_descr;
  op.op_extra = basename;
  if (!strcmp(suffixPtr, ".lif") || !strcmp(suffixPtr, ".LIF")) {
    /* Read in lightfield format */
    printf("Analyzing lightfield %s, writing result to %s.\n",
	   basename, destname);
    op.op_aux = lfBeginReadLif(ops, &op);
  } else if (!strcmp(suffixPtr, ".lid") || !strcmp(suffixPtr, ".LID")) {
    /* Read in .lid format. */
    printf("Analyzing lightfield description %s, writing results to %s.\n", 
	   basename, destname);
    op.op_aux = lfBeginReadSliceFile(ops, &op);
  } else {
    fprintf(stderr, "Error: input file %s is not a .lif or .lid file.\n",
	    basename);
    Usage(progname);
    exit(-1);
  }

  /* override sizes or use the one from the .lid file? */
  if(width >= 0) {
    ops->op_slabs[0]->shared->ns = width;
  } else {
    width = ops->op_slabs[0]->shared->ns;
  }
  if(height >= 0) {
    ops->op_slabs[0]->shared->nt = height;
  } else {
    height = ops->op_slabs[0]->shared->nt;
  }
  __LFRowMajor = TRUE;
  for(i = 0; i < LF_MAXPARAMS; i++) {
    if(Minimum[i] == INT_MIN) {
      Minimum[i] = params->params[i].min;
    }
    if(Maximum[i] == INT_MIN) {
      Maximum[i] = params->params[i].max;
    }
    if(values[i] == 0) {
      values[i] = Maximum[i] - Minimum[i] + 1;
    }
  }

  /*
  ** init renderer
  */
  strcpy(fnm, basename);
  renderer->lidDir = dirname(fnm);
  renderer->flags = (read_descr[2] == LF_READ_NO_RENDERER) ? LFRF_NO_RENDERING : 0;
  if(Quality != -1) {
    renderer->quality = Quality;
  }
  if(!lfRDAlloc(renderer)) {
    lfError("read .lif/.lid: cannot initialize renderer %d for \"%s\"\n",
	    renderer->type, renderer->config);
    exit(1);
  }

  /* dump histogram */
  if(Histogram) {
    printf("\nHistogram (before scaling):\n");
    for(i = 0; i < 256; i++) {
      printf("%d %d\n", i, volHistogram[i]);
    }
  }

  /* open result file and dump program arguments */
  if(!(res = fopen(destname, "w"))) {
    fprintf(stderr, "cannot create destfile %s\n", destname);
    exit(1);
  }
  fprintf(res, "#");
  for(i = 0; i < argc_orig; i++) {
    fprintf(res, " %s", argv_orig[i]);
  }
  fprintf(res, "\n\n");

  for(i = 0; i < LF_MAXPARAMS; i++) {
    if(values[i] <= 0) {
      values[i] = 1;
    } else if (values[i] > 1) {
      Dimensions++;
    }
  }

  /* Basis can only be given for one dimensional ranges. */
  if((Basis || Dummy) && (Dimensions > 1)) {
    lfError("cannot print basis functions for more than one dimension\n");
    exit(1);
  }

  if(Dimensions <= 1 && values[0] <= 1) {
    int tmp;
    /* draw column as row */
    tmp = values[1];
    values[1] = values[0];
    values[0] = tmp;
    tmp = Maximum[0];
    Maximum[0] = Maximum[1];
    Maximum[1] = tmp;
    tmp = Minimum[0];
    Minimum[0] = Minimum[1];
    Minimum[1] = tmp;
    Swapped = TRUE;
  } else {
    Swapped = FALSE;
  }

  NumChannels = 3;
  for(i = 0; i < 3; i++) {
    if(!(Channels & (1<<i))) {
      NumChannels--;
    }
  }

  InitOpacity();
  InitArray();
  if(!Interactive && read_descr[2] != LF_READ_NO_RENDERER) {
    lfRDDone(renderer);
  }
  ImageSize -= (3 - NumChannels) * Pixels;
  if(ImageSize <= 0) {
    lfError("no channels left\n");
    exit(1);
  }
  fprintf(res, "# %d rows: ", Rows);
  for(i = 0; i < Rows; i++) {
    fprintf(res, "%d ", POSITION(i,0).p[1]);
  }
  fprintf(res, "\n# %d columns: ", Columns);
  for(i = 0; i < Columns; i++) {
    fprintf(res, "%d ", POSITION(0,i).p[0]);
  }
  fprintf(res, "\n");
  absv = 0;
  absh = 0;
  absb = 0;
  iabsv = 0;
  iabsh = 0;
  iabsb = 0;
  devh = 0;
  devv = 0;
  devb = 0;
  maxh = 0;
  maxv = 0;
  maxb = 0;
  for(i = 0; i < Rows; i++) {
    for(e = 0; e < Columns; e++) {
      double diffv = 0, diffh = 0;
      int contrib = 0;

      if(i + 1 < Rows) {
	diffv = diff_mem(POSITION(i,e).image,POSITION(i+1,e).image, ImageSize);
	absv += diffv;
	iabsv++;
	contrib++;
	if(diffv > maxv) {
	  maxv = diffv;
	}
      }
      if(e + 1 < Columns) {
	diffh = diff_mem(POSITION(i,e).image,POSITION(i,e+1).image, ImageSize);
	absh += diffh;
	iabsh++;
	contrib++;
	if(diffh > maxh) {
	  maxh = diffh;
	}
      }
      if(contrib >= 2) {
	double diffb = (diffv + diffh) / contrib;
	absb += diffb;
	iabsb++;
	if(diffb > maxb) {
	  maxb = diffb;
	}
      }
    }
  }
  avgh = absv / iabsv;
  avgv = absh / iabsh;
  avgb = absb / iabsb;
  for(i = 0; i < Rows; i++) {
    for(e = 0; e < Columns; e++) {
      double diffv = 0, diffh = 0, ldevv, ldevh;
      int contrib = 0;

      if(e == 0) {
	fprintf(res, "#vdiffs \n#hdiffs \n#bdiffs \n");
      }
      if(i + 1 < Rows) {
	diffv = diff_mem(POSITION(i,e).image,POSITION(i+1,e).image, ImageSize);
	ldevv = diffv - avgv;
	fprintf(res, "#vdiffs %d %d %e %e\n",
		POSITION(i,0).p[1], POSITION(0,e).p[0],
		(double)diffv, (double)ldevv);
	contrib++;
	devv += ldevv*ldevv;
      }
      if(e + 1 < Columns) {
	diffh = diff_mem(POSITION(i,e).image,POSITION(i,e+1).image, ImageSize);
	ldevh = diffh - avgh;
	fprintf(res, "#hdiffs %d %d %e %e\n",
		POSITION(i,0).p[1], POSITION(0,e).p[0],
		(double)diffh, (double)ldevh);
	contrib++;
	devh += ldevh*ldevh;
      }
      if(contrib >= 2) {
	double diffb = (diffv + diffh) / contrib;
	double ldevb = diffb - avgb;
	devb += ldevb*ldevb;
	fprintf(res, "#bdiffs %d %d %e %e\n",
		POSITION(i,0).p[1], POSITION(0,e).p[0],
		(double)diffb, (double)ldevb);
      }
    }
  }
  fprintf(res, "#vdiffsstat %d %d %d %.2f %.2f %.2f %.2f\n",
	  Rows * Columns, Columns - 1, Rows - 1, absv, avgv, sqrt(devv/iabsv), maxv);
  fprintf(res, "#hdiffsstat %d %d %d %.2f %.2f %.2f %.2f\n",
	  Rows * Columns, Columns - 1, Rows - 1, absh, avgh, sqrt(devh/iabsh), maxh);
  fprintf(res, "#bdiffsstat %d %d %d %.2f %.2f %.2f %.2f\n",
	  Rows * Columns, Columns - 1, Rows - 1, absb, avgb, sqrt(devb/iabsb), maxb);
  fprintf(res, "\n\n");

  /* now do analysis */
  for(CurrentSet = 0; CurrentSet < NumSubsets; CurrentSet++) {
    bool_t old_linear = Linear;
    Linear = Linear || GLinear[CurrentSet];

    logging("set %d %s, size %d/%d\n",
	    CurrentSet,
	    Linear || GLinear[CurrentSet] ? "linear" : "polynomial",
	    GSize[CurrentSet], ImageSize);
    switch(GType[CurrentSet].type) {
    case REGULAR:
      logging("regular %d/%d\n", GType[CurrentSet].dx, GType[CurrentSet].dy);
      break;
    case RANDOM:
      logging("random %f\n", GType[CurrentSet].percent);
      break;
    case POISSON:
      logging("poisson %f\n", GType[CurrentSet].percent);
      break;
    default:
      fprintf(stderr, "unknown grid type\n");
      exit(10);
    }
    Dropped = GDropped[CurrentSet];
    for(i = 0; i < Rows; i++) {
      TNeigh[i] = i;
      BNeigh[i] = i;
    }
    for(e = 0; e < Columns; e++) {
      RNeigh[e] = e;
      LNeigh[e] = e;
    }

    for(i = 0; i < Rows; i++) {
      for(e = 0; e < Columns; e++) {
	struct Position *position = &POSITION(i,e);
	double **cf;

	for(j = 0, cf = position->cx; j < Columns + 2; j++) {
	  free(*cf);
	  *cf++ = NULL;
	}
	for(j = 0, cf = position->cy; j < Rows + 2; j++) {
	  free(*cf);
	  *cf++ = NULL;
	}
	position->sampled = TRUE;
	position->left = e - 1;
	position->right = e + 1;
	position->bottom = i - 1;
	position->top = i + 1;
	memset(position->contribs, 0, Rows * Columns * sizeof(err_t));
      }
    }


    if(Interactive && !NoDisplay) {
      win_width = Columns * width;
      win_height = Rows * height;
      glutInitWindowSize(win_width, win_height);
      glutInitDisplayMode(0);
      MW = glutCreateWindow("panalysis");
      glutDisplayFunc(optError);
      /* unfortunately this call never returns */
      glutMainLoop();
    } else {
      optError();
    }

    Linear = old_linear;
  }

  /* print result for all grids */
  fprintf(res, "\n\n\n# dropped positions and errors for all grids\n");
  for(CurrentSet = 0; CurrentSet < NumSubsets; CurrentSet++) {
    int entry;

    fprintf(res, "#dropped set %d %s, %d/%d: ",
	    CurrentSet, 
	    Linear || GLinear[CurrentSet] ? "linear" : "polynomial",
	    GSize[CurrentSet], ImageSize);
    switch(GType[CurrentSet].type) {
    case REGULAR:
      fprintf(res, "regular %d/%d\n", GType[CurrentSet].dx, GType[CurrentSet].dy);
      break;
    case RANDOM:
      fprintf(res, "random %f\n", GType[CurrentSet].percent);
      break;
    case POISSON:
      fprintf(res, "poisson %f\n", GType[CurrentSet].percent);
      break;
    default:
      fprintf(stderr, "unknown grid type\n");
      exit(10);
    }
    memset(Left, 0, ImageSize);
    for(i = 0; i < GSize[CurrentSet]; i++) {
      Left[Grid[CurrentSet][i]] = 0xFF;
    }
    for(i = 0, entry = 0; i < ImageSize; i++) {
      if(!(entry % 20)) {
	fprintf(res, "#grid%02d ", CurrentSet);
      }
      fprintf(res, "%d ", Left[i]);
      if(++entry >= 20) {
	entry = 0;
      }
    }
    fprintf(res, "\n");
  }
  for(i = 0; i < Rows * Columns; i++) {
    fprintf(res, "#position %d ", i);
    for(CurrentSet = 0; CurrentSet < NumSubsets; CurrentSet++) {
      fprintf(res, "%d %e %e ",
	      GDropped[CurrentSet][i],
	      GErrors[CurrentSet][i],
	      GErrors[CurrentSet][i] / (i + 1));
    }
    fprintf(res, "\n");
  }

  PrintInterpolation(res);

  return 0;
}

static void InitOpacity(void)
{
  int i,e,j;

  /* dump average opacities */
  Columns = (Maximum[0] - Minimum[0] + 1);
  Rows = (Maximum[1] - Minimum[1] + 1);
  if(!(OpacitySteep = malloc(2 * Columns * sizeof(OpacitySteep[0]))) ||
     !(ExactValues[0] = malloc(values[0] * sizeof(ExactValues[0][0]))) ||
     !(ExactValues[1] = malloc(values[1] * sizeof(ExactValues[1][0]))) ||
     !(OpacityPos = malloc(2 * Rows * sizeof(OpacityPos[0]))) ||
     !(Opacity = malloc(sizeof(float) * Rows * Columns))) {
    lfError("no mem!\n");
    exit(10);
  }
  OpacityPosMax = &OpacityPos[Rows];
  OpacitySteepMax = &OpacitySteep[Columns];
  for(i = Minimum[1]; i <= Maximum[1]; i++) {
    for(e = Minimum[0]; e <= Maximum[0]; e++) {
      lfVolSetRampVals(i,e);
      Opacity[(i - Minimum[1]) * Columns + e - Minimum[0]] = lfVolCalcOpacity(volHistogram, density_ramp, DENSITY_MAX + 1);
    }
  }
  OpacityPos[0] =
    OpacityPosMax[0] = 0;
  for(i = 1; i < Rows; i++) {
    double avg_diff = 0;
    double max_diff = 0, diff;
    for(e = 0; e < Columns; e++) {
      diff = fabs(Opacity[i * Columns + e] - Opacity[(i - 1) * Columns + e]);
      avg_diff += diff;
      if(diff > max_diff) {
	max_diff = diff;
      }
    }
    OpacityPos[i] = avg_diff / Columns;
    OpacityPosMax[i] = max_diff;
    fprintf(res, "#opacpos %d %e %e\n", i + Minimum[1], OpacityPos[i], max_diff);
  }
  OpacitySteep[0] =
    OpacitySteepMax[0] = 0;
  for(e = 1; e < Columns; e++) {
    double avg_diff = 0;
    double max_diff = 0, diff;
    for(i = 0; i < Rows; i++) {
      diff = fabs(Opacity[i * Columns + e] - Opacity[i * Columns + (e - 1)]);
      avg_diff += diff;
      if(diff > max_diff) {
	max_diff = diff;
      }
    }
    OpacitySteep[e] = avg_diff / Rows;
    OpacitySteepMax[e] = max_diff;
    fprintf(res, "#opacsteep %d %e %e\n", e + Minimum[0], OpacitySteep[e], max_diff);
  }
  DivideArray(OpacityPos, Minimum[1], Maximum[1], ExactValues[1], values[1]);
  for(j = 0; j < values[1]; j++) {
    fprintf(res, "#opacposavgvals %d %e\n",
	    ExactValues[1][j], OpacityPos[ExactValues[1][j] - Minimum[1]]);
  }
  DivideArray(OpacityPosMax, Minimum[1], Maximum[1], ExactValues[1], values[1]);
  for(j = 0; j < values[1]; j++) {
    fprintf(res, "#opacposmaxvals %d %e\n",
	    ExactValues[1][j], OpacityPosMax[ExactValues[1][j] - Minimum[1]]);
  }
  DivideArray(OpacitySteep, Minimum[0], Maximum[0], ExactValues[0], values[0]);
  for(j = 0; j < values[0]; j++) {
    fprintf(res, "#opacsteepavgvals %d %e\n",
	    ExactValues[0][j], OpacitySteep[ExactValues[0][j] - Minimum[0]]);
  }
  DivideArray(OpacitySteepMax, Minimum[0], Maximum[0], ExactValues[0], values[0]);
  for(j = 0; j < values[0]; j++) {
    fprintf(res, "#opacsteepmaxvals %d %e\n",
	    ExactValues[0][j], OpacitySteepMax[ExactValues[0][j] - Minimum[0]]);
  }

  if(OpacityOnly) {
    exit(0);
  }
}

/* remove entries that have no bit set */
static void FilterChannels(unsigned char *data, int size, int channels)
{
  unsigned char *source = data, *target = data;

  size /= 3;
  while(--size >= 0) {
    if(channels & 1) {
      *target++ = *source;
    }
    source++;
    if(channels & 2) {
      *target++ = *source;
    }
    source++;
    if(channels & 4) {
      *target++ = *source;
    }
    source++;
  }
}

static LFParamSamples p0, p1;
static void RenderImage(int p[2], unsigned char *image)
{
  p1.samples[0].p = p[1];
  p0.samples[0].p = p[0];
  
  op.op_func(ops->op_slabs, ops->op_slab_cnt, op.op_pos,
	     op.op_descr, NULL, 0,
	     image, ImageSize, op.op_aux);
}

static void InitEntry(int i, int e)
{
  struct Position *position = &POSITION(i, e);
  int j;

  lfOutput("%d/%d = %d/%d\n", e,i, position->p[0],position->p[1]);
  if(!(position->image = calloc(ImageSize, 1))) {
    fprintf(stderr, "no memory!\n");
    exit(1);
  }
  position->sampled = TRUE;
 
  if(Dummy) {
    for(j = 0; j < NumDummyFuncs; j++) {
      position->image[j] = DummyFuncs[j](position->p[0], position->p[1]);
    }
  } else {
    RenderImage(position->p, position->image);
    FilterChannels(position->image, ImageSize, Channels);
  }

  if(Analysis) {
    unsigned char *this, *other, diff;
    
    Intensity[(i * Columns + e) * 4 + 0] = lfCalcIntensity(position->image, width, height, 3);
    Intensity[(i * Columns + e) * 4 + 1] = lfCalcIntensityStride(&position->image[0], width, height, 3);
    Intensity[(i * Columns + e) * 4 + 2] = lfCalcIntensityStride(&position->image[1], width, height, 3);
    Intensity[(i * Columns + e) * 4 + 3] = lfCalcIntensityStride(&position->image[2], width, height, 3);
    if(Channel) {
      Channel[i * Columns + e] = position->image[MaxPixel];
    }
    
    /* add differences within line? */
    if(e > 0) {
      this = position->image;
      other = POSITION(i, e - 1).image;
      for(j = 0; j < ImageSize; j++) {
	diff = *other++ - *this++;
	AnaChannelsX[j] += abs(diff);
	if(diff > MaxXChange[j]) {
	  MaxXChange[j] = diff;
	  MaxXChangeIndex[j] = i * Columns + e;
	} 
      }
    }
    
    /* add difference within column? */
    if(i > 0) {
      this = position->image;
      other = POSITION(i - 1, e).image;
      for(j = 0; j < ImageSize; j++) {
	diff = *other++ - *this++;
	AnaChannelsY[j] += abs(diff);
	if(diff > MaxYChange[j]) {
	  MaxYChange[j] = diff;
	  MaxYChangeIndex[j] = i * Columns + e;
	}
      }
      
      /* delete previous row? */
      if(e == Columns - 1) {
	for(e = 0; e < Columns; e++) {
	  free(POSITION(i - 1, e).image);
	}
	e--;
      }
    }
  }
}

static void InsertRow(int i)
{
  Rows++;
  if(!(Positions = realloc(Positions, sizeof(Positions[0]) * Rows * Columns))) {
    fprintf(stderr, "no mem for new pos array!\n");
    exit(1);
  }
  memmove(&POSITION(i + 1, 0), &POSITION(i, 0), sizeof(Positions[0]) * Columns * (Rows - 1 - i));
  memset(&POSITION(i, 0), 0, sizeof(Positions[0]) * Columns);
  POSITION(i, 0).p[1] =
    UseOpac ? DivideRange(OpacityPosMax, Minimum[1], Maximum[1],
			  POSITION(i - 1, 0).p[1], POSITION(i + 1, 0).p[1]) :
    (POSITION(i + 1, 0).p[1] + POSITION(i - 1, 0).p[1]) / 2;
  lfOutput("new row #%d with parameter %d\n", i, POSITION(i, 0).p[1]);
}

static void InsertColumn(int e)
{
  int i;
  int left = e, right = Columns - e;
  struct Position *target, *source;

  if(!(Positions = realloc(Positions, sizeof(Positions[0]) * Rows * (Columns + 1)))) {
    fprintf(stderr, "no mem for new pos array!\n");
    exit(1);
  }

  source = &POSITION(Rows - 1, e);
  Columns++;
  target = &POSITION(Rows - 1, e + 1);
  for(i = Rows - 1; i >= 0; i--) {
    memmove(target, source, sizeof(Positions[0]) * right);
    memset(--target, 0, sizeof(*target));
    target -= left;
    source -= left;
    memmove(target, source, sizeof(Positions[0]) * left);
    target -= right;
    source -= right;
  }
  POSITION(0,e).p[0] =
    UseOpac ? DivideRange(OpacitySteepMax, Minimum[0], Maximum[0],
			  POSITION(0,e-1).p[0], POSITION(0,e+1).p[0]) :
    (POSITION(0,e+1).p[0] + POSITION(0,e-1).p[0]) / 2;
  lfOutput("new column #%d with parameter %d\n", e, POSITION(0,e).p[0]);
}

static void InitArray(void)
{
  int i,e;
  LFParams *params = __LF_GET_PARAMS();
  LFSlab *slab = ops->op_slabs[0];
  LFShared *sh = slab->shared;

  if(params->num > 2) {
    for(i = 2; i < LF_MAXPARAMS; i++) {
      if(values[i] > 1) {
	fprintf(stderr, "Cannot handle more than 2 parameters!\n");
	exit(1);
      }
    }
  }

  bzero(op.op_pos, sizeof(op.op_pos));
  if(NotCentered) {
    op.op_pos[LF_U_POS] = sh->nu / 2;
    op.op_pos[LF_V_POS] = sh->nv / 2;
  } else {
    /* center view */
    op.op_pos[LF_U_POS] = 0;
    op.op_pos[LF_V_POS] = 0;
    sh->nu = 1;
    sh->nv = 1;
  }
  sh->sample_size = 3;

  switch(Dummy) {
  case 0:
    NumDummyFuncs = 0;
    break;
  case 4:
    /* sin(exp(x)) and mirrored */
    DummyFuncs[0] = DummyFuncSinExp;
    DummyFuncNames[0] = DummyFuncSinExpName;
    DummyFuncs[1] = DummyFuncSinExpMirror;
    DummyFuncNames[1] = DummyFuncSinExpMirrorName;
    NumDummyFuncs = 2;
    break;
  case 3:
    /* sin(x) */
    DummyFuncs[0] = DummyFuncSin;
    DummyFuncNames[0] = DummyFuncSinName;
    NumDummyFuncs = 1;
    break;
  case 2:
    /* sin(exp(x)) and linear */
    DummyFuncs[0] = DummyFuncSinExp;
    DummyFuncNames[0] = DummyFuncSinExpName;
    DummyFuncs[1] = DummyFuncLinear;
    DummyFuncNames[1] = DummyFuncLinearName;
    NumDummyFuncs = 2;
    break;
  default:
    /* sin(exp(x)) */
    DummyFuncs[0] = DummyFuncSinExp;
    DummyFuncNames[0] = DummyFuncSinExpName;
    NumDummyFuncs = 1;
    break;
  }

  /* allocate original images */
  Pixels = width * height;
  ImageSize = !Dummy ? Pixels * sizeof(unsigned char) * 3 : NumDummyFuncs;
  Columns = values[0];
  Rows    = values[1];
  
  if(!(Positions = calloc(Rows * Columns, sizeof(struct Position))) ||
     !(Left = malloc(ImageSize * 6)) ||
     (Analysis &&
      ((MaxPixel != -1 && !(Channel = malloc(sizeof(unsigned char) * Rows * Columns))) ||
       (!(AnaChannelsX = calloc(ImageSize, sizeof(double)))) ||
       (!(AnaChannelsY = calloc(ImageSize, sizeof(double)))) ||
       (!(MaxXChange = calloc(ImageSize, sizeof(int)))) ||
       (!(MaxYChange = calloc(ImageSize, sizeof(int)))) ||
       (!(MaxXChangeIndex = calloc(ImageSize, sizeof(int)))) ||
       (!(MaxYChangeIndex = calloc(ImageSize, sizeof(int)))) ||
       (!(Intensity = malloc(sizeof(float) * 4 * Rows * Columns)))))) {
    fprintf(stderr, "no memory!\n");
    exit(1);
  }
  Right = &Left[ImageSize];
  Buffer = &Right[ImageSize];
  Interp = &Buffer[ImageSize];
  LeftBasis = &Interp[ImageSize];
  RightBasis = &LeftBasis[ImageSize];

  printf("********* reading original images *********\n");
  TimeStamp("image creation", 0);
  op.op_pos[LF_P_POS(1)] =
    op.op_pos[LF_P_POS(0)] = 0;
  p0.np = 1;
  p1.np = 1;
  if(Swapped) {
    p1.samples[0].psamples = &p0;
    p0.samples[1].psamples = NULL;
    sh->psamples = &p1;
  } else {
    p0.samples[0].psamples = &p1;
    p1.samples[1].psamples = NULL;
    sh->psamples = &p0;
  }

  /* initialize regular grid */
  /* parameter p1 */
  for(i = 0; i < Rows; i++) {
    POSITION(i,0).p[1] =
      UseOpac ? ExactValues[1][i] :
      Minimum[1] + (values[1] > 1 ?
		    (Maximum[1] - Minimum[1]) * i / (values[1] - 1) : 0);
  }
  /* parameter p0 */
  for(e = 0; e < Columns; e++) {
    POSITION(0,e).p[0] = 
      UseOpac ? ExactValues[0][e] :
      Minimum[0] + (values[0] > 1 ?
		    (Maximum[0] - Minimum[0]) * e / (values[0] - 1) : 0);
  }

  for(i = 0; i < Rows; i++) {
    for(e = 0; e < Columns; e++) {
      struct Position *position = &POSITION(i,e);
      bool_t restart = FALSE;

      position->i = i;
      position->e = e;
      position->p[1] = POSITION(i,0).p[1];
      position->p[0] = POSITION(0,e).p[0];

      if(!position->image) {
	InitEntry(i,e);
      }

      /* enlarge array? */
      if(!Analysis && i > 0) {
	err_t diff = diff_mem(position->image, POSITION(i-1,e).image, ImageSize);
	lfOutput("vert. diff %d/%d to %d/%d: %f\n", i,e, i-1,e, diff/ERR_FACTOR);
	if(diff > Threshold && (POSITION(i,0).p[1] - POSITION(i-1,e).p[1]) > 1) {
	  /* new row and fill it */
	  InsertRow(i);
	  i++; /* for check within row */
	  restart = TRUE;
	}
      }
      /* position pointer may be invalid now! */
      if(!Analysis && e > 0) {
	err_t diff = diff_mem(POSITION(i,e-1).image, POSITION(i,e).image, ImageSize);
	lfOutput("horiz. diff %d/%d to %d/%d: %f\n", i,e, i,e-1, diff/ERR_FACTOR);
	if(diff > Threshold && (POSITION(0,e).p[0] - POSITION(0,e - 1).p[0]) > 1) {
	  /* new column and fill it */
	  InsertColumn(e);
	  restart = TRUE;
	}
      }
      if(restart) {
	e = Columns;
	i = -1;
      }
    }
  }

  TimeStamp("got images", 0);

  if(Analysis) {
    PrintAnalysis();
    exit(0);
  }

  /* alloc remaining memory */
  for(i = 0; i < Rows; i++) {
    for(e = 0; e < Columns; e++) {
      struct Position *position = &POSITION(i,e);

      if(!Analysis && 
	 (!(position->contribs = malloc(Rows * Columns * sizeof(err_t))) ||
#ifdef CACHED
	  !(position->vert_buffer = calloc(Rows * Columns, sizeof(position->vert_buffer[0]))) ||
#endif
	  !(position->cx = calloc(Columns + 2, sizeof(double *))) ||
	  !(position->cy = calloc(Rows + 2, sizeof(double *))))) {
	fprintf(stderr, "no mem!\n");
	exit(10);
      }
    }
  }



  if(!(LNeigh = malloc(2 * Columns * sizeof(LNeigh[0]))) ||
     !(TNeigh = malloc(2 * Rows * sizeof(TNeigh[0])))) {
    fprintf(stderr, "no mem!\n");
    exit(10);
  }
  RNeigh = &LNeigh[Columns];
  BNeigh = &TNeigh[Rows];
  AllocXY();

  /* dump histogram? */
  if(Histogram) {
    printf("\nHistogram (after scaleing):\n");
    for(i = 0; i < 256; i++) {
      printf("%d %d\n", i, volHistogram[i]);
    }
  }

  /* now allocate remaining structures */
  for(i = 0; i < NumSubsets; i++) {
    switch(GType[i].type) {
    case REGULAR:
      GSize[i] = Dummy ? NumDummyFuncs : (width / GType[i].dx) * (height / GType[i].dy) * 3;
      break;
    case RANDOM:
    case POISSON:
      GSize[i] = width * height * 3 * GType[i].percent;
      break;
    default:
      fprintf(stderr, "unknown grid type\n");
      exit(10);
    }
    if(!(GErrors[i] = malloc(Rows * Columns * sizeof(GErrors[i][0]))) ||
       !(GDropped[i] = malloc(Rows * Columns * sizeof(GDropped[i][0]))) ||
       !(Grid[i] = malloc(GSize[i] * sizeof(Grid[i][0])))) {
      fprintf(stderr, "no mem\n");
      exit(10);
    }
    switch(GType[i].type) {
    case REGULAR:
      if(GSize[i]) {
	if(Dummy) {
	  int x;
	  for(x = 0; x < GSize[i]; x++) {
	    Grid[i][x] = x;
	  }
	} else {
	  int x,y;
	  int *grid = Grid[i];
	  for(y = 0; y < height; y += GType[i].dy) {
	    for(x = 0; x < width; x += GType[i].dx) {
	      *grid++ = (y * width + x) * NumChannels;
	      *grid++ = (y * width + x) * NumChannels + 1;
	      *grid++ = (y * width + x) * NumChannels + 2;
	    }
	  }
	  assert(grid - Grid[i] == GSize[i]);
	}
      }
      break;
    case RANDOM:
      RandomGrid(Grid[i], GSize[i], width, height, FALSE);
      break;
    case POISSON:
      RandomGrid(Grid[i], GSize[i], width, height, TRUE);
      break;
    default:
      fprintf(stderr, "unknown grid type\n");
      exit(10);
    }
  }
}

static void PrintAnalysis(void)
{
  double max_change = 0, max_red_change = 0, max_green_change = 0, max_blue_change = 0;
  int index = 0, index_red = 0, index_green = 0, index_blue = 0;
  double maxX_change = 0, maxX_red_change = 0, maxX_green_change = 0, maxX_blue_change = 0;
  int indexX = 0, indexX_red = 0, indexX_green = 0, indexX_blue = 0; 
  double maxY_change = 0, maxY_red_change = 0, maxY_green_change = 0, maxY_blue_change = 0;
  int indexY = 0, indexY_red = 0, indexY_green = 0, indexY_blue = 0; 
  int max_absX_change = 0, max_absY_change = 0;
  int absX_index = 0, absY_index = 0;
  struct Position *position;
  int i,e,j;

  /* maximum changing channel? */
  fprintf(res,
	  "# changes in the channels for %d*%d*%d diffs\n"
	  "# index, row, column | x/y/both red, green, blue, avg. change | maximum x change red/green/blue, maximum y change red,green,blue, x/y change index red/green/blue\n",
	  (Rows > 1 ? Rows - 1 : 1), (Columns > 1 ? Columns - 1 : 1), ImageSize);
  for(j = 0; j < ImageSize; j += 3) {
    double avg_X = (AnaChannelsX[j] + AnaChannelsX[j + 1] + AnaChannelsX[j + 2])/3;
    double avg_Y = (AnaChannelsY[j] + AnaChannelsY[j + 1] + AnaChannelsY[j + 2])/3;
    double avg = (avg_X + avg_Y) / 2;
    fprintf(res, "%d\t%d\t%d\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n",
	    j, j / width / 3, (j / 3) % width,
	    AnaChannelsX[j], AnaChannelsX[j + 1], AnaChannelsX[j + 2], avg_X,
	    AnaChannelsY[j], AnaChannelsY[j + 1], AnaChannelsY[j + 2], avg_Y,
	    (AnaChannelsX[j] + AnaChannelsY[j])/2, (AnaChannelsX[j + 1]+AnaChannelsY[j + 1])/2,
	    (AnaChannelsX[j + 2]+AnaChannelsY[j + 2])/2, avg,
	    MaxXChange[j], MaxXChange[j + 1], MaxXChange[j + 2],
	    MaxYChange[j], MaxYChange[j + 1], MaxYChange[j + 2],
	    MaxXChangeIndex[j], MaxXChangeIndex[j + 1], MaxXChangeIndex[j + 2],
	    MaxYChangeIndex[j], MaxYChangeIndex[j + 1], MaxYChangeIndex[j + 2]);
    
    if(avg_X > maxX_change) {
      maxX_change = avg;
      indexX = j;
    }
    if(AnaChannelsX[j] > maxX_red_change) {
      maxX_red_change = AnaChannelsX[j];
      indexX_red = j;
    }
    if(AnaChannelsX[j + 1] > maxX_green_change) {
      maxX_green_change = AnaChannelsX[j + 1];
      indexX_green = j;
    }
    if(AnaChannelsX[j + 2] > maxX_blue_change) {
      maxX_blue_change = AnaChannelsX[j + 2];
      indexX_blue = j;
    }
    
    if(avg_Y > maxY_change) {
      maxY_change = avg;
      indexY = j;
    }
    if(AnaChannelsY[j] > maxY_red_change) {
      maxY_red_change = AnaChannelsY[j];
      indexY_red = j;
    }
    if(AnaChannelsY[j + 1] > maxY_green_change) {
      maxY_green_change = AnaChannelsY[j + 1];
      indexY_green = j;
    }
    if(AnaChannelsY[j + 2] > maxY_blue_change) {
      maxY_blue_change = AnaChannelsY[j + 2];
      indexY_blue = j;
    }
    
    if(avg > max_change) {
      max_change = avg;
      index = j;
    }
    if((AnaChannelsX[j] + AnaChannelsX[j]) / 2 > max_red_change) {
      max_red_change = (AnaChannelsX[j] + AnaChannelsX[j]) / 2;
      index_red = j;
    }
    if((AnaChannelsX[j + 1] + AnaChannelsX[j + 1]) / 2 > max_green_change) {
      max_green_change = (AnaChannelsX[j + 1] + AnaChannelsX[j + 1]) / 2;
      index_green = j;
    }
    if((AnaChannelsX[j + 2] + AnaChannelsX[j + 2]) / 2 > max_blue_change) {
      max_blue_change = (AnaChannelsX[j + 2] + AnaChannelsX[j + 2]) / 2;
      index_blue = j;
    }
    
    for(i = 0; i < 3; i++) {
      if(MaxXChange[j + i] > max_absX_change) {
	max_absX_change = MaxXChange[j + i];
	absX_index = j + i;
      }
      if(MaxYChange[j + i] > max_absY_change) {
	max_absY_change = MaxYChange[j + i];
	absY_index = j + i;
      }
    }
  }
  fprintf(res,
	  "# XXXXX\n"
	  "# maximum changing average red: %d = Row %d, Column %d, change %e\n"
	  "# maximum changing average green: %d = Row %d, Column %d, change %e\n"
	  "# maximum changing average blue: %d = Row %d, Column %d, change %e\n"
	  "# maximum changing average overall: %d = Row %d, Column %d, change %e\n"
	  "# YYYYYY\n"
	  "# maximum changing average red: %d = Row %d, Column %d, change %e\n"
	  "# maximum changing average green: %d = Row %d, Column %d, change %e\n"
	  "# maximum changing average blue: %d = Row %d, Column %d, change %e\n"
	  "# maximum changing average overall: %d = Row %d, Column %d, change %e\n"
	  "# both\n"
	  "# maximum changing average red: %d = Row %d, Column %d, change %e\n"
	  "# maximum changing average green: %d = Row %d, Column %d, change %e\n"
	  "# maximum changing average blue: %d = Row %d, Column %d, change %e\n"
	  "# maximum changing average overall: %d = Row %d, Column %d, change %e\n"
	  "# maximum absolute X changing channel: %d = Row %d, Column %d, color %d, change %d in row/column %d/%d\n"
	  "# maximum absolute Y changing channel: %d = Row %d, Column %d, color %d, change %d in row/column %d/%d\n",
	  indexX_red, indexX_red / width / 3, (indexX_red / 3) % width, maxX_red_change,
	  indexX_green, indexX_green / width / 3, (indexX_green / 3) % width, maxX_green_change,
	  indexX_blue, indexX_blue / width / 3, (indexX_blue / 3) % width, maxX_blue_change,
	  indexX, indexX / width / 3, (indexX / 3) % width, maxX_change,
	  
	  indexY_red, indexY_red / width / 3, (indexY_red / 3) % width, maxY_red_change,
	  indexY_green, indexY_green / width / 3, (indexY_green / 3) % width, maxY_green_change,
	  indexY_blue, indexY_blue / width / 3, (indexY_blue / 3) % width, maxY_blue_change,
	  indexY, indexY / width / 3, (indexY / 3) % width, maxY_change,
	  
	  index_red, index_red / width / 3, (index_red / 3) % width, max_red_change,
	  index_green, index_green / width / 3, (index_green / 3) % width, max_green_change,
	  index_blue, index_blue / width / 3, (index_blue / 3) % width, max_blue_change,
	  index, index / width / 3, (index / 3) % width, max_change,
	  
	  absX_index, absX_index / width / 3, (absX_index / 3) % width, absX_index % 3, max_absX_change,
	  MaxXChangeIndex[absX_index] / Columns, MaxXChangeIndex[absX_index] % Columns, 
	  absY_index, absY_index / width / 3, (absY_index / 3) % width, absY_index % 3, max_absY_change,
	  MaxYChangeIndex[absY_index] / Columns, MaxYChangeIndex[absY_index] % Columns);
  
  fprintf(res, "# i, e, p0, p1, avg. intensity, avg. red, avg. green, avg. blue, avg. opacity, channel intensity\n");
  for(i = 0; i < Rows; i++) {
    for(e = 0; e < Columns; e++) {
      position = &POSITION(i,e);
      fprintf(res, "%d\t%d\t%d\t%d\t%f\t%f\t%f\t%f\t%e\t%d\n",
	      i, e, position->p[1], position->p[0],
	      Intensity[(i * Columns + e) * 4],
	      Intensity[(i * Columns + e) * 4 + 1],
	      Intensity[(i * Columns + e) * 4 + 2],
	      Intensity[(i * Columns + e) * 4 + 3],
	      Opacity[i * Columns + e],
	      Channel ? Channel[i * Columns + e] : 0);
    }
  }
}


/* This function makes sure that every element of the array between
 * first and last (inclusive) is a power of 2, between 1 and 256.
 */
int CheckPow2(float *array, int first, int last)
{
  int i;
  int pow;
  int errorP = 0;

  for (i=first; i <= last; i++) {
    /* find pow, the ceiling of log (base 2) of array[i]. */
    for (pow = 0; pow <= 8; pow++) {
      if (array[i] <= (1 << pow)) break;
    }	
    if (array[i] != (1 << pow)) {
      errorP = 1;
      array[i] = (1 << pow);
    }
  }
  return errorP;
}



void Usage(char *progName)
{
  fprintf
    (stderr, 
     "Usage:\n"
     "%s [options] <infield> <outfield>\n\n"
     "arguments:\n\n"
     "    -v                      Verbose mode on, prints progress.\n"
     "\n"
     "    -interactive [n]        Show pictures and stop after each step\n"
     "                            of level n or lower.\n"
     "\n"
     "    -no_display             Don't open window to display the images even\n"
     "                            if interactive mode has been enabled.\n"
     "\n"
     "    -no_images              The renderer is used even if an image exists.\n"
     "                            Images are still created if requested.\n"
     "\n"
     "    -create_images          Save missing images after rendering.\n"
     "\n"
     "    -no_renderer            Don't initialize renderer - all images have to exist.\n"
     "\n"
     "    <infield>               Should be either a .lid file\n"
     "                            (LIght field Description file), which\n"
     "                            specifies a set of .rgb files, or a .lif\n"
     "                            file, which is the native light field\n"
     "                            format.\n"
     "\n"
     "    <outfile>	          The output file.\n"
     "\n\n"
     "The following options override the corresponding settings in the input file:\n"
     "\n"
     "    -range p min max n      Compare pictures whose parameter p starts at\n"
     "                            min and goes to max in n steps.\n"
     "\n"
     "    -steps p n              Compare pictures with parameter p within the range\n"
     "                            given in the .lid file in n steps.\n"
     "\n"
     "    -size x y               Use images of this size.\n"
     "\n"
     "    -quality n              Use quality #n to render images.\n"
     "\n"
     "    -force n1 n2 n3 ...     Remove samples in p0 in this order.\n"
     "\n"
     "    -fimm                   Remove all -force samples at once.\n"
     "\n"
     "    -linear                 Only use linear interpolation.\n"
     "\n"
     "    -abs                    Use absolute difference between images,\n"
     "                            not squared ones.\n"
     "\n"

#ifdef INTERNAL_OPTIONS
     "\n\n"
     "The following options have been used for debugging and evaluation and\n"
     "are not really useful:\n"
     "\n"
     "    -noarea                 Don't use area of influence restriction,\n"
     "                            but used positions instead.\n"
     "\n"
     "    -dummy                  Use builtin function instead of images.\n"
     "\n"
     "    -analysis               Calculate and output average intensities and\n"
     "                            gradients instead.\n"
     "\n"
     "    -opacityonly            Only calculate average opacities.\n"
     "\n"
     "    -histogram              Print histogram to stdout.\n"
     "\n"
     "    -basis                  Output basis functions even if no dummies are\n"
     "                            used; only applicable to 1D ranges.\n"
     "\n"

#endif

     "Some other options are not even listed, see source...\n"
     "\n"
     , progName);
  
  exit(-1);
}

/*
** compare two memory areas and calculate quadratic difference
** or averaged absolute  
*/
static err_t diffMemSQ(unsigned char *a, unsigned char *b, int size)
{
  long result = 0;
  int osize = size;

  if(size == ImageSize) {
    while(--size >= 0) {
      int diff = *a++ - *b++;
      result += diff * diff;
    }
  } else {
    int *grid = Grid[CurrentSet];
    while(--size >= 0) {
      int diff = a[*grid] - b[*grid];
      grid++;
      result += diff * diff;
    }
  }    
  return(sqrt(((double)result)/osize));
}
static err_t diffMemABS(unsigned char *a, unsigned char *b, int size)
{
  long result = 0;
  int osize = size;

  if(size == ImageSize) {
    while(--size >= 0) {
      int diff = *a++ - *b++;
      if(diff < 0)
	result -= diff;
      else
	result += diff;
    }
  } else {
    int *grid = Grid[CurrentSet];
    while(--size >= 0) {
      int diff = a[*grid] - b[*grid];
      grid++;
      if(diff < 0)
	result -= diff;
      else
	result += diff;
    }
  }
  return(((err_t)result)/osize);
}

/* did position have an influence on i/e? */
bool_t InflInterp(struct Position *position, int i, int e)
{
  /* inside square? */
#if 0
  /* including borders */
  if(position->bottom <= i &&
     position->top >= i &&
     position->left <= e &&
     position->right >= e) {
      return TRUE;
  }
  return FALSE;
#else
  /* excluding borders */
  if(((position->left < 0 && position->right >= Columns) ?
      (position->bottom <= i && i <= position->top) :
      (position->bottom < i && i < position->top)) &&
     ((position->bottom < 0 && position->top >= Rows) ?
      (position->left <= e && e <= position->right) :
      (position->left < e && e < position->right))) {
      return TRUE;
  }
#endif


  /* one of the neighbors themselves? */
  if((i == position->i && (position->left == e || position->right == e)) ||
     (e == position->e && (position->bottom == i || position->top == i))) {
    return TRUE;
  }

  /* none of it, so no influence */
  return FALSE;
}

/* did dropping position change accumulated error of i/e? */
bool_t InflAccum(struct Position *position, int i, int e)
{
  struct Position *dropped = &POSITION(i,e);

  /* accumulation changes if there is a true overlap in both parameters */
  if(((position->bottom < dropped->top && dropped->top < position->top) ||
      (position->bottom < dropped->bottom && dropped->bottom < position->top) ||
      (dropped->bottom <= position->bottom && position->top <= dropped->top)) &&
     ((position->left < dropped->left && dropped->left < position->right) ||
      (position->left < dropped->right && dropped->right < position->right) ||
      (dropped->left <= position->left && position->right <= dropped->right)))
    return TRUE;
  else
    return FALSE;
}

static bool_t Depended(struct Position *sampled, struct Position *interpolated)
{
  int vert, i = sampled->i, e = sampled->e, j;

  if(NoDepended) {
    return TRUE;
  }

  for(vert = 0; vert <= 1; vert++) {
    for(j = 0; j < interpolated->interp.sources[vert].num; j++) {
      if(interpolated->interp.sources[vert].sources[j][0] == i &&
	 interpolated->interp.sources[vert].sources[j][1] == e) {
	return TRUE;
      }
    }
  }
  return FALSE;
}


/*
** list handling
*/
static void Insert(struct Position *position)
{
  err_t error = position->interp.error;
  struct Position **nextp;

  /* insert into list */
  for(nextp = &List;
      *nextp && ((*nextp)->interp.error <= error);
      nextp = &((*nextp)->next))
    ;
  position->next = *nextp;
  *nextp = position;
}

static void Remove(struct Position *position)
{
  struct Position **nextp;

  for(nextp = &List;
      *nextp && *nextp != position;
      nextp = &((*nextp)->next))
    ;
  if(*nextp) {
    *nextp = position->next;
  }
}

static double CalcFactor(double *c, int n, int x0, int x, int dx)
{
  double factor = 1;
  double diff, power;
  int i;

  if(Linear) {
    factor = 1 - (x - x0) / (double)dx;
  } else {
    if(Smooth) {
      for(i = 0, diff = (x - x0) / (double)dx, power = diff * diff;
	  i <= n;
	  i++, power *= diff) {
	factor += c[n - i] * power;
      }
    } else {
      for(i = 0, power = diff = (x - x0) / (double)dx;
	  i < n;
	  i++, power *= diff) {
	factor += c[n - i - 1] * power;
      }
    }
  }
  return factor;
}
  
/*
** calculate horizontal or vertical basis function for position (i,e)
** upwards or to the right so that the sample in the given distance
** is used
*/
static double **X;  /* nxn with n == max(Rows,Columns) */
static double **X_; /* inverse of X */
static double  *Y;  /* n    "                "         */
static bool_t AllocXY(void)
{
  int i, n = max(Columns, Rows) + 2;

  if(!(Y = malloc(n * sizeof(Y[0]))) ||
     !(X_ = malloc(n * sizeof(X_[0])))||
     !(X = malloc(n * sizeof(X[0]))))
    return FALSE;
  for(i = 0; i < n; i++) {
    if(!(X[i] = malloc(n * sizeof(X[0][0]))))
      return FALSE;
    if(!(X_[i] = malloc(n * sizeof(X_[0][0]))))
      return FALSE;
  }
  return TRUE;
}
static void InvertX(int n)
{
  int i, j, k;
  double v;

  /* setup identity */
  for(i = 0; i < n; i++) {
    for(j = 0; j < n; j++) {
      X_[i][j] = i == j ? 1.0 : 0.0;
    }
  }
  /* lower triangle */
  for(i = 0; i < n; i++) {
    if(!NoPivot) {
      /* search pivot */
      for(v = fabs(X[i][i]), k = i, j = i + 1;
	  j < n;
	  j++) {
	if(fabs(X[j][i]) > v) {
	  v = fabs(X[j][i]);
	  k = j;
	}
      }

      /* swap lines */
      for(j = 0; j < n; j++) {
	v = X[i][j];
	X[i][j] = X[k][j];
	X[k][j] = v;
	
	v = X_[i][j];
	X_[i][j] = X_[k][j];
	X_[k][j] = v;
      }
    }

    v = 1/X[i][i];
    /* make line start with 1 */
    for(k = 0; k < n; k++) {
      X[i][k] *= v;
      X_[i][k] *= v;
    }
    /* subtract line from all below it */
    for(j = i + 1; j < n; j++) {
      v = X[j][i];
      for(k = 0; k < n; k++) {
	X[j][k] -= v * X[i][k];
	X_[j][k] -= v * X_[i][k];
      }
    }
  }
  /* upper triangle */
  for(i = n - 1; i >= 0; i--) {
    /* subtract line from all above it */
    for(j = i - 1; j >= 0; j--) {
      v = X[j][i];
      for(k = 0; k < n; k++) {
	X[j][k] -= v * X[i][k];
	X_[j][k] -= v * X_[i][k];
      }
    }
  }
}
static void MulXY(double *c, int n, int weight)
{
  int i, j;

  for(i = 0; i < n; i++) {
    for(j = 0; j < n; j++) {
      c[i] += X_[i][j] * Y[j] * weight;
    }
  }
}
static void CalcBasis(int i, int e, bool_t vert, int d)
{
  int m,k;
  double **cp = vert ? &POSITION(i,e).cy[d] : &POSITION(i,e).cx[d];
  double *c;
  unsigned long contribs;

  /* interpolation possible? */
  if(vert ?
     i + d < Rows :
     e + d < Columns) {
    int gdiff = vert ?
	  POSITION(i+d,e).p[1] - POSITION(i,e).p[1]:
	  POSITION(i, e+d).p[0] - POSITION(i,e).p[0];

    if(*cp) {
      /* done already */
      return;
    }
    if(!(*cp = malloc((d + 2) * sizeof(double)))) {
      fprintf(stderr, "no memory!\n");
      exit(1);
    }
    c = *cp;

    if(!Linear) {
      /* yes, setup matrix */
      if(Smooth) {
	double power;

	for(k = d; k >= 0; k--) {
	  X[0][k] = (d + 2 - k);
	}
	for(m = 1; m <= d; m++) {
	  double diff = (vert ?
	    POSITION(i+m,e).p[1] - POSITION(i,e).p[1]:
	    POSITION(i, e+m).p[0] - POSITION(i,e).p[0]) / (double)gdiff;
	  for(k = d, power = diff * diff; k >= 0; k--, power *= diff) {
	    X[m][k] = power;
	  }
	}
      } else {
	for(m = 0;
	    m < d;
	    m++) {
	  double diff = (vert ?
	    POSITION(i+m+1,e).p[1] - POSITION(i,e).p[1]:
	    POSITION(i, e+m+1).p[0] - POSITION(i,e).p[0]) / (double)gdiff;
	  double power;
	  
	  for(k = d - 1, power = diff; k >= 0; k--, power *= diff) {
	    X[m][k] = power;
	  }
	}
      }
      /* invert matrix */
      InvertX(Smooth ? d + 1 : d);
      /* now calculate average coefficients */
      for(k = 0; k < d + 1; k++) {
	c[k] = 0.0;
      }
      for(m = 0, contribs = 0; m < GSize[CurrentSet]; m++) {
	/* set Y vector */
	int y_other = vert ? POSITION(i+d,e).IMAGE(m) : POSITION(i,e+d).IMAGE(m);
	int y_this = POSITION(i,e).IMAGE(m);
	int div = y_this - y_other;
	int weight = NoWeights ? 1 : abs(div);

	if(div) {
	  if(Smooth) {
	    Y[0] = 0;
	    for(k = 1; k <= d; k++) {
	      int y = vert ? POSITION(i+k,e).IMAGE(m) : POSITION(i,e+k).IMAGE(m);

	      if(Clamp) {
		if(y_this < y_other) {
		  if(y < y_this) {
		    y = y_this;
		    ClampCount++;
		  } else if(y > y_other) {
		    y = y_other;
		    ClampCount++;
		  }
		} else {
		  if(y < y_other) {
		    y = y_other;
		    ClampCount++;
		  } else if(y > y_this) {
		    y = y_this;
		    ClampCount++;
		  }
		}
	      }
		  
	      Y[k] = (y - y_this) / (double)div;
	    }
	    MulXY(c,d + 1, weight);
	  } else {
	    for(k = 0; k < d; k++) {
	      int y = vert ? POSITION(i+k,e).IMAGE(m) : POSITION(i,e+k).IMAGE(m);

	      if(Clamp) {
		if(y_this < y_other) {
		  if(y < y_this) {
		    y = y_this;
		    ClampCount++;
		  } else if(y > y_other) {
		    y = y_other;
		    ClampCount++;
		  }
		} else {
		  if(y < y_other) {
		    y = y_other;
		    ClampCount++;
		  } else if(y > y_this) {
		    y = y_this;
		    ClampCount++;
		  }
		}
	      }

	      Y[k] = ((vert ? POSITION(i+k+1,e).IMAGE(m) : POSITION(i,e+k+1).IMAGE(m)) - y_this) / (double)div;
	    }
	    MulXY(c,d,weight);
	  }
	  contribs += weight;
	}
      }
      if(contribs) {
	for(k = 0; k <= d; k++) {
	  c[k] /= contribs;
	}
      }
    } else {
      k = Smooth ? d : d - 1;
      c[k] = -1.0;
      for(--k; k >= 0; k--) {
	c[k] = 0.0;
      }
    }
    #if 0
    InterpolatePictures(LeftBasis, POSITION(i,e).image, POSITION(i,e+d).image,
			CalcFactor(c, d, POSITION(i,e).p[0], POSITION(i,e).p[0], gdiff));
    InterpolatePictures(RightBasis, POSITION(i,e).image, POSITION(i,e+d).image,
			CalcFactor(c, d, POSITION(i,e).p[0], POSITION(i,e+d).p[0], gdiff));
    printf("CalcBasis(%d,%d,%d) => left %d ?= %d, right %d ?= %d\n", i,e,d,
	   LeftBasis[ImageSize/2], POSITION(i,e).image[ImageSize/2],
	   RightBasis[ImageSize/2], POSITION(i,e+d).image[ImageSize/2]);
    #endif
  } else {
    /* no, full influence */
  }
}



/*
** interpolate between two pictures
*/
static float itp_factor;
static bool_t itp_debug;
static void InterpolatePictures(unsigned char *target,
				unsigned char *source1,
				unsigned char *source2,
				double factor)
{
  int i;
  int fp = factor * (1<<Bits) + 0.5;

  if(FInterp) {
    return;
  }

  if(itp_debug) {
    lfOutput("\nInterpolate: factor %e, fp %d/2**%d=%e\n",
	     factor, fp, Bits, ((double)fp)/(1<<Bits));
  }

  itp_factor = factor;
  i = ImageSize;
  while(--i >= 0) {
    *target++ = Dummy ?
      (unsigned char)(*source2 + (*source1 - *source2) * factor + 0.5) :
      RLERP(*source2, *source1, fp, Bits);
    if(itp_debug && !Dummy) {
      unsigned char exact = (*source2 + (*source1 - *source2) * factor + 0.5);
      if(exact != target[-1]) {
	lfOutput("   %d: %d != %d\n", i, target[-1], exact);
      }
    }
    source1++;
    source2++;
  }
}

void AddSource(struct sources *sources, int i, int e, double factor)
{
#ifdef DEBUG
  int j, checksum = 0;

  for(j = 0; j < ImageSize; j++) {
    checksum += POSITION(i,e).image[j];
  }

  sources->checksum[sources->num] = checksum;
#endif
  ADD_SOURCE(sources, i, e, factor);
}

static void AddFactor(struct sources *sources, double factor, int source, int target)
{
  sources->factor[sources->fnum] = factor;
  sources->source_indices[sources->fnum][0] = source;
  sources->source_indices[sources->fnum][1] = target;
  sources->fnum++;
}


/*
** interpolate the image for position(i,e) if necessary
** or enforced and store it in the argument buffer;
** return number of used samples and their indices
*/
static int InterpolateLine(int i, int e, int p, bool_t vert, unsigned char *image, struct sources *sources)
{
  int f = -1, s = -1, rem;
  int m;

  /* search for upper and lower neighbor */
  for(m = vert ? i : e; m < (vert ? Rows : Columns); m++) {
    if(vert ?
       (POSITION(m, e).p[1] >= p && POSITION(m, e).sampled) :
       (POSITION(i, m).p[0] >= p && POSITION(i, m).sampled)) {
      s = m;
      break;
    }
  }
  for(m = vert ? i : e; m >= 0; m--) {
    if(vert ?
       (POSITION(m, e).p[1] <= p && POSITION(m, e).sampled) :
       (POSITION(i, m).p[0] <= p && POSITION(i, m).sampled)) {
      f = m;
      break;
    }
  }

  if((rem = f) != -1 && (rem = s) != -1) {
    int d = s - f;

    /* interpolate or take over picture */
    if(d) {
      double factor;
      if(vert) {
	CalcBasis(f,e, TRUE, d);
	factor = CalcFactor(POSITION(f, e).cy[d], d,
			    POSITION(f, e).p[1], p,
			    POSITION(f + d, e).p[1] - POSITION(f, e).p[1]);
	InterpolatePictures(image,
			    POSITION(f, e).image, 
			    POSITION(s, e).image,
			    factor);
	AddSource(sources, f, e, factor);
	AddSource(sources, s, e, 1 - factor);
      } else {
	CalcBasis(i,f, FALSE, d);
	factor = CalcFactor(POSITION(i, f).cx[d], d,
			    POSITION(i, f).p[0], p,
			    POSITION(i, f + d).p[0] - POSITION(i, f).p[0]);
	InterpolatePictures(image,
			    POSITION(i, f).image, 
			    POSITION(i, s).image,
			    factor);
	AddSource(sources, i, f, factor);
	AddSource(sources, i, s, 1 - factor);
      }
      AddFactor(sources, factor, sources->num - 1, sources->num - 2);
      return 2;
    } else {
      if(!FInterp) {
	memcpy(image, vert ? POSITION(f, e).image : POSITION(i, f).image, ImageSize);
      }
      if(vert) {
	AddSource(sources, f, e, 1);
      } else {
	AddSource(sources, i, f, 1);
      }
      return 1;
    }
  } else if((rem = f) != -1 || (rem = s) != -1) {
    if(!FInterp) {
      memcpy(image, vert ? POSITION(rem, e).image : POSITION(i, rem).image, ImageSize);
    }
    if(vert) {
      AddSource(sources, rem, e, 1);
    } else {
      AddSource(sources, i, rem, 1);
    }
    return 1;
  } else {
    return 0;
  }
}




#if 0

/*
** get (interpolated) picture from neighboring columns/rows;
** returns 0 if neighbors have no entries, otherwise number of samples used
** and the line used within them plus the column/row index
*/
static int InterpolateNeighbors(int i, int e, int direction, bool_t vert,
				unsigned char *image,
				int *index, int *line1, int *line2)
{
  int n, num;

  for(n = (vert ? e : i) + direction;
      n >= 0 && n < (vert ? Columns : Rows);
      n += direction) {
    if((num = vert ?
	InterpolateLine(i, n, TRUE, image, line1, line2) :
	InterpolateLine(n, e, FALSE, image, line1, line2))) {
      *index = n;
      return num;
    }
  }
  return 0;
}
#endif
  
/*
** calculate interpolated picture for position(i,e)
** by trying requested direction only
*/
static int InterpolateDirection(int i, int e, int p[2], bool_t vert, unsigned char *image, struct sources *sources)
{
  int num_f, num_s;
  int findex;
  int sindex;
  int j;
  bool_t within_grid = POSITION(i,e).p[!vert] == p[!vert]; /* the parameter is aligned to the entry in the major direction */

  memset(sources, 0, sizeof(*sources));
  if((vert ? RNeigh[e] == e : BNeigh[i] == i)
     && (within_grid || (!vert ? i + 1 >= Rows : e + 1 >= Columns))) {
    /* entries within the line left */
    num_f = InterpolateLine(i, e, p[vert], vert, image, sources);
    assert(num_f > 0);
  } else {
    /* interpolate neighbors in primary direction */
    if((findex = (vert ? LNeigh[e] : BNeigh[i])) >= 0) {
      num_f = vert ? InterpolateLine(i, findex, p[1], vert, Left, sources)
	: InterpolateLine(findex, e, p[0], vert, Left, sources);
      assert(num_f > 0);
    } else {
      num_f = 0;
    }
    if(vert ? (sindex = RNeigh[within_grid ? e : e + 1]) < Columns :
       (sindex = TNeigh[within_grid ? i : i + 1]) < Rows) {
      num_s = vert ? InterpolateLine(i, sindex, p[1], vert, Right, sources)
	: InterpolateLine(sindex, e, p[0], vert, Right, sources);
      assert(num_s > 0);
    } else {
      num_s = 0;
    }
    if(num_f && num_s) {
      int d = sindex - findex;
      double factor;

      /* interpolate in secondary direction */
      if(!vert) {
	/* now vertical */
	CalcBasis(findex,e, TRUE, d);
	factor = CalcFactor(POSITION(findex, e).cy[d], d,
			    POSITION(findex, e).p[1], p[1],
			    POSITION(findex + d, e).p[1] -
			    POSITION(findex, e).p[1]);
	if(!within_grid && e + 1 < Columns) {
	  double sfactor;

	  CalcBasis(findex,e+1, TRUE, d);
	  sfactor = CalcFactor(POSITION(findex, e+1).cy[d], d,
			       POSITION(findex, e+1).p[1], p[1],
			       POSITION(findex + d, e+1).p[1] -
			       POSITION(findex, e+1).p[1]);
	  factor = factor + (sfactor - factor) *
	    (p[0] - POSITION(0,e).p[0]) / (POSITION(0,e+1).p[0] - POSITION(0,e).p[0]);
	}
	InterpolatePictures(image,
			    Left, Right,
			    factor);
      } else {
	/* now horizontal */
	CalcBasis(i,findex, FALSE, d);
	factor = CalcFactor(POSITION(i, findex).cx[d], d,
			    POSITION(i, findex).p[0], p[0],
			    POSITION(i, findex + d).p[0] -
			    POSITION(i, findex).p[0]);
	if(!within_grid && i + 1 < Rows) {
	  double sfactor;

	  CalcBasis(i+1,findex, FALSE, d);
	  sfactor = CalcFactor(POSITION(i+1, findex).cx[d], d,
			       POSITION(i+1, findex).p[0], p[0],
			       POSITION(i+1, findex + d).p[0] -
			       POSITION(i+1, findex).p[0]);
	  factor = factor + (sfactor - factor) * 
	    (p[1] - POSITION(i,0).p[1]) / (POSITION(i+1,0).p[1] - POSITION(i,0).p[1]);
	}
	InterpolatePictures(image,
			    Left, Right,
			    factor);
      }
      AddFactor(sources, factor, num_f, 0);
      j = 0;
      while(j < num_f) {
	sources->sfactor[j] *= factor;
	j++;
      }
      while(j < num_f + num_s) {
	sources->sfactor[j] *= 1 - factor;
	j++;
      }
    } else if(!FInterp) {
      if(num_f) {
	memcpy(image, Left, ImageSize);
      } else if(num_s) {
	memcpy(image, Right, ImageSize);
      }
    }
  }
  return sources->num;
}


/*
** calculate interpolated picture for position(i,e)
** by trying both directions
*/
static void InterpolatePosition(int i, int e, int dropped, bool_t cached, bool_t full)
{
  int num_vert, num_horiz;
  struct Position *position = &POSITION(i, e);
  bool_t vert, horiz;

#ifdef CACHED
  if(dropped < 0 || !cached) {
    vert = TRUE;
    horiz = TRUE;
  } else if(IS_SET(position->vert_buffer, dropped)) {
    vert = TRUE;
    horiz = FALSE;
  } else {
    vert = FALSE;
    horiz = TRUE;
  }
#else
  vert = TRUE;
  horiz = TRUE;
#endif

  /* try interpolation with columns and with rows as primary direction
     => we need the dependencies even if we know which direction would be used */
  if(Rows > 1) {
    num_vert  = vert ?
      InterpolateDirection(i, e, position->p, TRUE, Buffer, &position->interp.sources[TRUE]) : 0;
  } else {
    /* shortcut... */
    num_vert = 0;
    position->interp.sources[TRUE].num = 0;
  }
  num_horiz = horiz ?
    InterpolateDirection(i, e, position->p, FALSE, Interp, &position->interp.sources[FALSE]) : 0;

  /* now choose best interpolation */
  if(num_vert || num_horiz) {
    err_t diff_vert = 0, diff_horiz = 0;

#ifdef DEBUG
    position->interp.sources[TRUE].error =
#endif
      diff_vert = num_vert ? diff_mem(Buffer, position->image, full ? ImageSize : GSize[CurrentSet]) : 0.0;
#ifdef DEBUG
    position->interp.sources[FALSE].error = 
#endif
      diff_horiz = num_horiz ? diff_mem(Interp, position->image, full ? ImageSize : GSize[CurrentSet]) : 0.0;

    if(!memcmp(&position->interp.sources[FALSE],
	       &position->interp.sources[TRUE],
	       sizeof(position->interp.sources[FALSE]))) {
      /* same sources, use more obvious direction */
      if(BNeigh[i] != i || VertOnly) {
	/* no interpolation within row, disable horizontal */
	num_horiz = 0;
      } else {
	num_vert = 0;
      }
    }

    if(!num_vert || (!VertOnly && (num_horiz && diff_horiz <= diff_vert))) {
      position->interp.error = diff_horiz;
      position->interp.vert = FALSE;
    } else {
      memcpy(Interp, Buffer, ImageSize);
      position->interp.error = diff_vert;
      position->interp.vert = TRUE;
    }
  } else {
    memset(Interp, 0, ImageSize);
    position->interp.error = diff_mem(Interp, POSITION(i,e).image, GSize[CurrentSet]);
  }

  #ifdef CACHED
  if(dropped >= 0 && cached) {
    if(position->interp.vert) {
      assert(IS_SET(position->vert_buffer, dropped));
    } else {
      assert(!IS_SET(position->vert_buffer, dropped));
    }
  }

  if(dropped >= 0 && !cached) {
    if(position->interp.vert) {
      SET(position->vert_buffer, dropped);
    } else {
      CLEAR(position->vert_buffer, dropped);
    }
  }
  #endif
}


static lerr_t SumErrors(void)
{
  int i, e;
  lerr_t error = 0;

  for(i = 0; i < Rows; i++) {
    for(e = 0; e < Columns; e++) {
      if(!POSITION(i,e).sampled) {
	error += POSITION(i,e).interp.error;
	/*	lfOutput("  + %f (%d,%d) = %f\n",
		POSITION(i,e).interp.error, i,e, error);*/
      }
    }
  }
  return error;
}

/* find new minimum element */
int FME_Start, FME_Left;
static bool_t FindMinError(void)
{
  int i, e;
  struct Position *position;

  if(FME_Left) {
    MinErrorPos[0] = ForceIndices[FME_Start] / Columns;
    MinErrorPos[1] = ForceIndices[FME_Start] % Columns;
    FME_Start++;
    FME_Left--;
    return TRUE;
  }

  MinErrorPos[0] = -1;
  for(i = 0; i < Rows; i++) {
    for(e = 0; e < Columns; e++) {
      position = &POSITION(i,e);
      if(position->sampled &&
	 (MinErrorPos[0] == -1 || MinError > (position->sum + position->interp.error))) {
	MinErrorPos[0] = i;
	MinErrorPos[1] = e;
	MinError = position->sum + position->interp.error;
      }
    }
  }
  return MinErrorPos[0] != -1;
}

static void TimeStamp(char *fmt, int arg)
{
  time_t cur;
  static time_t first, prev;

  time(&cur);
  if(!first) {
    first = cur;
    prev = cur;
  }
  fprintf(res, "# time %f %f %d ",
	  difftime(cur, first), difftime(cur, prev),
	  NumInterp);
  if(NumInterp) {
    fprintf(res, "%d %d %d ", Dropped[NumInterp - 1], Dropped[NumInterp - 1] / Columns, Dropped[NumInterp - 1] % Columns);
  } else {
    fprintf(res, "0 0 0 ");
  }
  fprintf(res, "%e %e : ", Error/ERR_FACTOR, Error/ERR_FACTOR / (NumInterp ? NumInterp : 1));
  fprintf(res, fmt, arg);
  fprintf(res, "  - %s", ctime(&cur));
  fflush(res);
  prev = cur;
}

static void AddSample(int i, int e)
{
  int j;
  struct Position *position = &POSITION(i,e);
  
  if(BNeigh[i] == i) {
    /* update neighbors in row */
    for(j = e + 1; j < Columns && !POSITION(i,j).sampled; j++)
      ;
    position->right = j;
    if(j < Columns) {
      POSITION(i,j).left = e;
    }
    for(j = e - 1; j >= 0 && !POSITION(i,j).sampled; j--)
      ;
    position->left = j;
    if(j >= 0) {
      POSITION(i,j).right = e;
    }
  } else {
    /* single entry, update T/BNeigh[] */
     position->left = -1;
     position->right = Columns;
     BNeigh[i] = TNeigh[i] = i;
     for(j = i + 1; j < Rows && BNeigh[j] < i; j++) {
       BNeigh[j] = i;
     }
     for(j = i - 1; j >= 0 && TNeigh[j] > i; j--) {
       TNeigh[j] = i;
     }
  }

  if(LNeigh[e] == e) {
    /* update neighbors in row */
    for(j = i + 1; j < Rows && !POSITION(j,e).sampled; j++)
      ;
    position->top = j;
    if(j < Rows) {
      POSITION(j,e).bottom = i;
    }
    for(j = i - 1; j >= 0 && !POSITION(j,e).sampled; j--)
      ;
    position->bottom = j;
    if(j >= 0) {
      POSITION(j,e).top = e;
    }
  } else {
    /* single entry, update L/RNeigh[] */
     position->bottom = -1;
     position->top = Rows;
     LNeigh[e] = RNeigh[e] = e;
     for(j = e + 1; j < Columns && LNeigh[j] < e; j++) {
       LNeigh[j] = e;
     }
     for(j = e - 1; j >= 0 && RNeigh[j] > i; j--) {
       RNeigh[j] = e;
     }
  }

  position->sampled = TRUE;
}

static void DropSample(int i, int e, struct Position *head, bool_t drop)
{
  /* remember old state so that we can restore it */
  static int nl, nr, nt, nb, hl, hr, ht, hb;
  static bool_t old_sampled;
  static int lc, rc, tr, br;
  bool_t last_r, last_c;
  int tmp;

  struct Position *position = &POSITION(i,e);
  last_r = position->left < 0 && position->right >= Columns;
  last_c = position->bottom < 0 && position->top >= Rows;

  if(drop) {
    old_sampled = position->sampled;
    position->sampled = FALSE;

    if(old_sampled) {
      if(last_c) {
	lc = e;
	tmp = e + 1 < Columns ? RNeigh[e + 1] : Columns;
	do {
	  RNeigh[lc--] = tmp;
	} while(lc >= 0 && RNeigh[lc] == e);
	rc = e;
	tmp = e > 0 ? LNeigh[e - 1] : -1;
	do {
	  LNeigh[rc++] = tmp;
	} while(rc < Columns && LNeigh[rc] == e);
      }
      if(last_r) {
	br = i;
	tmp = i + 1 < Rows ? TNeigh[i + 1] : Rows;
	do {
	  TNeigh[br--] = tmp;
	} while(br >= 0 && TNeigh[br] == i);
	tr = i;
	tmp = i > 0 ? BNeigh[i - 1] : -1;
	do {
	  BNeigh[tr++] = tmp;
	} while(tr < Rows && BNeigh[tr] == i);
      }
    }
    /* update neighbors */
    if(position->top < Rows) {
      nb = POSITION(position->top, e).bottom;
      POSITION(position->top, e).bottom = position->bottom;
    }
    if(position->bottom >= 0) {
      nt = POSITION(position->bottom, e).top;
      POSITION(position->bottom, e).top = position->top;
    }
    if(position->right < Columns) {
      nl = POSITION(i, position->right).left;
      POSITION(i, position->right).left = position->left;
    }
    if(position->left >= 0) {
      nr = POSITION(i, position->left).right;
      POSITION(i, position->left).right = position->right;
    }

    if(head) {
      hl = head->left, hr = head->right,
	ht = head->top, hb = head->bottom;
      
      if(i == head->i) {
	if(head->right == e) {
	  head->right = position->right;
	} else if(head->left == e) {
	  head->left = position->left;
	}
      } else if(e == head->e) {
	if(head->bottom == i) {
	  head->bottom = position->bottom;
	} else if(head->top == i) {
	  head->top = position->top;
	}
      }
    }
  } else {
    /* undo dropping */

    if(old_sampled) {
      if(last_c) {
	for(++lc; lc <= e; lc++) {
	  RNeigh[lc] = e;
	}
	for(--rc; rc >= e; rc--) {
	  LNeigh[rc] = e;
	}
      }
      if(last_r) {
	for(++br; br <= i; br++) {
	TNeigh[br] = i;
	}
	for(--tr; tr >= i; tr--) {
	  BNeigh[tr] = i;
	}
      }
    }

    if(position->top < Rows) {
      POSITION(position->top, e).bottom = nb;
    }
    if(position->bottom >= 0) {
      POSITION(position->bottom, e).top = nt;
    }
    if(position->right < Columns) {
      POSITION(i, position->right).left = nl;
    }
    if(position->left >= 0) {
      POSITION(i, position->left).right = nr;
    }
    position->sampled = old_sampled;
    if(head) {
      head->left = hl;
      head->right = hr;
      head->bottom = hb;
      head->top = ht;
    }
  }
}
	    
static void RecalcAll(void)
{
  int i, e;
  struct Position *position;
  int f, s;
  
  for(i = 0; i < Rows; i++) {
    BNeigh[i] = -1;
    TNeigh[i] = Rows;
  }
  for(e = 0; e < Columns; e++) {
    LNeigh[e] = -1;
    RNeigh[e] = Columns;
  }
  for(i = 0; i < Rows; i++) {
    for(e = 0; e < Columns; e++) {
      if(POSITION(i,e).sampled) {
	BNeigh[i] = i;
	TNeigh[i] = i;
	RNeigh[e] = e;
	LNeigh[e] = e;
      }
    }
  }
  f = -1;
  s = 0;
  i = 0;
  while(i < Rows) {
    while(s < Rows && BNeigh[s] != s) {
      s++;
    }
    while(i < s) {
      if(BNeigh[i] != i) {
	BNeigh[i] = f;
	TNeigh[i] = s;
      }
      i++;
    }
    f = s;
    s++;
  }
  f = -1;
  s = 0;
  e = 0;
  while(e < Columns) {
    while(s < Columns && LNeigh[s] != s) {
      s++;
    }
    while(e < s) {
      if(LNeigh[e] != e) {
	LNeigh[e] = f;
	RNeigh[e] = s;
      }
      e++;
    }
    f = s;
    s++;
  }

  for(i = 0; i < Rows; i++) {
    for(e = 0; e < Columns; e++) {
      position = &POSITION(i,e);
      DropSample(i,e, NULL, TRUE);
      InterpolatePosition(i,e, 0, FALSE, FALSE);
      DropSample(i,e, NULL, FALSE);
    }
  }
  Error = SumErrors();

  /* recalculate _all_ accumulated errors */
  for(i = 0; i < Rows; i++) {
    for(e = 0; e < Columns; e++) {
      position = &POSITION(i,e);
      if(position->sampled) {
	lerr_t sum = 0;
	int i_, e_;
	struct interp buffer;
	
	/* simulate removal of this sample */
	lfOutput("   do recalc %d/%d\n", i,e);
	DropSample(i,e,NULL,TRUE);
	  
	/* accumulate error */
	for(i_ = 0; i_ < Rows; i_++) {
	  for(e_ = 0; e_ < Columns; e_++) {
	    POSITION(i_,e_).contribs[i * Columns + e] = 0;
	    if(!(i_ == i && e_ == e) &&
	       !POSITION(i_,e_).sampled) {
	      err_t old_error, error;
	    
	      buffer = POSITION(i_,e_).interp;
	      old_error = POSITION(i_,e_).interp.error;
	      InterpolatePosition(i_, e_, i * Columns + e, FALSE, FALSE);
	      error = POSITION(i_,e_).interp.error;
	      POSITION(i_,e_).contribs[i * Columns + e] = error - old_error;
	      lfOutput("      old/new error %d/%d: %f (%s) / %f (%s) => diff %f\n",
		       i_, e_,
		       old_error/ERR_FACTOR, buffer.vert ? "|" : "-",
		       error/ERR_FACTOR, POSITION(i_,e_).interp.vert ? "|" : "-",
		       (error - old_error)/ERR_FACTOR);
	      sum += error - old_error;
	      POSITION(i_,e_).interp = buffer;
	    }
	  }
	}
	/* set new sum */
	lfOutput("      => new sum %f\n", sum);
	position->sum = sum;
	
	/* undo simulation */
	DropSample(i,e,NULL,FALSE);
      }
    }
  }
}

static void DumpDropped(int current_set)
{
  int i, e;

  fprintf(res, "\n#griddropped%02d ", current_set);
  for(i = 0; i < Rows; i++) {
    for(e = 0; e < Columns; e++) {
      if(!POSITION(i,e).sampled) {
	fprintf(res, "%d ", i * Columns + e);
      }
    }
  }
  fprintf(res, "\n\n");
}


/* calculate all possible differences */
static void optError(void)
{
  struct Position *position, *head;
  int i, e;

  /* not really useful... */
  for(i = 0; i < Rows; i++) {
    for(e = 0; e < Columns; e++) {
      CalcBasis(i,e, TRUE, 1);
      CalcBasis(i,e, FALSE, 1);
    }
  }

  TimeStamp("init interpolations", 0);

  /* initialize all interpolations */
  NumInterp = 0;
  FME_Start = 0;
  FME_Left = NumForce;
  MinErrorPos[0] = -1;
  if(ForceImm) {
    while(FME_Left) {
      DropSample(ForceIndices[FME_Start] / Columns, ForceIndices[FME_Start] % Columns,
		 NULL, TRUE);
      FME_Start++;
      FME_Left--;
      NumInterp++;
    }
    RecalcAll();
  } else {
    lfOutput("************ setting up interpolations *************\n");
    Error = 0;
    /* parameter p1 */
    for(i = 0; i < Rows; i++) {
      /* parameter p0 */
      for(e = 0; e < Columns; e++) {
	position = &POSITION(i,e);
	DoInteraction("Init Interp", i, e, LEVEL_INIT_INTERP, FALSE);
	DropSample(i,e, NULL, TRUE);
	InterpolatePosition(i, e, i * Columns + e, FALSE,FALSE);
	DropSample(i,e, NULL, FALSE);
	position->sum = 0;
      }
    }
  }

  while(Error < MaxError &&
	(!NumInterp || Error/NumInterp < MaxAvgError) &&
	NumInterp < MaxInterp &&
	FindMinError()) {
    lerr_t summed_error;

    /* started browsing */
    if(BW) {
      return;
    }

    lfOutput("\nerror %f, %d/%d interpolated => avg. error %f\n",
	     Error / ERR_FACTOR, NumInterp, Rows * Columns, Error/NumInterp / ERR_FACTOR);
    TimeStamp("dropped %d", NumInterp);
    for(i = Rows - 1; i >= 0; i--) {
      for(e = 0; e < Columns; e++) {
	lfOutput(POSITION(i,e).sampled ? "X" : POSITION(i,e).interp.vert ? "|" : "-");
      }
      lfOutput("\n");
    }
    SaveState();

    /* remove head of List, then recalculate interpolations and the list */
    head = &POSITION(MinErrorPos[0], MinErrorPos[1]);
    summed_error = SumErrors();
    if(abs(Error - summed_error) > .00000001) {
      lfOutput("warning: calculated error %f smaller than current error %f!\n", Error/ERR_FACTOR, summed_error/ERR_FACTOR);
      fprintf(res, "\n# error warning: %f != %f\n", Error/ERR_FACTOR, summed_error/ERR_FACTOR);
    }
    if(DoInteraction("remove", head->i, head->e, LEVEL_REMOVE, FALSE)) {
      continue;
    }
    Error += head->sum + head->interp.error;
    DropSample(head->i, head->e, NULL, TRUE);


#ifdef USE_RECALCALL
    NumInterp++;
    RecalcAll();
    continue;
#endif

    for(i = 0; i < Rows; i++) {
      for(e = 0; e < Columns; e++) {
	if(InflInterp(head, i, e) &&
	   (i != head->i || e != head->e)) { /* don't reinterpolate the position itself */
	  err_t error;

	  /* (i,e) might have depended on head, so recalculate it */
	  if(Depended(head, &POSITION(i,e))) {
	    err_t diff;
	    bool_t old_sampled, old_vert;

	    position = &POSITION(i,e);
	    old_sampled = position->sampled;
	    old_vert = position->interp.vert;
	    error = position->interp.error;
	    DoInteraction("   reinterpolate", i,e, LEVEL_REINTERP, FALSE);
	    DropSample(i,e, NULL, TRUE);
	    InterpolatePosition(i,e, head->i * Columns + head->e, !old_sampled, FALSE);
	    diff = position->interp.error - error;
	    lfOutput("      error %f(%s) => %f(%s), %sdiff %f%s\n",
		     (double)error/ERR_FACTOR, old_vert ? "|" : "-",
		     (double)position->interp.error/ERR_FACTOR, position->interp.vert ? "|" : "-",
		     old_sampled ? "(" : "",
		     (double)diff/ERR_FACTOR,
		     old_sampled ? ")" : "");
	    if(!old_sampled &&
	       abs((error = position->contribs[head->i * Columns + head->e]) - diff) >= 0.0001) {
	      lfOutput("    error warning: expected %f instead\n", error/ERR_FACTOR);
	      fprintf(res, "# error warning reinterp %d %d after removing %d %d: expected %f not %f\n",
		      i, e, head->i, head->e, error/ERR_FACTOR, diff/ERR_FACTOR);
	    }
	    DropSample(i,e, NULL, FALSE);
	  } else {
	    lfOutput("   no reinterpolate for %d/%d\n", i, e);
	  }
	}
      }
    }

    Dropped[NumInterp] = MinErrorPos[0] * Columns + MinErrorPos[1];
    if(GSize[CurrentSet] == ImageSize) {
      GErrors[CurrentSet][NumInterp] = Error;
    } else {
      /* calculate real error */
      lerr_t error = 0;
      struct interp buffer;

      for(i = 0; i < Rows; i++) {
	for(e = 0; e < Columns; e++) {
	  buffer = POSITION(i,e).interp;
	  InterpolatePosition(i,e, -1, FALSE, TRUE);
	  error += POSITION(i,e).interp.error;
	  POSITION(i,e).interp = buffer;
	}
      }
      GErrors[CurrentSet][NumInterp] = error;
    }
    NumInterp++;

    /* recalculate the effect of sample removals for those samples
       that are affected be the previous removal */
    for(i = 0; i < Rows; i++) {
      for(e = 0; e < Columns; e++) {

	position = &POSITION(i,e);
	if(position->sampled &&
	   InflAccum(position, head->i, head->e)) {
	  lerr_t sum = 0;
	  int i_, e_;
	  struct interp buffer;
	  
	  lfOutput("   do recalc %d/%d\n", i,e);

	  /* simulate removal of this sample */
	  DropSample(i,e,head, TRUE);
	  
	  /* accumulate error */
	  for(i_ = 0; i_ < Rows; i_++) {
	    for(e_ = 0; e_ < Columns; e_++) {
	      if(!POSITION(i_,e_).sampled &&
		 (i_ != position->i || e_ != position->e)) {
		bool_t depended = TRUE;
		/* either use stored value or interpolate to find new error if it can
		   have changed - has to check the area of influence for head as if
		   position had been removed  */
		if(InflInterp(head, i_, e_) &&
		   InflInterp(position, i_, e_) &&
		   (depended = Depended(position, &POSITION(i_,e_)))) {
		  err_t old_error, error;
		  
		  buffer = POSITION(i_,e_).interp;
		  old_error = POSITION(i_,e_).interp.error;
		  InterpolatePosition(i_, e_, i * Columns + e, FALSE, FALSE);
		  error = POSITION(i_,e_).interp.error;
		  POSITION(i_,e_).contribs[i * Columns + e] = error - old_error;
		  lfOutput("      old/new error %d/%d: %f (%s) / %f (%s) => diff %f\n",
			   i_, e_,
			   old_error/ERR_FACTOR, buffer.vert ? "|" : "-",
			   error/ERR_FACTOR, POSITION(i_,e_).interp.vert ? "|" : "-",
			   (error - old_error)/ERR_FACTOR);
		  sum += error - old_error;
		  POSITION(i_,e_).interp = buffer;
		} else {
		  if(!depended) {
		    lfOutput("      no dependency:\n");
		  }
		  lfOutput("      old diff %d/%d: %d\n",
			   i_, e_, POSITION(i_,e_).contribs[i * Columns + e]);
		  sum += POSITION(i_,e_).contribs[i * Columns + e];
		}
	      }
	    }
	  }
	  /* set new sum */
	  lfOutput("      => new sum %f\n", sum);
	  position->sum = sum;
	  
	  /* undo simulation */
	  DropSample(i,e, head, FALSE);
	}
      }
    }
  }
  TimeStamp("dropped %d", NumInterp);
  DumpDropped(CurrentSet);

  lfOutput("\nerror %f, %d/%d interpolated => avg. error %f\n",
	   Error/ERR_FACTOR, NumInterp, Rows * Columns, Error/NumInterp/ERR_FACTOR);
  if(Clamp) {
    lfOutput("clamp counter %d\n", ClampCount);
  }
  DoInteraction("done", 0, 0, LEVEL_DONE, FALSE);
  if(!Interactive) {
    return;
  } else {
    exit(0);
  }
}


/****************  interaction part ****************/
enum {
  ORIGINAL,
  ERRORS,
  INTERP,
  SINFL,
  AINFL
} Display;
int InflI, InflE, RealInflI, RealInflE;
static unsigned char *GetCurSource(int i, int e)
{
  unsigned char *pixels;
  struct Position *position = &POSITION(i,e);
  int j;

  switch(Display) {
  case ERRORS:
    if(!position->sampled && position->image) {
      InterpolatePosition(i,e, -1, FALSE, FALSE);
      for(j = 0; j < ImageSize; j++) {
	Left[j] = abs(position->image[j] - Interp[j]);
      }
      pixels = Left;
    } else {
      pixels = NULL;
    }
    break;
  case INTERP:
    if(position->sampled) {
      pixels = position->image;
    } else {
      InterpolatePosition(i,e, -1,FALSE,FALSE);
      pixels = Interp;
    }
    break;
  default:
    pixels = position->image;
    break;
  }
  return pixels;
}

static bool_t DisplayDirection, DisplayBorders = TRUE;
static void Redisplay(void)
{
  int i,e, j;
  int p[2];

  if(NoDisplay)
    return;

  if(BW) {
    char buffer[80];
    sprintf(buffer, "%d/%d", OP[1],OP[0]);
    strcat(buffer, PVertOnly ? " v" : PHorizOnly ? " h" : "");
    strcat(buffer, Linear ? " l" : "");
    glutSetWindowTitle(buffer);
  }

  glClear(GL_COLOR_BUFFER_BIT);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(-1.0, -1.0, 0.0);
  glScalef(2.0/win_width, 2.0/win_height, 1.0);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glViewport(0,0, win_width, win_height);
  glColor3f(1.0, 1.0, 1.0);

  switch(Display) {
  case SINFL:
    RealInflI = -1;
    RealInflE = -1;
    for(i = 0; i < Rows; i++) {
      for(e = 0; e < Columns; e++) {
	if(POSITION(i,e).sampled) {
	  if(abs(i - InflI) < abs(i - RealInflI) ||
	     (i == RealInflI && abs(e - InflE) < abs(e - RealInflE))) {
	    RealInflE = e;
	    RealInflI = i;
	  }
	}
      }
    }
    for(p[1] = POSITION(0,0).p[1], i = 0; p[1] <= POSITION(Rows-1,0).p[1]; p[1]++) {
      int l,r,t,b, dp;

      if(POSITION(i + 1, 0).p[1] < p[1]) {
	i++;
      }
      dp = POSITION(i+1,0).p[1] - POSITION(i,0).p[1];
      if(p[1] == POSITION(0,0).p[1]) {
	b = 0;
      } else {
	b = ((p[1] - POSITION(i,0).p[1]) * height - height/2 + dp/2) / dp + i * height + height/2;
      }
      if(p[1] == POSITION(Rows-1,0).p[1]) {
	t = win_height;
      } else if(POSITION(i + 1, 0).p[1] < p[1] + 1) {
	t = (i + 2) * height;
      } else {
	t = ((p[1] + 1 - POSITION(i,0).p[1]) * height - height/2 + dp/2) / dp + i * height + height/2;
      }

      for(p[0] = POSITION(0,0).p[0], e = 0; p[0] <= POSITION(0,Columns-1).p[0]; p[0]++) {
	unsigned char color;

	if(POSITION(0,e + 1).p[0] < p[0]) {
	  e++;
	}
	color = INFLUENCE(&POSITION(RealInflI,RealInflE), p[1], p[0]);
	glColor3ub(color, color, color);

	dp = POSITION(0,e+1).p[0] - POSITION(0,e).p[0];
	if(p[0] == POSITION(0,0).p[0]) {
	  l = 0;
	} else {
	  l = ((p[0] - POSITION(0,e).p[0]) * width - width/2) / dp + e * width + width/2;
	}
	if(p[0] == POSITION(0,Columns-1).p[0]) {
	  r = win_width;
	} else if(POSITION(0,e + 1).p[0] < p[0] + 1) {
	  r = (e + 2) * width;
	} else {
	  r = ((p[0] + 1 - POSITION(0,e).p[0]) * width - width/2) / dp + e * width + width/2;
	}
	glRecti(l,b, r,t);
      }
    }
    break;
  default: 
    for(i = 0; i < Rows; i++) {
      for(e = 0; e < Columns; e++) {
	unsigned char *pixels;

	glRasterPos2f(e * width, i * height);
	pixels = GetCurSource(i,e);
	if(pixels) {
	  if(Dummy) {
	    glColor3ub(pixels[0], pixels[0], pixels[0]);
	    glRecti(e * width, i * height, (e + 1) * width, (i + 1) * height);
	  } else {
	    glDrawPixels(width, height, GL_RGB, GL_UNSIGNED_BYTE, pixels);
	  }
	}
      }
    }
    break;
  }

  for(i = 0; i < Rows; i++) {
    for(e = 0; e < Columns; e++) {
      struct Position *position = &POSITION(i,e);
      int ssize;
        
      if(!position->sampled) {
	glColor3f(1.0, 1.0, 1.0);
      } else {
	glColor3f(1.0, 0.0, 0.0);
	if(BW) {
	  for(j = 0; j < PSources.num; j++) {
	    if(PSources.sources[j][0] == i && PSources.sources[j][1] == e) {
	      glColor3f(0.0, PSources.sfactor[j], 0.0);
	      break;
	    }
	  }
	}
      }
      if(DisplayBorders) {
	glBegin(GL_LINE_STRIP);
	glVertex2i(e * width, i * height);
	glVertex2i(e * width, (i + 1) * height - 1);
	glVertex2i((e + 1) * width - 1, (i + 1) * height - 1);
	glVertex2i((e + 1) * width - 1, i * height);
	glVertex2i(e * width, i * height);
	glEnd();
      
	if(width > 32 && height > 32) {
	  ssize = 5;
	} else {
	  ssize = 3;
	}
	/* draw marks for sources */
	for(j = 0; j < position->interp.sources[position->interp.vert].num; j++) {
	  int x, y;
	  x = position->interp.sources[position->interp.vert].sources[j][1] > e ?
	    1 : position->interp.sources[position->interp.vert].sources[j][1] < e ? -1 : 0;
	  y = position->interp.sources[position->interp.vert].sources[j][0] > i ?
	    1 : position->interp.sources[position->interp.vert].sources[j][0] < i ? -1 : 0;
	  x = e * width + width / 2 + x * (width / 2 - ssize) - 1;
	  y = i * height + height / 2 + y * (height / 2 - ssize) - 1;
	  glRecti(x,y, x + ssize - 1, y + ssize - 1);
	}
      }

      if(!position->sampled && DisplayDirection) {
	if(position->interp.vert) {
	  glBegin(GL_LINE_STRIP);
	  glVertex2i(e * width + width/2, i * height);
	  glVertex2i(e * width + width/2, (i + 1) * height);
	  glEnd();
	} else {
	  glBegin(GL_LINE_STRIP);
	  glVertex2i(e * width, i * height + height / 2);
	  glVertex2i((e + 1) * width, i * height + height / 2);
	  glEnd();
	}
      }
    }
  }
  glFlush();
}

static void SaveImageBuffer(void)
{
  char buffer[1024];
  unsigned char *pixels, *source, *target;
  int line_offset = (Columns - 1) * 3 * width;
  int i, e, x, y;

  printf("Please enter filename: ");
  fflush(stdout);
  fgets(buffer, sizeof(buffer), stdin);
  target = strchr(buffer, '\n');
  if(target) {
    *target = 0;
  }
  if((pixels = malloc(ImageSize * Rows * Columns))) {
    if(0) {
      for(i = 0; i < Rows; i++) {
	for(e = 0; e < Columns; e++) {
	  source = GetCurSource(i,e);
	  target = &pixels[(i * height + e) * width * 3];
	  for(y = 0; y < height; y++, target += line_offset) {
	    for(x = 0; x < width; x++) {
	      *target++ = *source++;
	    }
	  }
	}
      }
    } else {
      glReadPixels(0, 0, win_width, win_height, GL_RGB, GL_UNSIGNED_BYTE, pixels);
    }
    WriteRGBImage(buffer, win_width, win_height, GL_RGB, pixels);
    free(pixels);
  } else {
    fprintf(stderr, "no memory!\n");
  }
}

static int GetPosition(bool_t reset)
{
  static char buffer[256];
  static char *pos;
  int index = -1;

  if(reset) {
    fprintf(stderr, "position(s) index|rows,column ...: ");
    fgets(buffer, sizeof(buffer), stdin);
    pos = buffer;
  }

  while(*pos) {
    while(isspace(*pos)) {
      pos++;
    }
    if(*pos) {
      index = atoi(pos);
      while(isdigit(*pos)) {
	pos++;
      }
      if(*pos == ',') {
	pos++;
	index = index * Columns + atoi(pos);
	while(isdigit(*pos)) {
	  pos++;
	}
      }
      while(isspace(*pos)) {
	pos++;
      }
      break;
    }
  }

  return index;
}

static bool_t DoInteraction(char *action, int i, int e, int level, bool_t display_only)
{
  bool_t cont = TRUE;

  while(cont) {
    char input[10];

    if(display_only || level < Interactive || !Interactive) {
      lfOutput("%s %d/%d\n", action, i, e);
      cont = FALSE;
    } else {
      Redisplay();
      printf("%s %d/%d, (e)rrors, (o)riginal, (i)nterpolated, (s)ave, (q)uit, (b)rowse, (c)hange, (level) %d/%d RETURN: ",
	     action, i, e, level, Interactive);
      fgets(input, sizeof(input), stdin);
      switch(input[0]) {
      case 0:
      case '\r':
      case '\n':
	cont = FALSE;
	break;
      case 'e':
	Display = ERRORS;
	break;
      case 'o':
	Display = ORIGINAL;
	break;
      case 'i':
	Display = INTERP;
	break;
      case 's':
	SaveImageBuffer();
	break;
      case 'c': {
	bool_t reset = TRUE;
	int index;

	while((index = GetPosition(reset)) >= 0 && index < Columns * Rows) {
	  reset = FALSE;
	  if(Positions[index].sampled) {
	    DropSample(Positions[index].i, Positions[index].e, NULL, TRUE);
	    NumInterp++;
	  } else {
	    AddSample(Positions[index].i, Positions[index].e);
	    NumInterp--;
	  }
	}
	if(!reset) {
	  RecalcAll();
	  return TRUE;
	}
	break;
      }
      case 'q':
	DumpDropped(CurrentSet);
	PrintInterpolation(res);
	exit(0);
	break;
      case 'b':
	DumpDropped(CurrentSet);
	InitBrowsing();
	return TRUE;
	break;
      default:
	if(isdigit(input[0])) {
	  Interactive = atoi(input);
	}
	break;
      }
    }
  }
  return FALSE;
}

static void DumpFunction(int scale, int cur, int next, bool_t neg)
{
  int d = next - cur, i;
  int gdiff = Positions[next].p[0] - Positions[cur].p[0];
  fprintf(res, "range(x, %d, %d) * %s(1",
	  Positions[cur].p[0],
	  Positions[next].p[0], neg ? "(1 - " : "");
  if(Smooth) {
    for(i = 0; i <= d; i++) {
      fprintf(res, " + %e * ((x - %d) / %d)**%d",
	      Positions[cur].cx[d][d - i],
	      Positions[cur].p[0],
	      gdiff, i + 2);
    }
  } else {
    for(i = 0; i < d; i++) {
      fprintf(res, " + %e * ((x - %d) / %d)**%d",
	      Positions[cur].cx[d][d - i - 1],
	      Positions[cur].p[0],
	      gdiff, i + 1);
    }
  }
  fprintf(res, neg ? "))" : ")");
}


static void SaveState(void)
{
  if(Dummy || Basis) {
    int left, right, scale;
    int i, j, first, prev, next, num_samples = 0, max_pos;
    double maximum, factor;
    int entries = Rows * Columns;

    left = Minimum[0] - (Maximum[0] - Minimum[0]) / values[0] / 2;
    right = Maximum[0] + (Maximum[0] - Minimum[0]) / values[0] / 2;
    if(Dummy) {
      scale = 1;
    } else {
      scale = width;
    }

    fprintf(res,
	    "\n\n# dropped %d samples\n"
	    "range(x, x1,x2) = (x >= x1 && x < x2) ? 1 : 0\n"
	    "# basis functions\n",
	    NumInterp);

    /* search first sampled position */
    for(i = 0; i < entries && !Positions[i].sampled; i++)
      ;
    first = i;

    /* define basis functions */
    for(i = first, prev = -1; i < entries; prev = i, i = next) {
      /* search next valid sample */
      for(next = i + 1; next < entries && !Positions[next].sampled; next++)
	;

      if(prev == -1) {
	fprintf(res, "B%dl(x) = range(x, %d, %d)\n", i, left, Positions[i].p[0]);
      }
      fprintf(res, "B%dr(x) = ", i);
      if(next < entries) {
	DumpFunction(scale, i, next, FALSE);
      } else {
	fprintf(res, "range(x, %d, %d)", Positions[i].p[0], right);
      }
      fprintf(res, "\nB%d(x) = (B%dl(x) + B%dr(x))\n", i, i, i);
      if(next < entries) {
	fprintf(res, "B%dl(x) = ", next);
	DumpFunction(scale, i, next, TRUE);
      }
      fprintf(res, "\n");
    }

    /* dump samples and errors for Dummy mode */
    if(Dummy) {
      for(i = first; i < entries; prev = i, i = next) {
	/* search next valid sample */
	for(next = i + 1; next < entries && !Positions[next].sampled; next++)
	  ;
	for(j = 0; j < NumDummyFuncs; j++) {
	  fprintf(res, "#samples%02d %d %d\n", NumInterp,
		  Positions[i].p[0], Positions[i].image[j]);
	}
	num_samples++;
      }
      for(i = 0; i < NumInterp; i++) {
	for(j = 0; j < NumDummyFuncs; j++) {
	  InterpolatePosition(0, Dropped[i], -1, FALSE,FALSE);
	  fprintf(res, "#errors%02d %d %d %d %d\n", NumInterp,
		  Positions[Dropped[i]].p[0], Interp[j],
		  Interp[j], Positions[Dropped[i]].image[j]);
	}
      }

      /* search scale factor for basis functions */
      for(i = first, prev = -1, maximum = 1; i < entries; prev = i, i = next) {
	/* search next valid sample */
	for(next = i + 1; next < entries && !Positions[next].sampled; next++)
	  ;
	if(next < entries) {
	  int d = next - i;
	  double *c = Positions[i].cx[d];
	  
	  for(j = Positions[i].p[0]; j < Positions[next].p[0]; j++) {
	    double val = CalcFactor(c, d, Positions[i].p[0], j, Positions[i + d].p[0] - Positions[i].p[0]);
	    val = max(fabs(val), fabs(1 - val));
	    if(val > maximum) {
	      maximum = val;
	      max_pos = j;
	    }
	  }
	}
      }
      factor = 254 / maximum;
    } else {
      itp_debug = TRUE;
      for(i = 0; i < NumInterp; i++) {
	InterpolatePosition(0, Dropped[i], -1, FALSE,FALSE);
      }
      itp_debug = FALSE;

      factor = 1;
    }

    /* plot: function */
    fprintf(res, "set samples %d\n", (right - left) * scale);
    if(Dummy) {
      fprintf(res, "set key %d,-150\n", right - (right - left) / 15);
      fprintf(res, "set title \"%d out of %d interpolated => error %.2f / basis functions scaled with %.3f\"\n",
	      NumInterp, entries, Error/ERR_FACTOR, factor);
    } else {
      fprintf(res, "set title \"%d out of %d (%.2f)%s",
	      NumInterp, entries, Error/ERR_FACTOR,
	      NumInterp ? ": dropped " : "");
      for(i = 0; i < NumInterp; i++) {
	InterpolatePosition(0, Dropped[i], -1,FALSE,FALSE);
	fprintf(res, "%d (%.2f)%s",
		Positions[Dropped[i]].p[0],
		(double)Positions[Dropped[i]].interp.error,
		i + 1 < NumInterp ? ", " : ""); 
      }
    }
    fprintf(res, "\nplot [%d : %d] %s \\\n", left, right - 1, Dummy ? "[-300 : 300]" : "");
    if(Dummy) {
      for(j = 0; j < NumDummyFuncs; j++) {
	fprintf(res, "%s with lines 1,\\\n", DummyFuncNames[j]);
      }

      /* interpolated function(s) */
      for(j = 0; j < NumDummyFuncs; j++) {
	for(i = first, prev = -1; i < entries; prev = i, i = next) {
	  /* search next valid sample */
	  for(next = i + 1; next < entries && !Positions[next].sampled; next++)
	    ;
	  fprintf(res, "%d*B%d(x)", Positions[i].image[j], i);
	  fprintf(res, 
		  next < entries ? "+" :
		  j == 0 ? " title \"interpolated functions\" with lines 2,\\\n"
		  : " notitle with lines 2,\\\n");
	}
      }

      /* samples */
      fprintf(res, "\"< grep \\^\\#samples%02d %s | cut -f2- -d' '\" title \"%d samples\",\\\n",
	      NumInterp, destname, num_samples);

      /* errors */
      if(NumInterp) {
	fprintf(res, "\"< grep \\^\\#errors%02d %s | cut -f2- -d' '\" title \"dropped", NumInterp, destname);
	for(i = 0; i < NumInterp; i++) {
	  fprintf(res, " %d", Positions[Dropped[i]].p[0]);
	}
	fprintf(res, "\" with errorbars,\\\n");
      }
    }

    /* basis functions */
    for(i = first, prev = -1; i < entries; prev = i, i = next) {
      /* search next valid sample */
      for(next = i + 1; next < entries && !Positions[next].sampled; next++)
	;

      fprintf(res, "%.2f*B%d(x) %s with lines 3",
	      factor, i, i == first ? "title \"basis functions\"" : "notitle");
      fprintf(res, next < entries ? ",\\\n" : "\n");
    }
    fprintf(res, "pause -1\n");
    fflush(res);
  }
}


/* interpolation of any parameter values based on the current sampling */

static void InterpolateParameters(void)
{
  int i, e;
  struct Position *position;
  int offset[2], ycloser;
  int sign;

  /* clamp parameters */
  if(P[0] < POSITION(0,0).p[0]) {
    P[0] = POSITION(0,0).p[0];
  } else if(P[0] > POSITION(0,Columns - 1).p[0]) {
    P[0] = POSITION(0,Columns - 1).p[0];
  }
  if(P[1] < POSITION(0,0).p[1]) {
    P[1] = POSITION(0,0).p[1];
  } else if(P[1] > POSITION(Rows - 1,0).p[1]) {
    P[1] = POSITION(Rows - 1,0).p[1];
  }
    

  /* find lower left edge of tile */
  for(i = 0; i + 1 < Rows && P[1] >= POSITION(i+1,0).p[1]; i++)
    ;
  for(e = 0; e + 1 < Columns && P[0] >= POSITION(0,e+1).p[0]; e++)
    ;
  position = &POSITION(i,e);

  offset[1] = i + 1 < Rows &&
    P[1] >= (POSITION(i+1,0).p[1] + POSITION(i,0).p[1] + 1)/2;
  offset[0] = e + 1 < Columns &&
    P[0] >= (POSITION(0,e+1).p[0] + POSITION(0,e).p[0] + 1)/2;
  sign = offset[0] == offset[1] ? 1 : -1;
  ycloser =  e + 1 >= Columns ||
    (i + 1 < Rows &&
     /* (p[1] - POSITION(i,0).p[1]) / (p[0] - POSITION(0,e).p[0]) >
	(POSITION(i+1,0).p[1] - POSITION(i,0).p[1]) / (POSITION(0,e+1) - POSITION(0,e).p[0]) */
     sign * (P[1] - POSITION(i+offset[1],0).p[1]) * (POSITION(0,e+(!offset[0])).p[0] - POSITION(0,e+offset[0]).p[0]) >
     sign * (POSITION(i+(!offset[1]),0).p[1] - POSITION(i+offset[1],0).p[1]) * (P[0] - POSITION(0,e+offset[0]).p[0]));

#if 0
  printf("%d/%d => ll %d/%d=%d/%d, y/xoff %d/%d, ycloser %d\n",
	 P[1], P[0], POSITION(i,0).p[1], POSITION(0,e).p[0], i,e,
	 offset[1], offset[0], ycloser);
#endif

  if(POSITION(i+offset[1],e+offset[0]).sampled) {
    offset[ycloser] = !offset[ycloser];
#if 0
    printf("   => y/x %d/%d\n",
	   i+offset[1], e+offset[0]);
#endif
    if(i+offset[1] >= Rows || e+offset[0] >= Columns ||
       POSITION(i+offset[1],e+offset[0]).sampled) {
      offset[ycloser] = !offset[ycloser];
      offset[!ycloser] = !offset[!ycloser];
#if 0
      printf("   => y/x %d/%d\n",
	     i+offset[1], e+offset[0]);
#endif
      if(i+offset[1] >= Rows || e+offset[0] >= Columns ||
	 POSITION(i+offset[1],e+offset[0]).sampled) {
	offset[ycloser] = !offset[ycloser];
#if 0
	printf("   => y/x %d/%d\n",
	       i+offset[1], e+offset[0]);
#endif
	if(i+offset[1] >= Rows || e+offset[0] >= Columns) {
	  offset[0] = 0;
	  offset[1] = 0;
	}
      }
    }
  }
  if(!POSITION(i+offset[1],e+offset[0]).sampled) {
    PVert = POSITION(i+offset[1],e+offset[0]).interp.vert;
  } else {
    if(P[0] == position->p[0]) {
      PVert = TRUE;
    } else {
      PVert = FALSE;
    }
  }
#if 0
  printf("   => y/x %d/%d, vert %d\n",
	 i+offset[1], e+offset[0], PVert);
#endif

  memset(&PSources, 0, sizeof(PSources));
  if(PHorizOnly) {
    PVert = FALSE;
  } else if(PVertOnly) {
    PVert = TRUE;
  }
  InterpolateDirection(i,e, P, PVert, Interp, &PSources);
}

static void PrintInterpPos(void)
{
  int i;
  bool_t area = __LF_GET_PARAMS()->num > 1;

  InterpolateParameters();
  fprintf(res, "a%d", PSources.num);
  for(i = 0; i < PSources.num; i++) {
    struct Position *position = &POSITION(PSources.sources[i][0], PSources.sources[i][1]);
    
    fprintf(res, "_%d", position->index[0]);
    if(area) {
      fprintf(res, "_%d", position->index[1]);
    }
  }
  fprintf(res, "_%d", PSources.fnum);
  for(i = PSources.fnum - 1; i >= 0; i--) {
    fprintf(res, "_%d_%d_%f",
	    PSources.source_indices[i][0], 
	    PSources.source_indices[i][1],
	    PSources.factor[i] < 0 ? 1.0 :
	    PSources.factor[i] > 1 ? 0.0 : 1.0 - PSources.factor[i]);
  }
  fprintf(res, "\n");
}

static void PrintInterpolation(FILE *res)
{
  int i, e;
  int index0 = 0, index1;
  bool_t area = __LF_GET_PARAMS()->num > 1;

  if(Swapped) {
    /* print out sampling positions and determine indices */
    for(i = 0; i < Rows; i++) {
      index1 = 0;
      for(e = 0; e < Columns; e++) {
	struct Position *position = &POSITION(i,e);
	
	if(position->sampled) {
	  fprintf(res, "SAMPLE_POINT %d", position->p[1]);
	  fprintf(res, area ? " %d\n" : "\n",
		  position->p[0]);
	  position->index[0] = index0;
	  position->index[1] = index1;
	  index1++;
	}
      }
      if(index1 > 0) {
	index0++;
      }
    }

    fprintf(res, "\nINTERPOLATION\n");
    for(P[0] = POSITION(0,0).p[0]; P[0] <= POSITION(0,Columns-1).p[0]; P[0]++) {
      for(P[1] = POSITION(0,0).p[1]; P[1] <= POSITION(Rows-1,0).p[1]; P[1]++) {
	PrintInterpPos();
      }
    }
  } else {
    /* print out sampling positions and determine indices */
    for(e = 0; e < Columns; e++) {
      index1 = 0;
      for(i = 0; i < Rows; i++) {
	struct Position *position = &POSITION(i,e);
	
	if(position->sampled) {
	  fprintf(res, "SAMPLE_POINT %d", position->p[0]);
	  fprintf(res, area ? " %d\n" : "\n",
		  position->p[1]);
	  position->index[0] = index0;
	  position->index[1] = index1;
	  index1++;
	}
      }
      if(index1 > 0) {
	index0++;
      }
    }

    fprintf(res, "\nINTERPOLATION\n");
    for(P[1] = POSITION(0,0).p[1]; P[1] <= POSITION(Rows-1,0).p[1]; P[1]++) {
      for(P[0] = POSITION(0,0).p[0]; P[0] <= POSITION(0,Columns-1).p[0]; P[0]++) {
	PrintInterpPos();
      }
    }
  }
}


/* browsing the current solution with a second window */

void MWkbf(unsigned char key, int x, int y)
{
  bool_t redisplay = FALSE;
  bool_t Bredisp = FALSE;

  switch(key) {
  case 'e':
    Display = ERRORS;
    redisplay = TRUE;
    break;
  case 'o':
    Display = ORIGINAL;
    redisplay = TRUE;
    break;
  case 'i':
    Display = INTERP;
    redisplay = TRUE;
    break;
  case 'f':
    Display = SINFL;
    redisplay = TRUE;
    break;
  case 'F':
    Display = AINFL;
    redisplay = TRUE;
    break;
  case 's':
    SaveImageBuffer();
    break;
  case 'v':
    PVertOnly = !PVertOnly;
    if(PVertOnly) {
      PHorizOnly = FALSE;
    }
    Bredisp = TRUE;
    break;
  case 'h':
    PHorizOnly = !PHorizOnly;
    if(PHorizOnly) {
      PVertOnly = FALSE;
    }
    Bredisp = TRUE;
    break;
  case 'l':
    Linear = !Linear;
    Bredisp = TRUE;
    break;
  case 'd':
    DisplayDirection = !DisplayDirection;
    redisplay = TRUE;
    break;
  case 'b':
    DisplayBorders = !DisplayBorders;
    redisplay = TRUE;
    break;
  }
   
  if(redisplay) {
    glutPostRedisplay();
  }
  if(Bredisp) {
    glutSetWindow(BW);
    glutPostRedisplay();
  }
}

static void MWsf(int key, int x, int y)
{
  bool_t redisplay = TRUE;

  switch(key) {
  case GLUT_KEY_LEFT:
    if(RealInflE > 0) {
      InflE = POSITION(RealInflI, RealInflE).left /*LNeigh[RealInflE - 1]*/;
    } else {
      InflE = 0;
    }
    break;
  case GLUT_KEY_RIGHT:
    if(RealInflE + 1 < Columns) {
      InflE = POSITION(RealInflI, RealInflE).right /*RNeigh[RealInflE + 1]*/;
    } else {
      InflE = Columns - 1;
    }
    break;
  case GLUT_KEY_DOWN:
    if(RealInflI > 0) {
      InflI = BNeigh[RealInflI - 1];
    } else {
      InflI = 0;
    }
    break;
  case GLUT_KEY_UP:
    if(RealInflI + 1 < Rows) {
      InflI = TNeigh[RealInflI + 1];
    } else {
      InflI = Rows - 1;
    }
    break;
  default:
    redisplay = FALSE;
      break;
  }
  if(redisplay) {
    glutPostRedisplay();
  }
}

static void MWmotionf(int x, int y)
{
  int i,e;

  y = win_height - y - 1;

  e = (x - width/2) / width;
  if(e < 0) {
    e = 0;
  }
  if(Columns > 1) {
    if(e + 1 >= Columns) {
      e = Columns - 2;
    }
    OP[0] = P[0] =
      ((x - e * width - width/2) * (POSITION(0,e+1).p[0] - POSITION(0,e).p[0]) + width/2) / width
      + POSITION(0,e).p[0];
  } else {
    OP[0] = P[0] = POSITION(0,e).p[0];
  }
  i = (y - height/2) / height;
  if(i < 0) {
    i = 0;
  }
  if(Rows > 1) {
    if(i + 1 >= Rows) {
      i = Rows - 2;
    }
    OP[1] = P[1] =
      ((y - i * height - height/2) * (POSITION(i+1,0).p[1] - POSITION(i,0).p[1]) + height/2) / height
      + POSITION(i,0).p[1];
  } else {
    OP[1] = P[1] = POSITION(i,0).p[1];
  }
  glutSetWindow(BW);
  glutPostRedisplay();
}

static void MWmousef(int button, int state, int x, int y)
{
  if(state == GLUT_DOWN) {
    MWMouseDown++;
    MWmotionf(x,y);
  } else {
    MWMouseDown--;
    if(!MWMouseDown) {
      glutSetWindow(BW);
      glutPostRedisplay();
    }
  }
}


static void BWdisplay(void)
{
  char buffer[80];

  glClear(GL_COLOR_BUFFER_BIT);
  InterpolateParameters();
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(-1.0, -1.0, 0.0);
  glScalef(1.0/width, 1.0/height, 1.0);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glViewport(0,0, 2*width, height);
  glRasterPos2f(0,0);
  glDrawPixels(width, height, GL_RGB, GL_UNSIGNED_BYTE, Interp);
  sprintf(buffer, "(%s%s) %d/%d",
	  PVert ? "|" : "-", Linear ? "l" : "", P[1],P[0]);
  if(!MWMouseDown) {
    RenderImage(P, Left);
    glRasterPos2f(width,0);
    glDrawPixels(width, height, GL_RGB, GL_UNSIGNED_BYTE, Left);
    sprintf(&buffer[strlen(buffer)], ", %.2f",
	    (double)diff_mem(Left,Interp,ImageSize));
  }
  glutSetWindowTitle(buffer);

  if(DisplayBorders) {
    glutSetWindow(MW);
    glutPostRedisplay();
  }
}

static void InitBrowsing(void)
{
  int i,e,j;

  /* change main window */
  glutDisplayFunc(Redisplay);
  glutKeyboardFunc(MWkbf);
  glutMotionFunc(MWmotionf);
  glutMouseFunc(MWmousef);
  glutSpecialFunc(MWsf);

  /* do new window */
  glutInitWindowSize(width * 2, height);
  glutInitDisplayMode(0);
  BW = glutCreateWindow("panalysis");
  glutDisplayFunc(BWdisplay);

  /* init influence maps */
  for(i = 0; i < Rows; i++) {
    for(e = 0; e < Columns; e++) {
      struct Position *position = &POSITION(i,e);

      if(position->sampled) {
	if(!(position->influence = calloc(INFL_SIZE, sizeof(position->influence[0])))) {
	  lfError("no mem!\n");
	  exit(10);
	}
      }
    }
  }
  for(P[1] = POSITION(0,0).p[1]; P[1] <= POSITION(Rows-1,0).p[1]; P[1]++) {
    for(P[0] = POSITION(0,0).p[0]; P[0] <= POSITION(0,Columns-1).p[0]; P[0]++) {
      InterpolateParameters();
      for(j = 0; j < PSources.num; j++) {
	INFLUENCE(&POSITION(PSources.sources[j][0], PSources.sources[j][1]),
		  P[1], P[0]) = PSources.sfactor[j] >= 1.0 ? 255 :
	  PSources.sfactor[j] <= 0.0 ? 0 : PSources.sfactor[j] * 255;
      }
    }
  }
}
